//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.
//
//	Look-and-feel inspired by Xmol 
//	(http://www.networkcs.com/msc/docs/xmol/).
//
//	OpenGL selection code and quaterion rotation matrix
//	formulation from molview demo program by Mark Kilgard
//	(http://reality.sgi.com/opengl/OpenGLforX.html). PostScript
//	rendering also from Mark Kilgard.
//
//	The printing routines are based on code from
//	Craig Fahrnbach <craignan@home.com> via
//	Uwe Kotyczka <kotyczka@bnm-gmbh.de>.
//
//	Covalent bond determination code based on VMD
//	http://www.ks.uiuc.edu/Research/vmd/
//
//	Stride (http://www.embl-heidelberg.de/stride/stride.html)
//	is used for secondary structure determination and has been 
//	included in Qmol with the kind permission of 
//	Dmitrij Frishman, PhD
//	Institute for Bioinformatics
//	GSF - Forschungszentrum fr Umwelt und Gesundheit, GmbH
//	Ingolstdter Landstrae 1,
//	D-85764 Neuherberg, Germany
//	Telephone: +49-89-3187-4201
//	Fax: +49-89-31873585
//	e-mail: d.frishman@gsf.de
//	WWW: http://mips.gsf.de/mips/staff/frishman/

#include "stdafx.h"

#include <math.h>
#include "ClipPlane.h"
#include "LinearAlgebra3D.h"

#define		MAX_CLIP_DIST		7.0f
#define		CLIP_PLANE_LENGTH	2.0f
#define		CLIP_EPSILON		1.0e-6f

ClipPlane::ClipPlane()
{
	// Use GL_CLIP_PLANE0 for now
	clip_id = GL_CLIP_PLANE0;

	stereo_rot = 0.0f;

	show_plane = FALSE;
	enable_clipping = FALSE;

	center[X] = 0.0f;
	center[Y] = 0.0f;
	center[Z] = 0.0f;

	// Tilt the clipping plane up to show off the
	// atom colors (for clipping spheres).
	normal[X] = 0.0f;
	normal[Y] = 0.1f;
	normal[Z] = 1.0f;

	SetEqn(center, normal);

	UpdateCoordSystem();

	// Default color (transparent grey)
	clip_plane_color[X] = 0.5f;
	clip_plane_color[Y] = 0.5f;
	clip_plane_color[Z] = 0.5f;
	clip_plane_color[3] = 0.5f;

	// Initialize the track ball
	trackball(0.0, 0.0, 0.0, 0.0);
}

void ClipPlane::Enable()
{
	enable_clipping = TRUE;

	glEnable(clip_id);
}

void ClipPlane::Disable()
{
	enable_clipping = FALSE;

	glDisable(clip_id);
}

void ClipPlane::EnableDisplay()
{
	show_plane = TRUE;
};

void ClipPlane::DisableDisplay()
{
	show_plane = FALSE;
};

// Clip the current view
void ClipPlane::Clip(GLfloat m_stereo_rot)
{
	stereo_rot = m_stereo_rot;

	// Allways apply the clipping equation since
	// the clip plane may be enabled later in the rendering
	// pipe-line (as when only clipping a molecular surface).
	//if(enable_clipping){
		glPushMatrix();

		glRotatef(stereo_rot, 0.0f, 1.0f, 0.0f);

		glClipPlane(clip_id, eqn);

		glPopMatrix();
	//}
};

// Center the clipping plane in the window viewport
void ClipPlane::CenterClip()
{
	center[X] = 0.0f;
	center[Y] = 0.0f;
	center[Z] = 0.0f;

	eqn[3] = 0.0f; // normal*center
}

void ClipPlane::SetEqn(GLfloat m_center[3], GLfloat m_normal[3],
					   BOOL m_norm)
{
	if(m_norm){
		GLfloat mag = DOT(m_normal, m_normal);

		if(mag <= 0)
			throw QmolError("Zero length normal vector!", 
				__FILE__, __LINE__);
		
		mag = 1.0f/(float)(sqrt(mag));

		normal[X] = m_normal[X]*mag;
		normal[Y] = m_normal[Y]*mag;
		normal[Z] = m_normal[Z]*mag;
	}
	else{
		normal[X] = m_normal[X];
		normal[Y] = m_normal[Y];
		normal[Z] = m_normal[Z];
	}

	center[X] = m_center[X];
	center[Y] = m_center[Y];
	center[Z] = m_center[Z];

	eqn[X] = -normal[X];
	eqn[Y] = -normal[Y];
	eqn[Z] = -normal[Z];
	eqn[3] = DOT(normal, center);

	UpdateCoordSystem();
}

void ClipPlane::SetColor(GLfloat m_color[4])
{
	clip_plane_color[X] = m_color[X];
	clip_plane_color[Y] = m_color[Y];
	clip_plane_color[Z] = m_color[Z];
	clip_plane_color[3] = m_color[3];
}

void ClipPlane::GetCenter(GLfloat m_center[3])
{
	center[X] = m_center[X];
	center[Y] = m_center[Y];
	center[Z] = m_center[Z];
}

void ClipPlane::GetNormal(GLfloat m_normal[3])
{
	normal[X] = m_normal[X];
	normal[Y] = m_normal[Y];
	normal[Z] = m_normal[Z];
}

void ClipPlane::UpdateCoordSystem()
{
	array<GLfloat> tmp(3);

	if(fabs(eqn[Z]) > CLIP_EPSILON){
		tmp[X] = 1.0f;
		tmp[Y] = 0.0f;
		tmp[Z] = 0.0f;
	}
	else{
		tmp[X] = 0.0f;
		tmp[Y] = 0.0f;
		tmp[Z] = 1.0f;
	}

	// A = normal | tmp;
	CROSS(A, normal, tmp);

	//B = A | normal;
	CROSS(B, A, normal);

	GLfloat mag = 1.0f/(GLfloat)(sqrt(DOT(A,A)));

	A[X] *= mag;
	A[Y] *= mag;
	A[Z] *= mag;

	mag = 1.0f/(GLfloat)(sqrt(DOT(B,B)));

	B[X] *= mag;
	B[Y] *= mag;
	B[Z] *= mag;
}

// Render the clip plane
void ClipPlane::RenderClipPlane()
{
	if(!show_plane)
		return;

	if(enable_clipping){
		// Make sure we don't clip the representation of the
		// clipping plane!
		glDisable(clip_id);
	}

	GLfloat tmp[3];

	glDisable(GL_CULL_FACE);

	glPushMatrix();

	glRotatef(stereo_rot, 0.0f, 1.0f, 0.0f);

#define ROUND_PLANE

#ifdef ROUND_PLANE
	glBegin(GL_TRIANGLE_FAN);
		glColor4fv(clip_plane_color);
		glNormal3fv(normal);

		// The center of the traingle fan
		glVertex3fv(center);

		// The edges of the fan ...
		const int clip_plane_res = 50;

		for(int i = 0;i <= clip_plane_res;i++){
			GLfloat tmp_a = (float)cos(i*2.0*PI/clip_plane_res)*CLIP_PLANE_LENGTH;
			GLfloat tmp_b = (float)sin(i*2.0*PI/clip_plane_res)*CLIP_PLANE_LENGTH;

			tmp[X] = center[X] + A[X]*tmp_a + B[X]*tmp_b;
			tmp[Y] = center[Y] + A[Y]*tmp_a + B[Y]*tmp_b;
			tmp[Z] = center[Z] + A[Z]*tmp_a + B[Z]*tmp_b;
			
			glVertex3fv(tmp);
		}

	glEnd();
#endif // ROUND_PLANE

#ifdef SQUARE_PLANE
	// Draw a quadrilateral to indicate the clipping plane
	glBegin(GL_QUADS);
		glColor4fv(clip_plane_color);
		glNormal3fv(normal);

		// Remember! The order we render the verticies determines
		// if this quad will (a) get culled and (b) receive the correct
		// lighting.
		tmp[X] = (-A[X] + B[X])*CLIP_PLANE_LENGTH + center[X];
		tmp[Y] = (-A[Y] + B[Y])*CLIP_PLANE_LENGTH + center[Y];
		tmp[Z] = (-A[Z] + B[Z])*CLIP_PLANE_LENGTH + center[Z];
		glVertex3fv(tmp);

		tmp[X] = (-A[X] - B[X])*CLIP_PLANE_LENGTH + center[X];
		tmp[Y] = (-A[Y] - B[Y])*CLIP_PLANE_LENGTH + center[Y];
		tmp[Z] = (-A[Z] - B[Z])*CLIP_PLANE_LENGTH + center[Z];
		glVertex3fv(tmp);

		tmp[X] = (A[X] - B[X])*CLIP_PLANE_LENGTH + center[X];
		tmp[Y] = (A[Y] - B[Y])*CLIP_PLANE_LENGTH + center[Y];
		tmp[Z] = (A[Z] - B[Z])*CLIP_PLANE_LENGTH + center[Z];
		glVertex3fv(tmp);

		tmp[X] = (A[X] + B[X])*CLIP_PLANE_LENGTH + center[X];
		tmp[Y] = (A[Y] + B[Y])*CLIP_PLANE_LENGTH + center[Y];
		tmp[Z] = (A[Z] + B[Z])*CLIP_PLANE_LENGTH + center[Z];
		glVertex3fv(tmp);

	glEnd();
#endif // SQUARE_PLANE

	glPopMatrix();

	glEnable(GL_CULL_FACE);

	if(enable_clipping){
		glEnable(clip_id);
	}
}

// Zoom
void ClipPlane::StartZoom(int x)
{
	begin_scale_x = x;
}

void ClipPlane::Zoom(int x, int width)
{
	GLfloat tmp = ((x-begin_scale_x)/(GLfloat)(width));

	// Translate the center vector along the normal vector
	center[X] -= normal[X]*tmp;
	center[Y] -= normal[Y]*tmp;
	center[Z] -= normal[Z]*tmp;

	// Make sure center doesn't get out of control
	tmp = DOT(center, center);

	if(tmp > MAX_CLIP_DIST*MAX_CLIP_DIST){
		tmp = MAX_CLIP_DIST/(GLfloat)(sqrt(tmp));
		center[X] *= tmp;
		center[Y] *= tmp;
		center[Z] *= tmp;
	}

	begin_scale_x = x;

	eqn[3] = DOT(center, normal);

	UpdateCoordSystem();
}

// Rotation
void ClipPlane::StartRotate(int x, int y)
{
	begin_rotation_x = x;
	begin_rotation_y = y;
}

void ClipPlane::Rotate(int x, int y, int width, int height)
{
	VirtualTrackBall::trackball( (2.0f*begin_rotation_x - width) / width,
    			(height - 2.0f*begin_rotation_y) / height,
    			(2.0f*x - width) / width,
    			(height - 2.0f*y) / height);
  	
	begin_rotation_x = x;
  	begin_rotation_y = y;

	GLfloat m[4][4];

	VirtualTrackBall::update();
	VirtualTrackBall::build_rotmatrix(m);

	// Transform the normal and plane equation according to the rotation matrix
	normal[X] = m[X][Z];
	normal[Y] = m[Y][Z];
	normal[Z] = m[Z][Z];

	eqn[X] = -normal[X];
	eqn[Y] = -normal[Y];
	eqn[Z] = -normal[Z];
	eqn[3] = DOT(normal,center);

	UpdateCoordSystem();
}

BOOL ClipPlane::DepthTestDisk()
{

	// Perform depth testing on the clipping disks.
	// We must take into account both perspective and
	// stereo rotations.
	GLfloat proj;
	
	// A 2D rotation matrix to correct for the 
	// stereo rotation about the y-axis (if any).
	GLfloat stereo_angle = 2.0f*PI*stereo_rot/360.0f;
	GLfloat c = (GLfloat)(cos(stereo_angle));
	GLfloat s = (GLfloat)(sin(stereo_angle));

	// The displacement vector describing the relative location
	// of the clippling plane to the eye cooridante system,
	//GLfloat D[3];
	
	//D[X] = center[X];
	//D[Y] = center[Y];
	//D[Z] = center[Z];

	// Note that the eye coordinates are displaced 
	// QM_DEFAULT_DEPTH along the z-axis.
	//D[Z] += QM_DEFAULT_DEPTH;

	// Compute normal dot D
	//proj = (c*normal[X] + s*normal[Z])*D[X] + 
	//	   normal[Y]*D[Y] + 
	//	   (c*normal[Z] - s*normal[X])*D[Z];
	
	// Here is a simplified form of the above code that does
	// not require a temporary vector.
	proj = (c*normal[X] + s*normal[Z])*center[X] + 
		   normal[Y]*center[Y] + 
		   (c*normal[Z] - s*normal[X])*(center[Z] + QM_DEFAULT_DEPTH);

	if(proj > 0.0f)
		return FALSE;

	return TRUE;
}

// Render a disk onto spheres intersected by the clippling plane
void ClipPlane::RenderClipSphere(array<atom> &atoms, int resolution, BOOL solvent,
								 GLfloat scaleFactor, 
								 GLfloat c_x, GLfloat c_y, GLfloat c_z,
								 GLfloat M[4][4])
{

	// We only need to render clipping disks if they are visible in
	// eye coordinates
	if(DepthTestDisk() == FALSE)
		return;

	GLfloat R[3], disk[3];
	int i, j;

	glDisable(clip_id);
	glDisable(GL_CULL_FACE);

	// Many of the disks will be exactly overlapping, so disable
	// depth testing. The disks will appear on the screen in the
	// reverse order of their position in the atom list.
	// Not that we don't want to use glDisable(GL_DEPTH_TEST) to
	// disable depth testing since we need entries to be written into the
	// depth buffer for comparison with objects rendered later.
	glDepthFunc(GL_ALWAYS);	

	glPushMatrix();
	glLoadIdentity();

	glTranslatef(0, 0, QM_DEFAULT_DEPTH);

	glScalef(scaleFactor, scaleFactor, scaleFactor);

	glRotatef(stereo_rot, 0.0f, 1.0f, 0.0f);

	// Dive once, multiuply many times
	scaleFactor = 1.0f/scaleFactor;

	// Allocate memory for disk verticies
	GLfloat *disk_x = new GLfloat[resolution + 1];
	GLfloat *disk_y = new GLfloat[resolution + 1];

	// Compute disk verticies
	for(j = 0;j <= resolution;j++){
		GLfloat theta = (2.0f*PI*j)/resolution;
		disk_x[j] = (GLfloat)(cos(theta));
		disk_y[j] = (GLfloat)(sin(theta));
	}

	for(i = 0;i < atoms.Length();i++){
		atom &atom_ref = atoms[i];

		// Only render solvent cross sections if requested
		if(!solvent && atom_ref.isSolvent)
			continue;

		GLfloat radius = atom_ref.radius;

		// R = M*r, the position of the rotated atom center in the frame
		// of the clipping plane.
		R[X] = M[X][X]*atom_ref.coord[X] + M[X][Y]*atom_ref.coord[Y] + M[X][Z]*atom_ref.coord[Z];
		R[Y] = M[Y][X]*atom_ref.coord[X] + M[Y][Y]*atom_ref.coord[Y] + M[Y][Z]*atom_ref.coord[Z];
		R[Z] = M[Z][X]*atom_ref.coord[X] + M[Z][Y]*atom_ref.coord[Y] + M[Z][Z]*atom_ref.coord[Z];

		R[X] += c_x*scaleFactor;
		R[Y] += c_y*scaleFactor;
		R[Z] += c_z*scaleFactor;

		// The shortest distance from the atom center to the clip plane
		// along the normal. 
		GLfloat dist;

		// dist = (R - center/scaleFactor)*normal
		dist = (R[X] - center[X]*scaleFactor)*normal[X];
		dist += (R[Y] - center[Y]*scaleFactor)*normal[Y];
		dist += (R[Z] - center[Z]*scaleFactor)*normal[Z];

		// Does the plane intersect the sphere?
		if(fabs(dist) < radius){
			// Draw the intersection of the clipping plane and the sphere
			GLfloat disk_radius = 
				(GLfloat)(sqrt(fabs(radius*radius - dist*dist)));
			
			// The center of the disk
			//disk = R - normal*dist;
			disk[X] = R[X] - normal[X]*dist;
			disk[Y] = R[Y] - normal[Y]*dist;
			disk[Z] = R[Z] - normal[Z]*dist;

			// Render the disk as a triangle fan
			glBegin(GL_TRIANGLE_FAN);
				glNormal3fv(normal);
				glColor4ubv(atom_ref.rgb_color);
				
				glVertex3fv(disk);

				for(j = 0;j <= resolution;j++){
					// Use R as scratch space
					//R = A*x + B*y + disk;
					//R[X] = A[X]*x + B[X]*y + disk[X];
					//R[Y] = A[Y]*x + B[Y]*y + disk[Y];
					//R[Z] = A[Z]*x + B[Z]*y + disk[Z];
					
					GLfloat tmp_x = disk_x[j]*disk_radius;
					GLfloat tmp_y = disk_y[j]*disk_radius;

					R[X] = A[X]*tmp_x + B[X]*tmp_y + disk[X];
					R[Y] = A[Y]*tmp_x + B[Y]*tmp_y + disk[Y];
					R[Z] = A[Z]*tmp_x + B[Z]*tmp_y + disk[Z];

					glVertex3fv(R);
				}

			glEnd();
		}
	}

	delete [] disk_x;
	delete [] disk_y;

	glPopMatrix();

	glEnable(clip_id);
	glEnable(GL_CULL_FACE);
	glDepthFunc(GL_LESS);
}

void ClipPlane::RenderClipSphere(int index, list< array<GLfloat> > &traj, 
		array<atom> &atoms, int resolution, BOOL solvent,
		GLfloat scaleFactor, GLfloat c_x, GLfloat c_y, GLfloat c_z,
		GLfloat M[4][4])
{

	// We only need to render clipping disks if they are visible in
	// eye coordinates
	if(DepthTestDisk() == FALSE)
		return;

	GLfloat R[3], disk[3];

	glDisable(clip_id);
	glDisable(GL_CULL_FACE);

	// Many of the disks will be exactly overlapping, so disable
	// depth testing. The disks will appear on the screen in the
	// reverse order of their position in the atom list.
	// Not that we don't want to use glDisable(GL_DEPTH_TEST) to
	// disable depth testing since we need entries to be written into the
	// depth buffer for comparison with objects rendered later.
	glDepthFunc(GL_ALWAYS);	

	glPushMatrix();
	glLoadIdentity();

	glTranslatef(0, 0, QM_DEFAULT_DEPTH);

	glScalef(scaleFactor, scaleFactor, scaleFactor);

	glRotatef(stereo_rot, 0.0f, 1.0f, 0.0f);

	for(int i = 0;i < atoms.Length();i++){
		atom &atom_ref = atoms[i];

		// Only render solvent cross sections if requested
		if(!solvent && atom_ref.isSolvent)
			continue;

		GLfloat radius = atom_ref.radius;

		// R = M*r, the position of the rotated atom center in the frame
		// of the clipping plane.
		R[X] = M[X][X]*traj[index][3*i] + M[X][Y]*traj[index][3*i + 1] + M[X][Z]*traj[index][3*i + 2];
		R[Y] = M[Y][X]*traj[index][3*i] + M[Y][Y]*traj[index][3*i + 1] + M[Y][Z]*traj[index][3*i + 2];
		R[Z] = M[Z][X]*traj[index][3*i] + M[Z][Y]*traj[index][3*i + 1] + M[Z][Z]*traj[index][3*i + 2];

		R[X] += c_x/scaleFactor;
		R[Y] += c_y/scaleFactor;
		R[Z] += c_z/scaleFactor;

		// The shortest distance from the atom center to the clip plane
		// along the normal. 
		GLfloat dist = 0.0f;

		// dist = (R - center/scaleFactor)*normal
		dist += (R[X] - center[X]/scaleFactor)*normal[X];
		dist += (R[Y] - center[Y]/scaleFactor)*normal[Y];
		dist += (R[Z] - center[Z]/scaleFactor)*normal[Z];

		// Does the plane intersect the sphere?
		if(fabs(dist) < radius){
			// Draw the intersection of the clipping plane and the sphere
			GLfloat disk_radius = 
				(GLfloat)(sqrt(fabs(radius*radius - dist*dist)));
			
			// The center of the disk
			//disk = R - normal*dist;
			disk[X] = R[X] - normal[X]*dist;
			disk[Y] = R[Y] - normal[Y]*dist;
			disk[Z] = R[Z] - normal[Z]*dist;

			// Render the disk as a triangle fan
			glBegin(GL_TRIANGLE_FAN);
				glNormal3fv(normal);
				glColor4ubv(atom_ref.rgb_color);
				
				glVertex3fv(disk);

				for(int j = 0;j <= resolution;j++){
					GLfloat theta = (2.0f*PI*j)/resolution;
					GLfloat x = (GLfloat)(cos(theta))*disk_radius;
					GLfloat y = (GLfloat)(sin(theta))*disk_radius;

					// Use R as scratch space
					//R = A*x + B*y + disk;
					R[X] = A[X]*x + B[X]*y + disk[X];
					R[Y] = A[Y]*x + B[Y]*y + disk[Y];
					R[Z] = A[Z]*x + B[Z]*y + disk[Z];

					glVertex3fv(R);
				}

			glEnd();
		}
	}

	glPopMatrix();

	glEnable(clip_id);
	glEnable(GL_CULL_FACE);
	glDepthFunc(GL_LESS);
}

// Does the given sphere intersect the clipping plane? Sort the given
// sphere into one of three catagories: NOT_RENDERED, RENDERED or CLIPED.
int ClipPlane::SphereClipTest(GLfloat coord[], GLfloat radius,
							   GLfloat scaleFactor, GLfloat c_x, GLfloat c_y, GLfloat c_z,
							   GLfloat M[4][4])
{
	GLfloat R[3];

	// Divide once, multiple many times
	scaleFactor = 1.0f/scaleFactor;

	// R = M*r, the position of the rotated atom center in the frame
	// of the clipping plane.
	R[X] = M[X][X]*coord[X] + M[X][Y]*coord[Y] + M[X][Z]*coord[Z];
	R[Y] = M[Y][X]*coord[X] + M[Y][Y]*coord[Y] + M[Y][Z]*coord[Z];
	R[Z] = M[Z][X]*coord[X] + M[Z][Y]*coord[Y] + M[Z][Z]*coord[Z];

	R[X] += c_x*scaleFactor;
	R[Y] += c_y*scaleFactor;
	R[Z] += c_z*scaleFactor;

	// The shortest distance from the atom center to the clip plane
	// along the normal. 
	GLfloat dist;

	// dist = (R - center*scaleFactor)*normal
	dist = (R[X] - center[X]*scaleFactor)*normal[X];
	dist += (R[Y] - center[Y]*scaleFactor)*normal[Y];
	dist += (R[Z] - center[Z]*scaleFactor)*normal[Z];

	// Does the plane intersect the sphere?
	if(fabs(dist) < radius)
		return CLIPED;

	if(dist < radius)
		return RENDERED;

	return NOT_RENDERED;
}