//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.
//
//	Look-and-feel inspired by Xmol 
//	(http://www.networkcs.com/msc/docs/xmol/).
//
//	OpenGL selection code and quaterion rotation matrix
//	formulation from molview demo program by Mark Kilgard
//	(http://reality.sgi.com/opengl/OpenGLforX.html). PostScript
//	rendering also from Mark Kilgard.
//
//	The printing routines are based on code from
//	Craig Fahrnbach <craignan@home.com> via
//	Uwe Kotyczka <kotyczka@bnm-gmbh.de>.
//
//	Covalent bond determination code based on VMD
//	http://www.ks.uiuc.edu/Research/vmd/
//
//	Stride (http://www.embl-heidelberg.de/stride/stride.html)
//	is used for secondary structure determination and has been 
//	included in Qmol with the kind permission of 
//	Dmitrij Frishman, PhD
//	Institute for Bioinformatics
//	GSF - Forschungszentrum fr Umwelt und Gesundheit, GmbH
//	Ingolstdter Landstrae 1,
//	D-85764 Neuherberg, Germany
//	Telephone: +49-89-3187-4201
//	Fax: +49-89-31873585
//	e-mail: d.frishman@gsf.de
//	WWW: http://mips.gsf.de/mips/staff/frishman/
#include "stdafx.h"

/*
 * A stripped down implementation of the DCD file format.
 */

/* See DCD.h for comments */

#include "DCD.h"

/*
 * Indexing for 3D data.
 */
enum {X = 0, Y, Z};

/*
 * Local function definitions
 */

size_t FRead(void *ptr, size_t size, size_t nobj, FILE *stream, int swap);

/*
 * Convert from big endian to little endian (and visa-versa)
 * when reading binary data. If swap == TRUE, then the data is byte
 * reversed when it is read. If swap == FALSE, then the data is
 * read without modification. Note that this function has NOT been
 * tested for reading structures in binary! It HAS been tested for reading
 * arrays of char, int, float, double and long double.
 */
size_t FRead(void *ptr, size_t size, size_t nobj, FILE *stream, int swap)
{   
    if(!swap || (size == sizeof(char)))
		return fread(ptr, size, nobj, stream);
    
	unsigned char *temp = (unsigned char*)ptr;

    size_t out = fread(temp, 1, size*nobj, stream);

    for(unsigned int i = 0;i < nobj;i++)
		for(unsigned int j = 0;j < size/2;j++){
	    	unsigned char c = temp[i*size + size - j - 1];
	    	temp[i*size + size - j - 1] = temp[i*size + j];
	    	temp[i*size + j] = c;
		}

    return out;
}

void dcd::read_dcd_header(int &npt, int &steps, int &size)
{
	if(!fin)
		throw QMOL_ERROR("Please open the file before reading!");
		
    char char_buffer[80]; /* Needs to be 80 characters long */
    int int_buffer[25];   /* Needs to be at least 23 "intergers" long */
    long file_position;
    
    /* Remove leading integer */
    fread(int_buffer, sizeof(int), 1, fin);
    
	if(int_buffer[0] != MAGIC_NUMBER1){
		rewind(fin);

		FRead(int_buffer, sizeof(int), 1, fin, TRUE);

		if(int_buffer[0] != MAGIC_NUMBER1)
			throw QMOL_ERROR("Corrupted DCD file header");
		
		swap_endian = TRUE;
	}

    /* Read 'CORD' */
    fread(char_buffer, sizeof(char), 4, fin);

	// Read the number of steps */
	FRead(&steps, sizeof(int), 1, fin, swap_endian);

    /* Read 22 integers*/
    FRead(int_buffer, sizeof(int), 22, fin, swap_endian);
	
	if((int_buffer[20] - 4)%80 != 0)
		throw QMOL_ERROR("Bad DCD file header");

	// int_buffer[21] stores the number of 80 byte title strings
	for(int i = 0;i < int_buffer[21];i++)
		// Read header title(s). 
		fread(char_buffer, sizeof(char), 80, fin);

    /*Read 5 integers including npt */
	FRead(int_buffer, sizeof(int), 5, fin, swap_endian);

    npt =  int_buffer[2];

    /*
     * Now calculate the size (in bytes) of each array element. This 
     * "brute force" calculation will hopefully be very robust.
     */
    
    /* Current position */
    file_position = ftell(fin);
    
    /* Move to the end of the file */
    fseek(fin, 0, SEEK_END);

	if((npt == 0) || (steps == 0)){
		// Don't divide by zero in the case of 
		// npt == 0 or steps == 0.
		if(npt == 0){
			size = 0;
		}
		else{
			// Assume size == sizeof(float) and compute the
			// number of steps (frames) in the file. Note that 
			// this is the same equation that appears below
			// in the formula for size. This is required to read
			// DCD files written by VMD (which, as far as I can tell)
			// do not specify the number of steps in the header.
			size = sizeof(float);
			steps = (ftell(fin) - file_position + 2*sizeof(int))/
					(3*npt*size + 6*sizeof(int));
		}
	}
	else{
		size = (ftell(fin) - file_position - 
			(6*steps - 2)*sizeof(int))/(3*npt*steps);
	}

    /* reset the file pointer */
    fseek(fin, file_position, SEEK_SET);
    
    if(ferror(fin)){
		error = "Error reading DCD file";
		return;
	}
	
    if(feof(fin)){
		error = "Unexpected end of file!";
		return;
	}
}

void dcd::read_dcd(void *dcd, int npt, int size)
{
	if(!fin)
		throw QMOL_ERROR("Please open the file before reading!");
	
    int i, dim;
    
    for(dim = X;dim <= Z;dim++){
        for(i = 0;i < npt;i++)
	    	/* Read one value at a time */
	    	FRead((char*)dcd  + size*(3*i + dim), size, 1 ,fin, swap_endian);
    
		fseek(fin, 2*sizeof(int), SEEK_CUR);
	
		if(ferror(fin)){
	    	error = "Error reading DCD file!";
			return;
    	}
	
		if(feof(fin)){
	    	error = "Unexpected end of file!";
			return;
		}
    }
}

/*
 * The write_dcd_header() function not only writes the necessary 
 * information at the start of the DCD file, but also inserts the
 * required FORTRAN array terminators throughout the file. This has
 * the effect of pre-allocating the DCD file (i.e. the file will 
 * occupy its final size in memory after this function is called).
 */
void dcd::write_dcd_header(int npt, int steps, int size)
{
	if(!fout)
		throw QMOL_ERROR("Please open the file before writing!");
	
    int i, d;
    char char_buffer[50];
    int int_buffer[25]; /* Needs to be at least 23 "intergers" long */

    int MagicNumber1 = MAGIC_NUMBER1;
    int MagicNumber3 = MAGIC_NUMBER3;
    int MagicNumber4 = MAGIC_NUMBER4;
    int MagicNumber5 = size*npt;


    /* Initialize all arrays to nice default values */
    for(i = 0;i < 50;i++)
		char_buffer[i] = ' ';

    int_buffer[0] = steps;
    
    for(i = 1;i < 20;i++)
	int_buffer[i] = 0;
	
    int_buffer[20] = MagicNumber1;
    int_buffer[21] = MagicNumber1;
    int_buffer[22] = MagicNumber3;

    /* Write leading integer  -> 4 bytes running total */
    fwrite(&MagicNumber1, sizeof(int), 1, fout);

    /* Write 'CORD'  -> 8 bytes running total */
    fwrite("CORD", sizeof(char), 4, fout);

    /* Write 23 integers -> 100 bytes running total */
    fwrite(int_buffer, sizeof(int), 23, fout);

    /* 
     * Write 'A dummy title to dynamics coordinates'. 
     * Note that this string must be padded with blanks 
     * to make it 80 characters long -> 180 bytes running total.
     */
    fwrite("A dummy title to dynamics coordinates", sizeof(char), 37, 
		fout);
    fwrite(char_buffer, sizeof(char), 43, fout);

    /*Write 5 integers including npt -> 200 bytes running total*/
    int_buffer[0] = MagicNumber1;
    int_buffer[1] = MagicNumber4;
    int_buffer[2] = npt;
    int_buffer[3] = MagicNumber4;
    int_buffer[4] = MagicNumber5;
    fwrite(int_buffer, sizeof(int), 5, fout);

    /*
     * A total of 200 bytes of header information have been written.
     * Now the FORTRAN array terminators must be written.
     */
    int_buffer[0] = int_buffer[1] = MagicNumber5;
    
    for(i = 0;i < steps;i++)
		for(d = X;d <= Z;d++){
	    	fseek(fout, npt*size, SEEK_CUR);
	    	fwrite(int_buffer, sizeof(int), 2, fout);
		}
    
    /*
     * Return the file pointer to the end of the header section.
     */
    fseek(fout, DCD_HEADER_LENGTH, SEEK_SET);
}

void dcd::write_dcd_step(void *data, int npt, int size)
{
	if(!fout)
		throw QMOL_ERROR("Please open the file before writing!");
	
    int i, dim;
    int int_buffer[2];
    
    for(i = 0;i < 2;i++)
		int_buffer[i] = size*npt;

    for(dim = X;dim <= Z;dim++){
		for(i = 0;i < npt;i++)
	    	fwrite((char*)data + size*(3*i + dim), size, 1, fout);

		fwrite(int_buffer, sizeof(int), 2, fout);
    }
}

