//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.
//
//	Look-and-feel inspired by Xmol 
//	(http://www.networkcs.com/msc/docs/xmol/).
//
//	OpenGL selection code and quaterion rotation matrix
//	formulation from molview demo program by Mark Kilgard
//	(http://reality.sgi.com/opengl/OpenGLforX.html). PostScript
//	rendering also from Mark Kilgard.
//
//	The printing routines are based on code from
//	Craig Fahrnbach <craignan@home.com> via
//	Uwe Kotyczka <kotyczka@bnm-gmbh.de>.
//
//	Covalent bond determination code based on VMD
//	http://www.ks.uiuc.edu/Research/vmd/
//
//	Stride (http://www.embl-heidelberg.de/stride/stride.html)
//	is used for secondary structure determination and has been 
//	included in Qmol with the kind permission of 
//	Dmitrij Frishman, PhD
//	Institute for Bioinformatics
//	GSF - Forschungszentrum fr Umwelt und Gesundheit, GmbH
//	Ingolstdter Landstrae 1,
//	D-85764 Neuherberg, Germany
//	Telephone: +49-89-3187-4201
//	Fax: +49-89-31873585
//	e-mail: d.frishman@gsf.de
//	WWW: http://mips.gsf.de/mips/staff/frishman/
// 
#include "stdafx.h"
#include "Qmol.h"
#include "QmolDoc.h"
#include "LinearAlgebra3D.h"
#include <math.h>

// Energetic cut off for the Kabsch & Sander H-bond definition
#define		HBOND_CUT_OFF	(-0.5f) // kcal/mol

#define		Q1				(0.42f)
#define		Q2				(0.20f)
#define		F_HB			(332.0f)
#define		DIST_CUT_OFF	(2.1f*2.1f)

#define		HBOND_DONOR		0
#define		HBOND_ACCEPTOR	1

class HBondAtom{
public:
	molecule *mol;
	char hbondType;
	int atomIndex;
};

// Two molecules
BOOL IsHBond(HBondAtom &dnr, HBondAtom &acc);

// One molecule
BOOL IsHBond(int dnr, int acc, molecule *mol);

// Two molecules
float HBondEnergy(int dnr_1, int dnr_2, molecule *dnr_mol, 
				  int acc_1, int acc_2, molecule *acc_mol);

// One molecule
float HBondEnergy(int dnr_1, int dnr_2, int acc_1, int acc_2, molecule *mol);

// Two molecules
float HBondEnergy(int dnr_1, array<float> &r_h, molecule *dnr_mol,
				  int acc_1, int acc_2, molecule *acc_mol);

// One molecule
float HBondEnergy(int dnr_1, array<float> &r_h, 
				  int acc_1, int acc_2, molecule *mol);

// Two molecules
BOOL PlaceHydrogen(array<float> &r_h, HBondAtom &dnr, HBondAtom &acc);

// One molecule
BOOL PlaceHydrogen(array<float> &r_h, int dnr, int acc, molecule *mol);

void CQmolDoc::UpdateHydrogenBonding()
{
	// Update the hbond array by removing any previously computed
	// hydrogen bonds
	while(!hbonds.IsEmpty())
		hbonds.Pop();

	BOOL use_mol_frame = FALSE;

	// Are there any hydrogen bonds to compute?
	switch(molList.Length()){
		case 0:
			// No hbonds to compute
			return;
		case 1:
			// Don't compute atom positions in the world frame
			use_mol_frame = TRUE;
			break;
	};

	int num_hbonds = 0;
	
	float r[3], tmp_r[3];
  	int i, j, k;
  	float boxsizex, boxsizey, boxsizez;
  	int xb, yb, zb, xytotb, totb, axb, ayb, azb, aindex, bindex;
	float xmax, xmin, ymax, ymin, zmax, zmin;
	int xi, yi, zi;
	float tmp_xsize, tmp_ysize, tmp_zsize;

	// Find min/max bounds of all molecular coordinates.
	ListIterator<molecule*> next(molList);
	molecule **tmpMol;

	xmax = xmin = ymax = ymin = zmax = zmin = 0.0f;

	while(tmpMol = next()){
		float tmp_xmax, tmp_xmin, tmp_ymax, tmp_ymin, tmp_zmax, tmp_zmin;

		(*tmpMol)->GetCurrentMolDim(tmp_xmax, tmp_xmin, 
									tmp_ymax, tmp_ymin, 
									tmp_zmax, tmp_zmin,
									use_mol_frame);

		xmax = MAX(xmax, tmp_xmax);
		xmin = MIN(xmin, tmp_xmin);

		ymax = MAX(ymax, tmp_ymax);
		ymin = MIN(ymin, tmp_ymin);

		zmax = MAX(zmax, tmp_zmax);
		zmin = MIN(zmin, tmp_zmin);
	}

	// From size of molecule, break up space into boxes of side length=4 A
  	// (or, if the molecule size is too large in any direction, the boxes are
  	// increased so that a maximum of 50 boxes are used in any direction)
	tmp_xsize = xmax - xmin;
	tmp_ysize = ymax - ymin;
	tmp_zsize = zmax - zmin;

  	boxsizex = tmp_xsize >= 200.0f ? tmp_xsize/50.0f : 4.0f;
  	boxsizey = tmp_ysize >= 200.0f ? tmp_ysize/50.0f : 4.0f;
  	boxsizez = tmp_zsize >= 200.0f ? tmp_zsize/50.0f : 4.0f;

  	xb = (int)(tmp_xsize / boxsizex) + 1;
  	yb = (int)(tmp_ysize / boxsizey) + 1;
  	zb = (int)(tmp_zsize / boxsizez) + 1;

  	xytotb = yb*xb;
  	totb = xytotb*zb;

	list<HBondAtom> **boxatom = new list<HBondAtom>* [totb];
	list<HBondAtom> *tmpbox, *nbrbox;

	for(i = 0; i < totb;i++)
		boxatom[i] = NULL;

	// Re-initialize the list iterator
	next.Init(molList);

	while(tmpMol = next()){
		
		float m[4][4];

		if(!use_mol_frame)
			(*tmpMol)->BuildRotationMatrix(m);

		// For all the atoms, put them in their particular box
  		for(i = 0; i < (*tmpMol)->NumAtoms(); i++) {

			// Is this atom a possible hydrogen bond donor or acceptor?
			BOOL isDonor = (*tmpMol)->IsHBondDonor(i);
			BOOL isAcceptor = FALSE;

			if(!isDonor){
				isAcceptor = (*tmpMol)->IsHBondAcceptor(i);

				if(!isAcceptor)
					continue;
			}
			
			if((*tmpMol)->current_step == 0){
				(*tmpMol)->atoms[i].GetCoord(tmp_r);
			}
			else{
				memcpy(tmp_r, &((*tmpMol)->traj[(*tmpMol)->current_step - 1][3*i]), 3*sizeof(float));
			}
			
			if(!use_mol_frame){

				TRANSLATE(tmp_r, (*tmpMol)->R);

				r[X] = m[X][X]*tmp_r[X] + m[Y][X]*tmp_r[Y] + m[Z][X]*tmp_r[Z];
				r[Y] = m[X][Y]*tmp_r[X] + m[Y][Y]*tmp_r[Y] + m[Z][Y]*tmp_r[Z];
				r[Z] = m[X][Z]*tmp_r[X] + m[Y][Z]*tmp_r[Y] + m[Z][Z]*tmp_r[Z];
			}
			else{
				r[X] = tmp_r[X];
				r[Y] = tmp_r[Y];
				r[Z] = tmp_r[Z];
			}

    		axb = (int)((r[X] - xmin) / boxsizex);
    		ayb = (int)((r[Y] - ymin) / boxsizey);
    		azb = (int)((r[Z] - zmin) / boxsizez);
			
    		aindex = azb*xytotb + ayb*xb + axb;

			if(boxatom[aindex] == NULL)
				// Only allocate memory for a list if there is at least one
				// atom in the box
				boxatom[aindex] = new list<HBondAtom>;

			// Add the current record to box aindex.
			HBondAtom tmpAtom;
			
			tmpAtom.mol = (*tmpMol);
			tmpAtom.atomIndex = i;

			if(isDonor)
				tmpAtom.hbondType = HBOND_DONOR;
			else // isAcceptor
				tmpAtom.hbondType = HBOND_ACCEPTOR;

			boxatom[aindex]->Push(tmpAtom);
		}
	}

	// Now, for each box, check atoms in (positive) neighbor boxes for bonds
  	aindex = 0;
  	for(zi = 0;zi < zb; zi++){
		for(yi = 0;yi < yb; yi++){
      		for(xi = 0;xi < xb; xi++){
        		tmpbox = boxatom[aindex];
				
				// Only check boxes that contain at least one atom
				if(tmpbox != NULL){
					for(j = 0;j < 14; j++){
	  					switch(j){
	    					case 0: 
								bindex = aindex; 
								break; // same
	    					case 1: 
								if(xi < (xb - 1)) 
									bindex = aindex + 1;
								else
									continue;
								break; // X
	    					case 2: 
								if(yi < (yb - 1)) 
									bindex = aindex + xb;
								else
									continue;
								break; // Y
            				case 3:
								if(zi < (zb - 1))
									bindex = aindex + xytotb;
								else
									continue;
								break; // Z		
	    					case 4:
								if(xi < (xb - 1) && yi < (yb - 1))
	              					bindex = aindex + xb + 1;
								else
									continue;
		    					break; // XY
	    					case 5: 
								if(xi < (xb - 1) && zi < (zb - 1))
	              					bindex = aindex + xytotb + 1;
								else
									continue;
		    					break; // XZ
	    					case 6: 
								if(yi < (yb - 1) && zi < (zb - 1))
	              					bindex = aindex + xytotb + xb;
								else
									continue;
		    					break; // YZ
	    					case 7:
								if(xi < (xb - 1) && yi > 0)
	              					bindex = aindex - xb + 1;
								else
									continue;
		    					break; // X-Y
	    					case 8:
								if(xi > 0 && zi < (zb - 1))
	              					bindex = aindex + xytotb - 1;
								else
									continue;
		    					break; // -XZ
	    					case 9: 
								if(yi > 0 && zi < (zb - 1))
	              					bindex = aindex + xytotb - xb;
								else
									continue;
		    					break; // -YZ
	    					case 10: 
								if(xi < (xb - 1) && yi < (yb - 1) && zi < (zb - 1))
	              					bindex = aindex + xytotb + xb + 1;
								else
									continue;
		    					break; // XYZ
	    					case 11:
								if(xi > 0 && yi < (yb - 1) && zi < (zb - 1))
	              					bindex = aindex + xytotb + xb - 1;
								else
									continue;
		    					break; // -XYZ
	    					case 12:
								if(xi < (xb - 1) && yi > 0 && zi < (zb - 1))
	              					bindex = aindex + xytotb - xb + 1;
								else
									continue;
		    					break; // X-YZ
	    					case 13: 
								if(xi > 0 && yi > 0 && zi < (zb - 1))
	              					bindex = aindex + xytotb - xb - 1;
								else
									continue;
		    					break; //-X-YZ
	  					};
	  					
						if((nbrbox = boxatom[bindex]) == NULL)
							// Skip to the next box if the current box is
							// empty
							continue;
							
						// The number of atoms in box a and box b.
						int num_inbox_a = tmpbox->Length();
						int num_inbox_b = nbrbox->Length();
							
						// The current atom in box a and box b.
						HBondAtom curr_a, curr_b;

            			for(i = 0;i < num_inbox_a; i++){
							curr_a = (*tmpbox)[i];
							
							// Is the current atom a donor or acceptor?
							if(curr_a.hbondType == HBOND_DONOR){
								// This is a donor atom!

								for(k = 0;k < num_inbox_b; k++){
									curr_b = (*nbrbox)[k];
									
									if(curr_b.hbondType == HBOND_DONOR)
										// This is a donor atom -- skip it
										continue;

									// Test for hydrogen bonding here
									if(IsHBond(curr_a, curr_b)){
										QmolHBond tmp_bond;
										
										tmp_bond.indexA = curr_a.atomIndex;
										tmp_bond.molA = curr_a.mol;

										tmp_bond.indexB = curr_b.atomIndex;
										tmp_bond.molB = curr_b.mol;

										hbonds.Push(tmp_bond);
									}
	      						}
							}
							else{
								// Don't double count the bonds
								if(bindex == aindex)
									continue;

								// This is an acceptor atom

								for(k = 0;k < num_inbox_b; k++){
									curr_b = (*nbrbox)[k];
									
									if(curr_b.hbondType == HBOND_ACCEPTOR)
										// This is an acceptor atom -- skip it
										continue;

									// Test for hydrogen bonding here
									if(IsHBond(curr_b, curr_a)){
										QmolHBond tmp_bond;
										
										tmp_bond.indexA = curr_a.atomIndex;
										tmp_bond.molA = curr_a.mol;

										tmp_bond.indexB = curr_b.atomIndex;
										tmp_bond.molB = curr_b.mol;

										hbonds.Push(tmp_bond);
									}
	      						}	
							}
	  					}
					}
				}
				aindex++;
			}
		}
  	}
  
	// free up the storage space allocted for the grid search
	for(i = 0;i < totb; i++)
    	if(boxatom[i])  
			delete boxatom[i];
  
	delete [] boxatom;
}

BOOL molecule::IsHBondDonor(int i)
{
	// Hydrogen atoms that are bonded to oxygen or nitrogen 
	// can be donors
	if(atoms[i].elem == ELEM_H){
		if(mol_connect[i].Length() != 1)
			return FALSE;
		
		int tmp_elem = atoms[mol_connect[i][0]].elem;

		if((tmp_elem == ELEM_O) || 
		   (tmp_elem == ELEM_N))
			return TRUE;

		return FALSE;
	}

	// For structures where the hydrogen atoms have not been provided,
	// the structure of the peptide plane is used to add the hydrogen
	// that is attached to the nitrogen.
	if(atoms[i].elem == ELEM_N)
		return TRUE;

	return FALSE;
}

BOOL molecule::IsHBondAcceptor(int i)
{
	// Double bonded Oxygen atoms (O == C) may be hydrogen bond acceptors
	if(atoms[i].elem == ELEM_O && (mol_connect[i].Length() == 1))
		return TRUE;

	return FALSE;
}

BOOL IsHBond(HBondAtom &dnr, HBondAtom &acc)
{
	if(dnr.mol == acc.mol)
		return IsHBond(dnr.atomIndex, acc.atomIndex, dnr.mol);

	// Only singly bonded oxygen atoms are valid acceptors.
	// Find the atom number if the carbon atom attached to the
	// oxygen atom
	int acc_c;
	array<unsigned int> &accConnectRef = acc.mol->mol_connect[acc.atomIndex];
	array<unsigned int> &dnrConnectRef = dnr.mol->mol_connect[dnr.atomIndex];

	if(accConnectRef.Length() != 1)
		return FALSE;

	acc_c = accConnectRef[0];

	// The atom number of the nitrogen or oxygen attached to the
	// hydrogen
	int dnr_n_or_o;

	if(dnr.mol->atoms[dnr.atomIndex].elem == ELEM_H){
			
		dnr_n_or_o = dnrConnectRef[0];

		// Compute the hbond energy
		if(HBondEnergy(dnr_n_or_o, dnr.atomIndex, dnr.mol, 
					   acc.atomIndex, acc_c, acc.mol) < HBOND_CUT_OFF)
			return TRUE;

		return FALSE;
	}

	// If the donor is a nitrogen, we need to
	// (a) Make sure a hydrogen is not already present and
	// (b) compute the location of the "virtual" hydrogen from the
	// structure if the peptide plane.
	if(dnr.mol->atoms[dnr.atomIndex].elem == ELEM_N){
		// Look for attached hydrogen
		int len = dnrConnectRef.Length();

		for(int i = 0;i < len;i++)
			if(dnr.mol->atoms[dnrConnectRef[i]].elem == ELEM_H)
				return FALSE;
		
		array<float> r_h(3);

		// Given the atom number of the non-hydrogen donor atom, place the
		// "virtual" hydrogen in the correct location.
		if(!PlaceHydrogen(r_h, dnr, acc))
			// Unable to place the hydrogen -- bad or unknown geometery
			return FALSE;

		// Compute the hbond energy
		if(HBondEnergy(dnr.atomIndex, r_h, dnr.mol,
					   acc.atomIndex, acc_c, acc.mol) < HBOND_CUT_OFF)
			return TRUE;

		return FALSE;
	}

	return FALSE;
}

BOOL IsHBond(int dnr, int acc, molecule *mol)
{
	// Hydrogen bonds are not allowed between different models
	if(mol->atoms[dnr].model != mol->atoms[acc].model)
		return FALSE;

	// Only singly bonded oxygen atoms are valid acceptors.
	// Find the atom number if the carbon atom attached to the
	// oxygen atom
	int acc_c;
	
	if(mol->mol_connect[acc].Length() != 1)
		return FALSE;

	acc_c = mol->mol_connect[acc][0];

	// The atom number of the nitrogen or oxygen attached to the
	// hydrogen
	int dnr_n_or_o;

	if(mol->atoms[dnr].elem == ELEM_H){
			
		dnr_n_or_o = mol->mol_connect[dnr][0];

		// Compute the hbond energy
		if(HBondEnergy(dnr_n_or_o, dnr, acc, acc_c, mol) < HBOND_CUT_OFF)
			return TRUE;

		return FALSE;
	}

	// If the donor is a nitrogen, we need to
	// (a) Make sure a hydrogen is not already present and
	// (b) compute the location of the "virtual" hydrogen from the
	// structure if the peptide plane.
	if(mol->atoms[dnr].elem == ELEM_N){
		// Look for attached hydrogen
		int len = mol->mol_connect[dnr].Length();

		for(int i = 0;i < len;i++)
			if(mol->atoms[mol->mol_connect[dnr][i]].elem == ELEM_H)
				return FALSE;
		
		array<float> r_h(3);

		// Given the atom number of the non-hydrogen donor atom, place the
		// "virtual" hydrogen in the correct location.
		if(!PlaceHydrogen(r_h, dnr, acc, mol))
			// Unable to place the hydrogen -- bad or unknown geometery
			return FALSE;

		if(HBondEnergy(dnr, r_h, acc, acc_c, mol) < HBOND_CUT_OFF)
			return TRUE;

		return FALSE;
	}

	return FALSE;
}

float HBondEnergy(int dnr_1, int dnr_2, molecule *dnr_mol, 
				  int acc_1, int acc_2, molecule *acc_mol)
{
	// Use the H-bond energy function and cutoff from W. Kabsch and C. Sander
	// "Dictionary of Protien Secondary Structure: Pattern Recognition of Hydrogen-
	// Bonded and Geometrical Features", (1983) Biopolymers 22, 2577-2637
	//
	// dnr_1 -- dnr_2 .... acc_1 == acc_2
	//   N   --   H   ....   O   ==  C
	//
	//	An additional pairwise distance constraint between dnr_2 and acc_1 is imposed:
	//	r_{dnr-acc} <= 2.1 angstroms
	//	The cut off value of 2.1 is an upper bound to the length of a hydrogen bond
	//	(see "Proteins: Structures and molecular properties" 2nd Ed. by T. E. Creighton).

	if(dnr_mol == acc_mol)
		// No correction for different molecular frames is required!
		return HBondEnergy(dnr_1, dnr_2, acc_1, acc_2, dnr_mol);

	float r_ON, r_CH, r_OH, r_CN;

	float dnr_m[4][4];
	float acc_m[4][4];
	float delta[3];
	float dnr_1_v[3], dnr_2_v[3];
	float acc_1_v[3], acc_2_v[3];

	dnr_mol->BuildRotationMatrix(dnr_m);
	acc_mol->BuildRotationMatrix(acc_m);

	// Use delta as a temporary vector for performing translations and
	// rotations to the world frame.
	if(dnr_mol->current_step == 0){
		memcpy(delta, dnr_mol->atoms[dnr_1].coord, 3*sizeof(float));
		TRANSLATE(delta, dnr_mol->R);
		ROTATE(dnr_1_v, dnr_m, delta);

		memcpy(delta, dnr_mol->atoms[dnr_2].coord, 3*sizeof(float));
		TRANSLATE(delta, dnr_mol->R);
		ROTATE(dnr_2_v, dnr_m, delta);
	}
	else{
		int dnr_step = dnr_mol->current_step - 1;
		
		memcpy(delta, &(dnr_mol->traj[dnr_step][3*dnr_1]), 3*sizeof(float));
		TRANSLATE(delta, dnr_mol->R);
		ROTATE(dnr_1_v, dnr_m, delta);

		memcpy(delta, &(dnr_mol->traj[dnr_step][3*dnr_2]), 3*sizeof(float));
		TRANSLATE(delta, dnr_mol->R);
		ROTATE(dnr_2_v, dnr_m, delta);
	}

	if(acc_mol->current_step == 0){
		memcpy(delta, acc_mol->atoms[acc_1].coord, 3*sizeof(float));
		TRANSLATE(delta, acc_mol->R);
		ROTATE(acc_1_v, acc_m, delta);

		memcpy(delta, acc_mol->atoms[acc_2].coord, 3*sizeof(float));
		TRANSLATE(delta, acc_mol->R);
		ROTATE(acc_2_v, acc_m, delta);
	}
	else{
		int acc_step = acc_mol->current_step - 1;

		memcpy(delta, &(acc_mol->traj[acc_step][3*acc_1]), 3*sizeof(float));
		TRANSLATE(delta, acc_mol->R);
		ROTATE(acc_1_v, acc_m, delta);

		memcpy(delta, &(acc_mol->traj[acc_step][3*acc_2]), 3*sizeof(float));
		TRANSLATE(delta, acc_mol->R);
		ROTATE(acc_2_v, acc_m, delta);
	}

	// r_OH == dnr_2 -> acc_1
	DIFF(delta, dnr_2_v, acc_1_v);
	
	r_OH = DOT(delta, delta);
	
	if(r_OH > DIST_CUT_OFF)
		return HBOND_CUT_OFF + 1.0f;
	
	r_OH = (float)(sqrt(r_OH));
	
	// r_ON == dnr_1 -> acc_1
	DIFF(delta, dnr_1_v, acc_1_v);
	
	r_ON = (float)(sqrt(DOT(delta, delta)));
	
	// r_CH == dnr_2 -> acc_2
	DIFF(delta, dnr_2_v, acc_2_v);
	
	r_CH = (float)(sqrt(DOT(delta, delta)));
	
	// r_CN == dnr_1 -> acc_2
	DIFF(delta, dnr_1_v, acc_2_v);
	
	r_CN = (float)(sqrt(DOT(delta, delta)));

	return Q1*Q2*F_HB*(1.0f/r_ON + 1.0f/r_CH - 1.0f/r_OH - 1.0f/r_CN);
}

float HBondEnergy(int dnr_1, array<float> &r_h, molecule *dnr_mol,
				  int acc_1, int acc_2, molecule *acc_mol)
{
	// Use the H-bond energy function and cutoff from W. Kabsch and C. Sander
	// "Dictionary of Protien Secondary Structure: Pattern Recognition of Hydrogen-
	// Bonded and Geometrical Features", (1983) Biopolymers 22, 2577-2637
	//
	// dnr_1 -- dnr_2 .... acc_1 == acc_2
	//   N   --   H   ....   O   ==  C
	//
	//	An additional pairwise distance constraint between dnr_2 and acc_1 is imposed:
	//	r_{dnr-acc} <= 2.1 angstroms
	//	The cut off value of 2.1 is an upper bound to the length of a hydrogen bond
	//	(see "Proteins: Structures and molecular properties" 2nd Ed. by T. E. Creighton).

	if(dnr_mol == acc_mol)
		// No correction for different molecular frames is required!
		return HBondEnergy(dnr_1, r_h, acc_1, acc_2, dnr_mol);

	float r_ON, r_CH, r_OH, r_CN;

	float dnr_m[4][4];
	float acc_m[4][4];
	float delta[3];
	float dnr_1_v[3];
	float acc_1_v[3], acc_2_v[3];

	dnr_mol->BuildRotationMatrix(dnr_m);
	acc_mol->BuildRotationMatrix(acc_m);

	if(dnr_mol->current_step == 0){
		memcpy(delta, dnr_mol->atoms[dnr_1].coord, 3*sizeof(float));
		TRANSLATE(delta, dnr_mol->R);
		ROTATE(dnr_1_v, dnr_m, delta);
	}
	else{
		int dnr_step = dnr_mol->current_step - 1;
		
		memcpy(delta, &(dnr_mol->traj[dnr_step][3*dnr_1]), 3*sizeof(float));
		TRANSLATE(delta, dnr_mol->R);
		ROTATE(dnr_1_v, dnr_m, delta);
	}

	if(acc_mol->current_step == 0){
		memcpy(delta, acc_mol->atoms[acc_1].coord, 3*sizeof(float));
		TRANSLATE(delta, acc_mol->R);
		ROTATE(acc_1_v, acc_m, delta);

		memcpy(delta, acc_mol->atoms[acc_2].coord, 3*sizeof(float));
		TRANSLATE(delta, acc_mol->R);
		ROTATE(acc_2_v, acc_m, delta);
	}
	else{
		int acc_step = acc_mol->current_step - 1;

		memcpy(delta, &(acc_mol->traj[acc_step][3*acc_1]), 3*sizeof(float));
		TRANSLATE(delta, acc_mol->R);
		ROTATE(acc_1_v, acc_m, delta);

		memcpy(delta, &(acc_mol->traj[acc_step][3*acc_2]), 3*sizeof(float));
		TRANSLATE(delta, acc_mol->R);
		ROTATE(acc_2_v, acc_m, delta);
	}

	// r_OH == dnr_2 -> acc_1
	DIFF(delta, r_h, acc_1_v);
	
	r_OH = DOT(delta, delta);
	
	if(r_OH > DIST_CUT_OFF)
		return HBOND_CUT_OFF + 1.0f;
	
	r_OH = (float)(sqrt(r_OH));
	
	// r_ON == dnr_1 -> acc_1
	DIFF(delta, dnr_1_v, acc_1_v);
	
	r_ON = (float)(sqrt(DOT(delta, delta)));
	
	// r_CH == dnr_2 -> acc_2
	DIFF(delta, r_h, acc_2_v);
	
	r_CH = (float)(sqrt(DOT(delta, delta)));
	
	// r_CN == dnr_1 -> acc_2
	DIFF(delta, dnr_1_v, acc_2_v);
	
	r_CN = (float)(sqrt(DOT(delta, delta)));

	return Q1*Q2*F_HB*(1.0f/r_ON + 1.0f/r_CH - 1.0f/r_OH - 1.0f/r_CN);
}


float HBondEnergy(int dnr_1, int dnr_2, int acc_1, int acc_2, molecule *mol)
{
	// Use the H-bond energy function and cutoff from W. Kabsch and C. Sander
	// "Dictionary of Protien Secondary Structure: Pattern Recognition of Hydrogen-
	// Bonded and Geometrical Features", (1983) Biopolymers 22, 2577-2637
	//
	// dnr_1 -- dnr_2 .... acc_1 == acc_2
	//   N   --   H   ....   O   ==  C
	//
	//	An additional pairwise distance constraint between dnr_2 and acc_1 is imposed:
	//	r_{dnr-acc} <= 2.1 angstroms
	//	The cut off value of 2.1 is an upper bound to the length of a hydrogen bond
	//	(see "Proteins: Structures and molecular properties" 2nd Ed. by T. E. Creighton).

	float x, y, z;
	float r_ON, r_CH, r_OH, r_CN;

	if(mol->current_step == 0){
		// r_OH == dnr_2 -> acc_1
		x = mol->atoms[dnr_2].coord[X] - mol->atoms[acc_1].coord[X];
		y = mol->atoms[dnr_2].coord[Y] - mol->atoms[acc_1].coord[Y];
		z = mol->atoms[dnr_2].coord[Z] - mol->atoms[acc_1].coord[Z];
		
		r_OH = x*x + y*y + z*z;
		
		if(r_OH > DIST_CUT_OFF)
			return HBOND_CUT_OFF + 1.0f;
		
		r_OH = (float)(sqrt(r_OH));
		
		// r_ON == dnr_1 -> acc_1
		x = mol->atoms[dnr_1].coord[X] - mol->atoms[acc_1].coord[X];
		y = mol->atoms[dnr_1].coord[Y] - mol->atoms[acc_1].coord[Y];
		z = mol->atoms[dnr_1].coord[Z] - mol->atoms[acc_1].coord[Z];
		
		r_ON = (float)(sqrt(x*x + y*y + z*z));
		
		// r_CH == dnr_2 -> acc_2
		x = mol->atoms[dnr_2].coord[X] - mol->atoms[acc_2].coord[X];
		y = mol->atoms[dnr_2].coord[Y] - mol->atoms[acc_2].coord[Y];
		z = mol->atoms[dnr_2].coord[Z] - mol->atoms[acc_2].coord[Z];
		
		r_CH = (float)(sqrt(x*x + y*y + z*z));
		
		// r_CN == dnr_1 -> acc_2
		x = mol->atoms[dnr_1].coord[X] - mol->atoms[acc_2].coord[X];
		y = mol->atoms[dnr_1].coord[Y] - mol->atoms[acc_2].coord[Y];
		z = mol->atoms[dnr_1].coord[Z] - mol->atoms[acc_2].coord[Z];
		
		r_CN = (float)(sqrt(x*x + y*y + z*z));
	}
	else{ // mol->current_step != 0
		int mol_step = mol->current_step - 1;

		// r_OH == dnr_2 -> acc_1
		x = mol->traj[mol_step][3*dnr_2] - mol->traj[mol_step][3*acc_1];
		y = mol->traj[mol_step][3*dnr_2 + Y] - mol->traj[mol_step][3*acc_1 + Y];
		z = mol->traj[mol_step][3*dnr_2 + Z] - mol->traj[mol_step][3*acc_1 + Z];
		
		r_OH = x*x + y*y + z*z;
		
		if(r_OH > DIST_CUT_OFF)
			return HBOND_CUT_OFF + 1.0f;
		
		r_OH = (float)(sqrt(r_OH));
		
		// r_ON == dnr_1 -> acc_1
		x = mol->traj[mol_step][3*dnr_1] - mol->traj[mol_step][3*acc_1];
		y = mol->traj[mol_step][3*dnr_1 + Y] - mol->traj[mol_step][3*acc_1 + Y];
		z = mol->traj[mol_step][3*dnr_1 + Z] - mol->traj[mol_step][3*acc_1 + Z];
		
		r_ON = (float)(sqrt(x*x + y*y + z*z));
		
		// r_CH == dnr_2 -> acc_2
		x = mol->traj[mol_step][3*dnr_2] - mol->traj[mol_step][3*acc_2];
		y = mol->traj[mol_step][3*dnr_2 + Y] - mol->traj[mol_step][3*acc_2 + Y];
		z = mol->traj[mol_step][3*dnr_2 + Z] - mol->traj[mol_step][3*acc_2 + Z];
		
		r_CH = (float)(sqrt(x*x + y*y + z*z));
		
		// r_CN == dnr_1 -> acc_2
		x = mol->traj[mol_step][3*dnr_1] - mol->traj[mol_step][3*acc_2];
		y = mol->traj[mol_step][3*dnr_1 + Y] - mol->traj[mol_step][3*acc_2 + Y];
		z = mol->traj[mol_step][3*dnr_1 + Z] - mol->traj[mol_step][3*acc_2 + Z];
		
		r_CN = (float)(sqrt(x*x + y*y + z*z));
	}

	return Q1*Q2*F_HB*(1.0f/r_ON + 1.0f/r_CH - 1.0f/r_OH - 1.0f/r_CN);
}

float HBondEnergy(int dnr_1, array<float> &r_h, 
				  int acc_1, int acc_2, molecule *mol)
{
	// Use the H-bond energy function and cutoff from W. Kabsch and C. Sander
	// "Dictionary of Protien Secondary Structure: Pattern Recognition of Hydrogen-
	// Bonded and Geometrical Features", (1983) Biopolymers 22, 2577-2637
	//
	// dnr_1 -- dnr_2 .... acc_1 == acc_2
	//   N   --   H   ....   O   ==  C
	//
	//	An additional pairwise distance constraint between dnr_2 and acc_1 is imposed:
	//	r_{dnr-acc} <= 2.1 angstroms
	//	The cut off value of 2.1 is an upper bound to the length of a hydrogen bond
	//	(see "Proteins: Structures and molecular properties" 2nd Ed. by T. E. Creighton).

	float x, y, z;
	float r_ON, r_CH, r_OH, r_CN;

	if(mol->current_step == 0){
		// r_OH == dnr_2 -> acc_1
		x = r_h[X] - mol->atoms[acc_1].coord[X];
		y = r_h[Y] - mol->atoms[acc_1].coord[Y];
		z = r_h[Z] - mol->atoms[acc_1].coord[Z];
		
		r_OH = x*x + y*y + z*z;
		
		if(r_OH > DIST_CUT_OFF)
			return HBOND_CUT_OFF + 1.0f;
		
		r_OH = (float)(sqrt(r_OH));
		
		// r_ON == dnr_1 -> acc_1
		x = mol->atoms[dnr_1].coord[X] - mol->atoms[acc_1].coord[X];
		y = mol->atoms[dnr_1].coord[Y] - mol->atoms[acc_1].coord[Y];
		z = mol->atoms[dnr_1].coord[Z] - mol->atoms[acc_1].coord[Z];
		
		r_ON = (float)(sqrt(x*x + y*y + z*z));
		
		// r_CH == dnr_2 -> acc_2
		x = r_h[X] - mol->atoms[acc_2].coord[X];
		y = r_h[Y] - mol->atoms[acc_2].coord[Y];
		z = r_h[Z] - mol->atoms[acc_2].coord[Z];
		
		r_CH = (float)(sqrt(x*x + y*y + z*z));
		
		// r_CN == dnr_1 -> acc_2
		x = mol->atoms[dnr_1].coord[X] - mol->atoms[acc_2].coord[X];
		y = mol->atoms[dnr_1].coord[Y] - mol->atoms[acc_2].coord[Y];
		z = mol->atoms[dnr_1].coord[Z] - mol->atoms[acc_2].coord[Z];
		
		r_CN = (float)(sqrt(x*x + y*y + z*z));
	}
	else{ // mol_mol->current_step != 0
		int mol_step = mol->current_step - 1;
		
		// r_OH == dnr_2 -> acc_1
		x = r_h[X] - mol->traj[mol_step][3*acc_1];
		y = r_h[Y] - mol->traj[mol_step][3*acc_1 + Y];
		z = r_h[Z] - mol->traj[mol_step][3*acc_1 + Z];
		
		r_OH = x*x + y*y + z*z;
		
		if(r_OH > DIST_CUT_OFF)
			return HBOND_CUT_OFF + 1.0f;
		
		r_OH = (float)(sqrt(r_OH));
		
		// r_ON == dnr_1 -> acc_1
		x = mol->traj[mol_step][3*dnr_1] - mol->traj[mol_step][3*acc_1];
		y = mol->traj[mol_step][3*dnr_1 + Y] - mol->traj[mol_step][3*acc_1 + Y];
		z = mol->traj[mol_step][3*dnr_1 + Z] - mol->traj[mol_step][3*acc_1 + Z];
		
		r_ON = (float)(sqrt(x*x + y*y + z*z));
		
		// r_CH == dnr_2 -> acc_2
		x = r_h[X] - mol->traj[mol_step][3*acc_2];
		y = r_h[Y] - mol->traj[mol_step][3*acc_2 + Y];
		z = r_h[Z] - mol->traj[mol_step][3*acc_2 + Z];
		
		r_CH = (float)(sqrt(x*x + y*y + z*z));
		
		// r_CN == dnr_1 -> acc_2
		x = mol->traj[mol_step][3*dnr_1] - mol->traj[mol_step][3*acc_2];
		y = mol->traj[mol_step][3*dnr_1 + Y] - mol->traj[mol_step][3*acc_2 + Y];
		z = mol->traj[mol_step][3*dnr_1 + Z] - mol->traj[mol_step][3*acc_2 + Z];
		
		r_CN = (float)(sqrt(x*x + y*y + z*z));
	}

	return Q1*Q2*F_HB*(1.0f/r_ON + 1.0f/r_CH - 1.0f/r_OH - 1.0f/r_CN);
}

BOOL PlaceHydrogen(array<float> &r_h, HBondAtom &dnr, HBondAtom &acc)
{
	// Are both donor and acceptor atoms from the same molecule?
	// If so, there is no need to work in the world coordinate system
	// (i.e. we can stay in the molecular coordinate system and save a few
	// rotations and translations).
	if(dnr.mol == acc.mol)
		return PlaceHydrogen(r_h, dnr.atomIndex, acc.atomIndex, dnr.mol);

	float dnr_m[4][4];
	float acc_m[4][4];
	float tmp_dnr[3];
	float tmp_acc[3];
	float tmp[3];

	dnr.mol->BuildRotationMatrix(dnr_m);
	acc.mol->BuildRotationMatrix(acc_m);

	if(dnr.mol->atoms[dnr.atomIndex].elem == ELEM_N){
		// What is the connectivity of the donor atom?
		switch(dnr.mol->mol_connect[dnr.atomIndex].Length()){
			case 1:
				{
					// This atom belongs to a side chain. Assume the hydrogen atoms are
					// free to align (subject to a bond angle constraint)
					// with the acceptor atom when computing the hydrogen location.
					//
					//			 A
					//		    .
					//		   .
					//		  .
					//	     H
					//      /
					// C-- D
					float r_DA[3], r_CD[3], r_perp[3];
					float mag;
					//const float h_angle = 1.95f;
					const float h_angle = 1.2f;
					
					int dnr_index = dnr.mol->mol_connect[dnr.atomIndex][0];

					if(dnr.mol->current_step == 0){
						tmp_dnr[X] = dnr.mol->atoms[dnr.atomIndex].coord[X] - 
								dnr.mol->atoms[dnr_index].coord[X];
						tmp_dnr[Y] = dnr.mol->atoms[dnr.atomIndex].coord[Y] - 
								dnr.mol->atoms[dnr_index].coord[Y];
						tmp_dnr[Z] = dnr.mol->atoms[dnr.atomIndex].coord[Z] - 
								dnr.mol->atoms[dnr_index].coord[Z];
						
						TRANSLATE(tmp_dnr, dnr.mol->R);
						ROTATE(r_CD, dnr_m, tmp_dnr);

						tmp_dnr[X] = dnr.mol->atoms[dnr.atomIndex].coord[X];
						tmp_dnr[Y] = dnr.mol->atoms[dnr.atomIndex].coord[Y];
						tmp_dnr[Z] = dnr.mol->atoms[dnr.atomIndex].coord[Z];

						TRANSLATE(tmp_dnr, dnr.mol->R);
						ROTATE(r_h, dnr_m, tmp_dnr);

						if(acc.mol->current_step == 0){
							memcpy(tmp_acc, acc.mol->atoms[acc.atomIndex].coord, 3*sizeof(float));
							TRANSLATE(tmp_acc, acc.mol->R);
							ROTATE(r_DA, acc_m, tmp_acc);

							memcpy(tmp_dnr, dnr.mol->atoms[dnr.atomIndex].coord, 3*sizeof(float));
							TRANSLATE(tmp_dnr, dnr.mol->R);
							ROTATE(tmp, dnr_m, tmp_dnr);

							DIFF(r_DA, r_DA, tmp);
						}
						else{
							int acc_step = acc.mol->current_step - 1;

							memcpy(tmp_acc, &(acc.mol->traj[acc_step][3*acc.atomIndex]), 3*sizeof(float));
							TRANSLATE(tmp_acc, acc.mol->R);
							ROTATE(r_DA, acc_m, tmp_acc);

							memcpy(tmp_dnr, dnr.mol->atoms[dnr.atomIndex].coord, 3*sizeof(float));
							TRANSLATE(tmp_dnr, dnr.mol->R);
							ROTATE(tmp, dnr_m, tmp_dnr);

							DIFF(r_DA, r_DA, tmp);
						}
					}
					else{
						int dnr_step = dnr.mol->current_step - 1;

						tmp[X] = dnr.mol->traj[dnr_step][3*dnr.atomIndex] - 
								dnr.mol->traj[dnr_step][3*dnr_index];
						tmp[Y] = dnr.mol->traj[dnr_step][3*dnr.atomIndex + Y] - 
								dnr.mol->traj[dnr_step][3*dnr_index + Y];
						tmp[Z] = dnr.mol->traj[dnr_step][3*dnr.atomIndex + Z] - 
								dnr.mol->traj[dnr_step][3*dnr_index + Z];

						TRANSLATE(tmp, dnr.mol->R);
						ROTATE(r_CD, dnr_m, tmp);

						tmp[X] = dnr.mol->traj[dnr_step][3*dnr.atomIndex];
						tmp[Y] = dnr.mol->traj[dnr_step][3*dnr.atomIndex + Y];
						tmp[Z] = dnr.mol->traj[dnr_step][3*dnr.atomIndex + Z];

						TRANSLATE(tmp, dnr.mol->R);
						ROTATE(r_h, dnr_m, tmp);

						if(acc.mol->current_step == 0){
							memcpy(tmp_acc, acc.mol->atoms[acc.atomIndex].coord, 3*sizeof(float));
							TRANSLATE(tmp_acc, acc.mol->R);
							ROTATE(r_DA, acc_m, tmp_acc);

							memcpy(tmp_dnr, &(dnr.mol->traj[dnr_step][3*dnr.atomIndex]), 3*sizeof(float));
							TRANSLATE(tmp_dnr, dnr.mol->R);
							ROTATE(tmp, dnr_m, tmp_dnr);

							DIFF(r_DA, r_DA, tmp);
						}
						else{
							int acc_step = acc.mol->current_step - 1;

							memcpy(tmp_acc, &(acc.mol->traj[acc_step][3*acc.atomIndex]), 3*sizeof(float));
							TRANSLATE(tmp_acc, acc.mol->R);
							ROTATE(r_DA, acc_m, tmp_acc);

							memcpy(tmp_dnr, &(dnr.mol->traj[dnr_step][3*dnr.atomIndex]), 3*sizeof(float));
							TRANSLATE(tmp_dnr, dnr.mol->R);
							ROTATE(tmp, dnr_m, tmp_dnr);

							DIFF(r_DA, r_DA, tmp);
						}
					}
					
					mag = (float)(sqrt(DOT(r_DA, r_DA)));

					if(mag < PRECISION)
						return FALSE;

					mag = 1.0f/mag;

					r_DA[X] *= mag;
					r_DA[Y] *= mag;
					r_DA[Z] *= mag;

					mag = (float)(sqrt(DOT(r_CD,r_CD)));

					if(mag < PRECISION)
						return FALSE;
					
					mag = 1.0f/mag;

					r_CD[X] *= mag;
					r_CD[Y] *= mag;
					r_CD[Z] *= mag;
					
					mag = DOT(r_CD, r_DA);

					r_perp[X] = r_DA[X] - r_CD[X]*mag;
					r_perp[Y] = r_DA[Y] - r_CD[Y]*mag;
					r_perp[Z] = r_DA[Z] - r_CD[Z]*mag;

					mag = (float)(sqrt(DOT(r_perp,r_perp)));

					if(mag < PRECISION)
						return FALSE;	
					
					mag = 1.0f/mag;

					r_perp[X] *= mag;
					r_perp[Y] *= mag;
					r_perp[Z] *= mag;
					
					float s = (float)(sin(h_angle));
					float c = (float)(cos(h_angle));

					r_h[X] = r_h[X] + r_perp[X]*s + r_CD[X]*c;
					r_h[Y] = r_h[Y] + r_perp[Y]*s + r_CD[Y]*c;
					r_h[Z] = r_h[Z] + r_perp[Z]*s + r_CD[Z]*c;

					return TRUE;
				}
				break;
			case 2:
				{
					// This atom belongs to the peptide plane. The hydrogen position is fixed
					// by the peptide plane.
					float A[3], B[3];

					if(dnr.mol->current_step == 0){
						
						int dnr_index_0 = dnr.mol->mol_connect[dnr.atomIndex][0];
						int dnr_index_1 = dnr.mol->mol_connect[dnr.atomIndex][1];

						A[X] = dnr.mol->atoms[dnr.atomIndex].coord[X] - 
							dnr.mol->atoms[dnr_index_0].coord[X];
						A[Y] = dnr.mol->atoms[dnr.atomIndex].coord[Y] - 
							dnr.mol->atoms[dnr_index_0].coord[Y];
						A[Z] = dnr.mol->atoms[dnr.atomIndex].coord[Z] - 
							dnr.mol->atoms[dnr_index_0].coord[Z];

						B[X] = dnr.mol->atoms[dnr.atomIndex].coord[X] - 
							dnr.mol->atoms[dnr_index_1].coord[X];
						B[Y] = dnr.mol->atoms[dnr.atomIndex].coord[Y] - 
							dnr.mol->atoms[dnr_index_1].coord[Y];
						B[Z] = dnr.mol->atoms[dnr.atomIndex].coord[Z] - 
							dnr.mol->atoms[dnr_index_1].coord[Z];

						r_h[X] = dnr.mol->atoms[dnr.atomIndex].coord[X];
						r_h[Y] = dnr.mol->atoms[dnr.atomIndex].coord[Y];
						r_h[Z] = dnr.mol->atoms[dnr.atomIndex].coord[Z];
					}
					else{ // dnr.mol->current_step != 0
						int dnr_step = dnr.mol->current_step - 1;

						int dnr_index_0 = dnr.mol->mol_connect[dnr.atomIndex][0];
						int dnr_index_1 = dnr.mol->mol_connect[dnr.atomIndex][1];

						A[X] = dnr.mol->traj[dnr_step][3*dnr.atomIndex] - 
							dnr.mol->traj[dnr_step][3*dnr_index_0];
						A[Y] = dnr.mol->traj[dnr_step][3*dnr.atomIndex + Y] - 
							dnr.mol->traj[dnr_step][3*dnr_index_0 + Y];
						A[Z] = dnr.mol->traj[dnr_step][3*dnr.atomIndex + Z] - 
							dnr.mol->traj[dnr_step][3*dnr_index_0 + Z];

						B[X] = dnr.mol->traj[dnr_step][3*dnr.atomIndex] - 
							dnr.mol->traj[dnr_step][3*dnr_index_1];
						B[Y] = dnr.mol->traj[dnr_step][3*dnr.atomIndex + Y] - 
							dnr.mol->traj[dnr_step][3*dnr_index_1 + Y];
						B[Z] = dnr.mol->traj[dnr_step][3*dnr.atomIndex + Z] - 
							dnr.mol->traj[dnr_step][3*dnr_index_1 + Z];

						r_h[X] = dnr.mol->traj[dnr_step][3*dnr.atomIndex];
						r_h[Y] = dnr.mol->traj[dnr_step][3*dnr.atomIndex + Y];
						r_h[Z] = dnr.mol->traj[dnr_step][3*dnr.atomIndex + Z];
					}

					A[X] += B[X];
					A[Y] += B[Y];
					A[Z] += B[Z];
					
					float mag = (float)(sqrt(DOT(A,A)));

					// Don't divide by zero!
					if(mag < PRECISION)
						return FALSE;
					
					mag = 1.0f/mag;

					r_h[X] += A[X]*mag;
					r_h[Y] += A[Y]*mag;
					r_h[Z] += A[Z]*mag;

					return TRUE;
				}
				break;
		};

		return FALSE;
	}
	
	return FALSE;
}


BOOL PlaceHydrogen(array<float> &r_h, int dnr, int acc, molecule *mol)
{
	if(mol->atoms[dnr].elem == ELEM_N){
		// What is the connectivity of the donor atom?
		switch(mol->mol_connect[dnr].Length()){
			case 1:
				{
					// This atom belongs to a side chain. Assume the hydrogen atoms are
					// free to align (subject to a bond angle constraint)
					// with the acceptor atom when computing the hydrogen location.
					//
					//			 A
					//		    .
					//		   .
					//		  .
					//	     H
					//      /
					// C-- D
					float r_DA[3], r_CD[3], r_perp[3];
					float mag;
					//const float h_angle = 1.95f;
					const float h_angle = 1.2f;
					
					int dnr_index = mol->mol_connect[dnr][0];

					if(mol->current_step == 0){
						r_CD[X] = mol->atoms[dnr].coord[X] - 
								mol->atoms[dnr_index].coord[X];
						r_CD[Y] = mol->atoms[dnr].coord[Y] - 
								mol->atoms[dnr_index].coord[Y];
						r_CD[Z] = mol->atoms[dnr].coord[Z] - 
								mol->atoms[dnr_index].coord[Z];

						r_h[X] = mol->atoms[dnr].coord[X];
						r_h[Y] = mol->atoms[dnr].coord[Y];
						r_h[Z] = mol->atoms[dnr].coord[Z];

						r_DA[X] = mol->atoms[acc].coord[X] - 
							mol->atoms[dnr].coord[X];
						r_DA[Y] = mol->atoms[acc].coord[Y] - 
							mol->atoms[dnr].coord[Y];
						r_DA[Z] = mol->atoms[acc].coord[Z] - 
							mol->atoms[dnr].coord[Z];
					}
					else{ // mol->current_step != 0
						int mol_step = mol->current_step - 1;

						r_CD[X] = mol->traj[mol_step][3*dnr] - 
								mol->traj[mol_step][3*dnr_index];
						r_CD[Y] = mol->traj[mol_step][3*dnr+ Y] - 
								mol->traj[mol_step][3*dnr_index + Y];
						r_CD[Z] = mol->traj[mol_step][3*dnr + Z] - 
								mol->traj[mol_step][3*dnr_index + Z];

						r_h[X] = mol->traj[mol_step][3*dnr];
						r_h[Y] = mol->traj[mol_step][3*dnr + Y];
						r_h[Z] = mol->traj[mol_step][3*dnr + Z];

						r_DA[X] = mol->traj[mol_step][3*acc] - 
							mol->traj[mol_step][3*dnr];
						r_DA[Y] = mol->traj[mol_step][3*acc + Y] - 
							mol->traj[mol_step][3*dnr + Y];
						r_DA[Z] = mol->traj[mol_step][3*acc + Z] - 
							mol->traj[mol_step][3*dnr + Z];
					}
					
					mag = (float)(sqrt(DOT(r_DA, r_DA)));

					if(mag < PRECISION)
						return FALSE;

					mag = 1.0f/mag;

					r_DA[X] *= mag;
					r_DA[Y] *= mag;
					r_DA[Z] *= mag;

					mag = (float)(sqrt(DOT(r_CD,r_CD)));

					if(mag < PRECISION)
						return FALSE;
					
					mag = 1.0f/mag;

					r_CD[X] *= mag;
					r_CD[Y] *= mag;
					r_CD[Z] *= mag;
					
					mag = DOT(r_CD, r_DA);

					r_perp[X] = r_DA[X] - r_CD[X]*mag;
					r_perp[Y] = r_DA[Y] - r_CD[Y]*mag;
					r_perp[Z] = r_DA[Z] - r_CD[Z]*mag;

					mag = (float)(sqrt(DOT(r_perp,r_perp)));

					if(mag < PRECISION)
						return FALSE;	
					
					mag = 1.0f/mag;

					r_perp[X] *= mag;
					r_perp[Y] *= mag;
					r_perp[Z] *= mag;
					
					float s = (float)(sin(h_angle));
					float c = (float)(cos(h_angle));

					r_h[X] = r_h[X] + r_perp[X]*s + r_CD[X]*c;
					r_h[Y] = r_h[Y] + r_perp[Y]*s + r_CD[Y]*c;
					r_h[Z] = r_h[Z] + r_perp[Z]*s + r_CD[Z]*c;

					return TRUE;
				}
				break;
			case 2:
				{
					// This atom belongs to the peptide plane. The hydrogen position is fixed
					// by the peptide plane.
					float A[3], B[3];

					if(mol->current_step == 0){
						
						int dnr_index_0 = mol->mol_connect[dnr][0];
						int dnr_index_1 = mol->mol_connect[dnr][1];

						A[X] = mol->atoms[dnr].coord[X] - 
							mol->atoms[dnr_index_0].coord[X];
						A[Y] = mol->atoms[dnr].coord[Y] - 
							mol->atoms[dnr_index_0].coord[Y];
						A[Z] = mol->atoms[dnr].coord[Z] - 
							mol->atoms[dnr_index_0].coord[Z];

						B[X] = mol->atoms[dnr].coord[X] - 
							mol->atoms[dnr_index_1].coord[X];
						B[Y] = mol->atoms[dnr].coord[Y] - 
							mol->atoms[dnr_index_1].coord[Y];
						B[Z] = mol->atoms[dnr].coord[Z] - 
							mol->atoms[dnr_index_1].coord[Z];

						r_h[X] = mol->atoms[dnr].coord[X];
						r_h[Y] = mol->atoms[dnr].coord[Y];
						r_h[Z] = mol->atoms[dnr].coord[Z];
					}
					else{ // mol->current_step != 0
						int mol_step = mol->current_step - 1;

						int dnr_index_0 = mol->mol_connect[dnr][0];
						int dnr_index_1 = mol->mol_connect[dnr][1];

						A[X] = mol->traj[mol_step][3*dnr] - 
							mol->traj[mol_step][3*dnr_index_0];
						A[Y] = mol->traj[mol_step][3*dnr + Y] - 
							mol->traj[mol_step][3*dnr_index_0 + Y];
						A[Z] = mol->traj[mol_step][3*dnr + Z] - 
							mol->traj[mol_step][3*dnr_index_0 + Z];

						B[X] = mol->traj[mol_step][3*dnr] - 
							mol->traj[mol_step][3*dnr_index_1];
						B[Y] = mol->traj[mol_step][3*dnr + Y] - 
							mol->traj[mol_step][3*dnr_index_1 + Y];
						B[Z] = mol->traj[mol_step][3*dnr + Z] - 
							mol->traj[mol_step][3*dnr_index_1 + Z];

						r_h[X] = mol->traj[mol_step][3*dnr];
						r_h[Y] = mol->traj[mol_step][3*dnr + Y];
						r_h[Z] = mol->traj[mol_step][3*dnr + Z];
					}

					A[X] += B[X];
					A[Y] += B[Y];
					A[Z] += B[Z];
					
					float mag = (float)(sqrt(DOT(A,A)));

					// Don't divide by zero!
					if(mag < PRECISION)
						return FALSE;
					
					mag = 1.0f/mag;

					r_h[X] += A[X]*mag;
					r_h[Y] += A[Y]*mag;
					r_h[Z] += A[Z]*mag;

					return TRUE;
				}
				break;
		};

		return FALSE;
	}
	
	return FALSE;
}