//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.
//
//	Look-and-feel inspired by Xmol 
//	(http://www.networkcs.com/msc/docs/xmol/).
//
//	OpenGL selection code and quaterion rotation matrix
//	formulation from molview demo program by Mark Kilgard
//	(http://reality.sgi.com/opengl/OpenGLforX.html). PostScript
//	rendering also from Mark Kilgard.
//
//	The printing routines are based on code from
//	Craig Fahrnbach <craignan@home.com> via
//	Uwe Kotyczka <kotyczka@bnm-gmbh.de>.
//
//	Covalent bond determination code based on VMD
//	http://www.ks.uiuc.edu/Research/vmd/
//
//	Stride (http://www.embl-heidelberg.de/stride/stride.html)
//	is used for secondary structure determination and has been 
//	included in Qmol with the kind permission of 
//	Dmitrij Frishman, PhD
//	Institute for Bioinformatics
//	GSF - Forschungszentrum fr Umwelt und Gesundheit, GmbH
//	Ingolstdter Landstrae 1,
//	D-85764 Neuherberg, Germany
//	Telephone: +49-89-3187-4201
//	Fax: +49-89-31873585
//	e-mail: d.frishman@gsf.de
//	WWW: http://mips.gsf.de/mips/staff/frishman/
// 
#ifndef __QM_LIGHT
#define __QM_LIGHT

#include <GL\gl.h>
#include "Excepts.h"

// Six faces of the cube + ambient light
#define	QM_NUM_LIGHTS	7
#define	QM_AMBIENT		(QM_NUM_LIGHTS - 1)

class LightSource {
private:
	GLfloat ambient_color[4];
	GLfloat diffuse_color[4];
	GLfloat specular_color[4];
	GLfloat position[4];

	int light_num;
	static int s_light_num;

public:
	LightSource()
	{
		light_num = s_light_num;
		s_light_num++;

		if((s_light_num - GL_LIGHT0) >= GL_MAX_LIGHTS)
			throw QmolError("Ehausted supply of OpenGL lights!", __FILE__, __LINE__);

		for(int i = 0;i < 4;i++){
			ambient_color[i] = 0.0f;
			diffuse_color[i] = 0.0f;
			specular_color[i] = 0.0f;
			position[i] = 0.0f;
		}
	};
	
	inline void SetPosition(GLfloat x, GLfloat y, GLfloat z)
	{
		position[0] = x;
		position[1] = y;
		position[2] = z;
	};

	inline void SetAmbient(GLfloat r, GLfloat g, GLfloat b)
	{
		ambient_color[0] = r;
		ambient_color[1] = g;
		ambient_color[2] = b;
	};

	inline void SetDiffuse(GLfloat r, GLfloat g, GLfloat b)
	{
		diffuse_color[0] = r;
		diffuse_color[1] = g;
		diffuse_color[2] = b;
	};

	inline void SetSpecular(GLfloat r, GLfloat g, GLfloat b)
	{
		specular_color[0] = r;
		specular_color[1] = g;
		specular_color[2] = b;
	};

	inline void Apply()
	{
		glLightfv(light_num, GL_POSITION, position);
		glLightfv(light_num, GL_AMBIENT, ambient_color);
		glLightfv(light_num, GL_DIFFUSE, diffuse_color);
		glLightfv(light_num, GL_SPECULAR, specular_color);
	};

	inline void GetAmbient(GLfloat &r, GLfloat &g, GLfloat &b)
	{
		r = ambient_color[0];
		g = ambient_color[1];
		b = ambient_color[2];
	};

	inline void GetDiffuse(GLfloat &r, GLfloat &g, GLfloat &b)
	{
		r = diffuse_color[0];
		g = diffuse_color[1];
		b = diffuse_color[2];
	};

	void GetSpecular(GLfloat &r, GLfloat &g, GLfloat &b)
	{
		r = specular_color[0];
		g = specular_color[1];
		b = specular_color[2];
	};

	inline void Enable()
	{
		glEnable(light_num);
	};

	inline void Disable()
	{
		glDisable(light_num);
	};
};

class LightBox {
private:
	LightSource light_rack[QM_NUM_LIGHTS];
	BOOL enabled[QM_NUM_LIGHTS];

public:
	LightBox()
	{
		// Initialize the light positions
		// (by default, we're looking down the z axis)
		light_rack[0].SetPosition(0.0f, 0.0f, 1.0f);
		light_rack[1].SetPosition(1.0f, 0.0f, 0.0f);
		light_rack[2].SetPosition(0.0f, 1.0f, 0.0f);
		light_rack[3].SetPosition(0.0f, 0.0f, -1.0f);
		light_rack[4].SetPosition(-1.0f, 0.0f, 0.0f);
		light_rack[5].SetPosition(0.0f, -1.0f, 0.0f);

		// The ambient light source
		light_rack[6].SetPosition(0.0f, 0.0f, 1.0f);

		// Turn off all the lights
		for(int i = 0;i < QM_NUM_LIGHTS;i++)
			enabled[i] = FALSE;
		
		enabled[6] = TRUE;
	}

	inline void Apply()
	{
		for(int i = 0;i < QM_NUM_LIGHTS;i++)
			light_rack[i].Apply();
	}

	inline void TurnOn(int i)
	{
		#ifdef _DEBUG
		if((i < 0) || (i >= QM_NUM_LIGHTS))
			throw QmolError("Light index out of bounds!", __FILE__, __LINE__);
		#endif // _DEBUG
		
		enabled[i] = TRUE;
	};

	inline void TurnOff(int i)
	{
		#ifdef _DEBUG
		if((i < 0) || (i >= QM_NUM_LIGHTS))
			throw QmolError("Light index out of bounds!", __FILE__, __LINE__);
		#endif // _DEBUG
		
		enabled[i] = FALSE;
	};

	inline BOOL Status(int i)
	{
		#ifdef _DEBUG
		if((i < 0) || (i >= QM_NUM_LIGHTS))
			throw QmolError("Light index out of bounds!", __FILE__, __LINE__);
		#endif // _DEBUG
		
		return enabled[i];
	};

	inline void Enable()
	{
		for(int i = 0;i < QM_NUM_LIGHTS;i++)
			if(enabled[i])
				light_rack[i].Enable();
	};

	inline void Disable()
	{
		for(int i = 0;i < QM_NUM_LIGHTS;i++)
			if(enabled[i])
				light_rack[i].Disable();
	};

	inline void SetSpecular(int i, GLfloat r, GLfloat g, GLfloat b)
	{
		#ifdef _DEBUG
		if((i < 0) || (i >= QM_NUM_LIGHTS))
			throw QmolError("Light index out of bounds!", __FILE__, __LINE__);
		#endif // _DEBUG

		light_rack[i].SetSpecular(r, g, b);
	};

	inline void SetAmbient(int i, GLfloat r, GLfloat g, GLfloat b)
	{
		#ifdef _DEBUG
		if((i < 0) || (i >= QM_NUM_LIGHTS))
			throw QmolError("Light index out of bounds!", __FILE__, __LINE__);
		#endif // _DEBUG

		light_rack[i].SetAmbient(r, g, b);
	};

	inline void SetDiffuse(int i, GLfloat r, GLfloat g, GLfloat b)
	{
		#ifdef _DEBUG
		if((i < 0) || (i >= QM_NUM_LIGHTS))
			throw QmolError("Light index out of bounds!", __FILE__, __LINE__);
		#endif // _DEBUG

		light_rack[i].SetDiffuse(r, g, b);
	};

	inline void GetSpecular(int i, GLfloat &r, GLfloat &g, GLfloat &b)
	{
		#ifdef _DEBUG
		if((i < 0) || (i >= QM_NUM_LIGHTS))
			throw QmolError("Light index out of bounds!", __FILE__, __LINE__);
		#endif // _DEBUG

		light_rack[i].GetSpecular(r, g, b);
	};

	inline void GetAmbient(int i, GLfloat &r, GLfloat &g, GLfloat &b)
	{
		#ifdef _DEBUG
		if((i < 0) || (i >= QM_NUM_LIGHTS))
			throw QmolError("Light index out of bounds!", __FILE__, __LINE__);
		#endif // _DEBUG

		light_rack[i].GetAmbient(r, g, b);
	};

	inline void GetDiffuse(int i, GLfloat &r, GLfloat &g, GLfloat &b)
	{
		#ifdef _DEBUG
		if((i < 0) || (i >= QM_NUM_LIGHTS))
			throw QmolError("Light index out of bounds!", __FILE__, __LINE__);
		#endif // _DEBUG

		light_rack[i].GetDiffuse(r, g, b);
	};
};

#endif // __QM_LIGHT
