//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.
//
//	Look-and-feel inspired by Xmol 
//	(http://www.networkcs.com/msc/docs/xmol/).
//
//	OpenGL selection code and quaterion rotation matrix
//	formulation from molview demo program by Mark Kilgard
//	(http://reality.sgi.com/opengl/OpenGLforX.html). PostScript
//	rendering also from Mark Kilgard.
//
//	The printing routines are based on code from
//	Craig Fahrnbach <craignan@home.com> via
//	Uwe Kotyczka <kotyczka@bnm-gmbh.de>.
//
//	Covalent bond determination code based on VMD
//	http://www.ks.uiuc.edu/Research/vmd/
//
//	Stride (http://www.embl-heidelberg.de/stride/stride.html)
//	is used for secondary structure determination and has been 
//	included in Qmol with the kind permission of 
//	Dmitrij Frishman, PhD
//	Institute for Bioinformatics
//	GSF - Forschungszentrum fr Umwelt und Gesundheit, GmbH
//	Ingolstdter Landstrae 1,
//	D-85764 Neuherberg, Germany
//	Telephone: +49-89-3187-4201
//	Fax: +49-89-31873585
//	e-mail: d.frishman@gsf.de
//	WWW: http://mips.gsf.de/mips/staff/frishman/
// 
/////////////////////////////////////////////////////////////////////////////
#if !defined(AFX_QMOLDOC_H__979F19CC_DCC9_11D3_9099_00A0C9A3671F__INCLUDED_)
#define AFX_QMOLDOC_H__979F19CC_DCC9_11D3_9099_00A0C9A3671F__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#include "Qmol.h"
#include "stack.h"
#include "list.h"
#include "array.h"
#include "MoleculeState.h"
#include "pdb.h"
#include "surface.h"
#include "3dObject.h"
#include "SelectObject.h"
#include "Measure.h"

#define MAX(a,b) ((a)>(b)?(a):(b))
#define MIN(a,b) ((a)<(b)?(a):(b))

// The bond class stores the indicies of the
// two bonded atoms within a single molecule.
class bond{
	public:
		int a, b;
		BOOL colorBond;
		
		// Does this bond  belong to a solvent molecule?
		BOOL solvent;
		
		int operator==(const bond &temp)
		{
			if(((a == temp.a) && (b == temp.b)) ||
			   ((a == temp.b) && (b == temp.a)))
			   		return 1; // True
			
			return 0; // False
		};
};

// A visual link between two atoms
class monitor {
	public:
		int pairId;
		int atom_id[2];
		GLubyte color[3];
};

// An amino acid
class amino{
public:
	amino()
	{
		init();
	};

	void init()
	{
		aa = AA_DUMMY;
		ss = SS_DUMMY;
		resSeq = -1;
		c_alpha_index = -1;
		o_index = -1;
		major = 1.0;
		minor = 1.0;

		while(atom_id.Length())
			atom_id.Pop();
	};

	void SetAminoAcidType(char *m_type);
	BOOL SetCAlpha(char *m_name, int m_i);
	BOOL SetO(char *m_name, int m_i);
	void SetSSType(char m_ss);
	char *GetResName();

	// The amino acid type
	unsigned int aa;

	// Secondary structure type
	unsigned int ss;

	// The pdb amino acid number of this
	// residue.
	int resSeq;

	// The atom number of alpha carbon
	// (i.e. an index into the atoms array).
	int c_alpha_index;
	
	// The atom number of carbonyl oxygen
	// (i.e. an index into the atoms array).
	int o_index;

	// For ribbon and tube interpolations, the ellipse
	// dimensions for this residue
	float major, minor;

	// A list of the ids of every atom in the amino acid
	list<int> atom_id;
};

class amino_chain{
	public:
		amino_chain()
		{
			model = -1;
			chain = ' ';
		};

		// The id for this chain
		int model;
		char chain;

		// An array of amino acids that make up this chain.
		array<amino> sequence;
};

class molecule: public moleculeState, public DDD{
	private:
		void ZeroGCM(array<atom> &r);
		void ZeroGCM(list< array<float> > &r, int steps);

		void ZeroCM(array<atom> &r);
		void ZeroCM(list< array<float> > &r, array<atom> &ref, int steps);

	public:
		// Default constructor
		molecule()
		{
			size = 0;
			steps = 0;

			max_temp = 1.0;
			min_temp = 1.0;
			max_occ = 1.0;
			min_occ = 1.0;
		};
		
		molecule& operator=(molecule &copy);

		inline int NumAtoms()
		{
			return atoms.Length();
		};
		
		inline int NumBonds()
		{
			return bonds.GetLength();
		};
		
		// Return the number of amino acids
		int NumResidues(int &m_first);

		// A molecular surface
		MolSurface surf;

		// A fancy (i.e. overkill) way to implement an
		// array of atoms
		array<atom> atoms;
		
		// The protein sequence and secondary structure stored as
		// an array of models
		array<amino_chain> secondaryStruc;

		// Store the molecular connectivity as a stack of bonds
		stack<bond> bonds;

		// An list of atom monitors
		list<monitor> mon;
		
		// Dimensions of the molecule
		float xmin, ymin, zmin;
  		float xsize, ysize, zsize;
  		float maxdim;
		
		// The center of mass of each structure in the trajectory.
		// Stored as a 3*N array -> [cmx1, cmy1, cmz1, cmx2, cmy2, cmz2, ...]
		array<float> cm;
		
		// Trajectory Information
		int steps; // Number of time steps
		int size;  // size of the data sturcture (float, double, etc).
		
		// The maximum and minimum values of the temperature and occupancy (of a pdb
		// file).
		float max_temp, min_temp;
		float max_occ, min_occ;

		// Atomic radii
		float elem_radii_scheme[ELEM_LAST];

		// The trajectory data
		list <array <float> > traj;
		
		// Store the molecular connectivity.
		// The molecular connectivity is currently determined
		// within the ReadPdb() function - should be more general.
		array< array<unsigned int> > mol_connect;

		// Recompute the hydrogen bonds
		void UpdateHBonds(int cur_step);

		// Update the max temperature
		void UpdateTemperature();

		// Update the max occupancy
		void UpdateOccupancy();

		void UpdateBondColoring();

		// Color the residue "index" in chain chain_index according to
		// (R, G, B).
		void SetResidueColor(int index, int chain_index, 
					char R, char G, char B, char A);
		
		void SetAtomicRadiiScheme(RADIUS_FORMAT format, int elem, int index,
			float radius);

		// Compute the dimesions of the molecule (including all
		// trajectory elements) in the molecular frame.
		void GetMolDim(int trajProcess = QM_VERBATIM);

		// Get the dimensions of the current trajectory step
		// in the world frame.
		void GetCurrentMolDim(float &m_maxX, float &m_minX,
							  float &m_maxY, float &m_minY,
							  float &m_maxZ, float &m_minZ,
							  BOOL use_mol_frame = FALSE);

		// Add member functions for reading molecules (and other stuff) here.
		int AutoDetectFileType(const char *name);

		// Read is the wrapper function for reading different molecules
		void ReadCoord(const char *name, BOOL m_multi_model_traj, 
			BOOL m_connect_on_read, int file_type = QM_AUTO);

		// Read pdb files.
		void ReadPdb(pdb_file &pdb, BOOL m_multi_model_traj, BOOL m_connect_on_read);
		
		void ReadTrajHeader(const char *name, int file_type);
		void ReadDCDHeader(const char *trajName);
		
		void ReadTrajStep(const char *trajName, int file_type);
		void ReadDCDStep(const char *trajName);
		
		// Add member functions for savinf molecules (and other stuff) here.
		// The code itself lives in mol_file.cc
		// Save is the wrapper function for saving different molecules
		void SaveCoord(FILE *fout, int file_type, int precision,
			BOOL write_conect = TRUE, BOOL normalize_T = TRUE,
			int offset = 0);
		
		// Save pdb files.
		void SavePdb(FILE *fout, int precision, 
			BOOL write_conect = TRUE, BOOL normalize_T = TRUE,
			int offset = 0);

		// Export a trajectory file as a PDB file
		void TrajExportPdb(const char *file_name, int start, int stop);

		// Convert a user space atom id number to an atom index number.
		// Return -1 if the id cannot be found.
		int IdToIndex(int id);

		// Convert a user space residue id number to an residue index number.
		// Return -1 if the id cannot be found. Return the chain index in 
		// chain_index.
		int ResIdToIndex(int id, int &chain_index);

		BOOL ComputeSecondaryStructure();

		// Call the Stride program to assign secondary structure
		BOOL StrideWrapper(int first_chain, int NChain);

		// Set the dimension of the ribbon
		void SetRibbonSize(int m_type);

		// Hydrogen bond info
		BOOL IsHBondDonor(int i);
		BOOL IsHBondAcceptor(int i);
};

// A class to handle hydrogen bond (which can be between
// different molecules).
class QmolHBond{
public:
	molecule *molA, *molB;
	int indexA, indexB;
};

class CQmolDoc : public CDocument
{
protected: // create from serialization only
	CQmolDoc();
	DECLARE_DYNCREATE(CQmolDoc)

// Attributes
public:
	// The list of currently loaded molecules
	list<molecule *> molList;

	// The primary molecule (certain single molecule operations
	// only apply to the active molecule).
	molecule *activeMol;
	
	// Measurements between molecules
	measure molMetric;

	// Store the hydrogen bonds as an array of bonds
	list<QmolHBond> hbonds;

	// The color scheme for the molecule(s)
	GLubyte color_scheme[ELEM_LAST][4];
	
	// The structure-based color scheme for the molecule(s)
	GLubyte color_scheme_struc[SS_LAST][4];
	
	// The amino acid property-based color scheme for the molecule(s)
	GLubyte color_scheme_amino[AA_LAST][4];
	
	// Should the molecular connectivity be computed 
	// when the molecule is read in?
	BOOL connect_on_read;
	
	// Should the secondary structure be assigned when the 
	// molecule is read in?
	BOOL secondary_struct_on_read;
	
	// Should we interpret multiple pdb models as a 
	// trajectory?
	BOOL multi_model_traj;

	// How should we process trajectory created by reading multiple
	// structures from a coordinate file?
	int multi_model_traj_process;

	// Atomic masses for the molecule(s)
	float elem_mass_scheme[ELEM_LAST];

	void SetColorScheme(int m_scheme);
	
	// Set the mass array to its default value
	void SetAtomicMassScheme();

	void ColorByStruc();
	void ColorByAmino();

	// Is the specified object valid? That is, does it refer to
	// a valid molecule and atom or bond?
	BOOL IsValidObj(SelectObj &obj);
	BOOL IsValidMol(molecule *m_mol);

	// Are all molecules currently selected?
	BOOL AllSelected();

	// Are all molecules currently free to move and rotate?
	BOOL AllDynamic();

	float GetMaxDim(BOOL m_scaleToFit);

// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CQmolDoc)
	//}}AFX_VIRTUAL

	BOOL ReadFile(LPCSTR filename);
	BOOL OnFileCoordOpen();
	BOOL OnNetFileCoordOpen();
	BOOL OnFileCoordSave();
	BOOL OnFileOpenCoordTraj();

	void OnFileTrajExport(const char *file_name, int file_type,
		int start, int stop);

	void UpdateHydrogenBonding();

	inline int CurrentTrajStep()
	{
		return currentStep;
	};

	void CurrentTrajStep(int m_step);

	inline int MaxTrajStep()
	{
		return maxStep;
	};

	void IncrementTrajStep(int m_direction);
	void UpdateTraj();

	// Select a single molecule
	void SelectMol(molecule *m_mol);

private:
	// A list of previously used servers for sending queries to
	// the Protein Data bank.
	list<CString> server_list;

	// A list of previous quiries to the Protein Data bank.
	list<CString> pdb_query_list;

	// The current step for animating molecular trajectories
	int currentStep;

	// The maximum number of frames for all molecular trajectories
	int maxStep;

// Implementation
public:
	virtual ~CQmolDoc();
#ifdef _DEBUG
	virtual void AssertValid() const;
	virtual void Dump(CDumpContext& dc) const;
#endif

protected:

// Generated message map functions
protected:
	//{{AFX_MSG(CQmolDoc)
	afx_msg void OnOptionsRead();
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};

/////////////////////////////////////////////////////////////////////////////

//{{AFX_INSERT_LOCATION}}
// Microsoft Developer Studio will insert additional declarations immediately before the previous line.

#endif // !defined(AFX_QMOLDOC_H__979F19CC_DCC9_11D3_9099_00A0C9A3671F__INCLUDED_)

