//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.
//
//	Look-and-feel inspired by Xmol 
//	(http://www.networkcs.com/msc/docs/xmol/).
//
//	OpenGL selection code and quaterion rotation matrix
//	formulation from molview demo program by Mark Kilgard
//	(http://reality.sgi.com/opengl/OpenGLforX.html). PostScript
//	rendering also from Mark Kilgard.
//
//	The printing routines are based on code from
//	Craig Fahrnbach <craignan@home.com> via
//	Uwe Kotyczka <kotyczka@bnm-gmbh.de>.
//
//	Covalent bond determination code based on VMD
//	http://www.ks.uiuc.edu/Research/vmd/
//
//	Stride (http://www.embl-heidelberg.de/stride/stride.html)
//	is used for secondary structure determination and has been 
//	included in Qmol with the kind permission of 
//	Dmitrij Frishman, PhD
//	Institute for Bioinformatics
//	GSF - Forschungszentrum fr Umwelt und Gesundheit, GmbH
//	Ingolstdter Landstrae 1,
//	D-85764 Neuherberg, Germany
//	Telephone: +49-89-3187-4201
//	Fax: +49-89-31873585
//	e-mail: d.frishman@gsf.de
//	WWW: http://mips.gsf.de/mips/staff/frishman/

#include "stdafx.h"
#include "QmolView.h"

extern int g_purge_mru;

// Helper functions
void WriteStringInt(char *defaultValue, int m_int);
void WriteStringBool(char *defaultValue, int m_bool);
void WriteStringFloat(char *defaultValue, float m_float);
void WriteStringRGB(char *defaultValue, float m_color[3]);
void WriteStringRGBA(char *defaultValue, float m_color[4]);

int ReadStringInt(int &m_int, const char *m_value);
int ReadStringBool(int &m_bool, const char *m_value);
int ReadStringFloat(float &m_float, const char *m_value);
int ReadStringRGB(float m_color[3], const char *m_value);
int ReadStringRGBA(float m_color[4], const char *m_value);

// Read in default values from the system registry. To make the registry
// values human readable, all entries are stored as strings. The default
// values set above serve as the default values in the event that no
// valid registry entry is found.
void CQmolView::LoadRegistry()
{
	char defaultValue[1024];
	CString regValue;
	int tmpInt;
	float tmpFloat;

	/////////////////////////////////////////////////////////////////////////
	// int print_color_depth
	// print_color_depth can only be 16 or 24
	/////////////////////////////////////////////////////////////////////////
	WriteStringInt(defaultValue, print_color_depth);
	regValue = AfxGetApp()->GetProfileString("Settings", "printColorDepth", defaultValue);
	ReadStringInt(tmpInt, LPCSTR(regValue));
	
	if((tmpInt == 16) || (tmpInt == 24)){
		print_color_depth = tmpInt;
	}

	/////////////////////////////////////////////////////////////////////////
	// The specular material settings
	/////////////////////////////////////////////////////////////////////////
	WriteStringRGBA(defaultValue, mat_specular);
	regValue = AfxGetApp()->GetProfileString("Settings", "matSpecular", defaultValue);
	ReadStringRGBA(mat_specular, LPCSTR(regValue));

	/////////////////////////////////////////////////////////////////////////
	// float mat_shininess[0]
	// The specular shininess value of the material settings
	/////////////////////////////////////////////////////////////////////////
	WriteStringFloat(defaultValue, mat_shininess[0]);
	regValue = AfxGetApp()->GetProfileString("Settings", "matShininess", defaultValue);
	ReadStringFloat(tmpFloat, LPCSTR(regValue));
	
	// 0.0f <= mat_shininess[0] <= 100.0
	if((tmpFloat >= 0.0f) && (tmpFloat <= 100.0f)){
		mat_shininess[0] = tmpFloat;
	}
	
	/////////////////////////////////////////////////////////////////////////g
	// The background color
	/////////////////////////////////////////////////////////////////////////
	WriteStringRGBA(defaultValue, background);
	regValue = AfxGetApp()->GetProfileString("Settings", "backgroundColor", defaultValue);
	ReadStringRGBA(background, LPCSTR(regValue));

	/////////////////////////////////////////////////////////////////////////
	// The selection color
	/////////////////////////////////////////////////////////////////////////
	WriteStringRGBA(defaultValue, selection_color);
	regValue = AfxGetApp()->GetProfileString("Settings", "selectionColor", defaultValue);
	ReadStringRGBA(selection_color, LPCSTR(regValue));

	/////////////////////////////////////////////////////////////////////////
	// The H bond color
	/////////////////////////////////////////////////////////////////////////
	WriteStringRGB(defaultValue, hbond_color);
	regValue = AfxGetApp()->GetProfileString("Settings", "hBondColor", defaultValue);
	ReadStringRGB(hbond_color, LPCSTR(regValue));
	
	/////////////////////////////////////////////////////////////////////////
	// The depthCue flag
	/////////////////////////////////////////////////////////////////////////
	WriteStringBool(defaultValue, depthCue);
	regValue = AfxGetApp()->GetProfileString("Settings", "depthCue", defaultValue);
	ReadStringBool(depthCue, LPCSTR(regValue));

	/////////////////////////////////////////////////////////////////////////
	// The lineAlias flag
	/////////////////////////////////////////////////////////////////////////
	WriteStringBool(defaultValue, lineAlias);
	regValue = AfxGetApp()->GetProfileString("Settings", "lineAlias", defaultValue);
	ReadStringBool(lineAlias, LPCSTR(regValue));

	/////////////////////////////////////////////////////////////////////////
	// The smooth3D flag
	/////////////////////////////////////////////////////////////////////////
	WriteStringBool(defaultValue, smooth3D);
	regValue = AfxGetApp()->GetProfileString("Settings", "smooth3D", defaultValue);
	ReadStringBool(smooth3D, LPCSTR(regValue));

	/////////////////////////////////////////////////////////////////////////
	// The displayStereo value
	/////////////////////////////////////////////////////////////////////////
	WriteStringInt(defaultValue, displayStereo);
	regValue = AfxGetApp()->GetProfileString("Settings", "displayStereo", defaultValue);
	ReadStringInt(tmpInt, LPCSTR(regValue));

	if((tmpInt == QM_STEREO_SPLIT_SCREEN) || (QM_STEREO_NONE)){
		displayStereo = tmpInt;
	}
	
	/////////////////////////////////////////////////////////////////////////
	// The clipView flag
	/////////////////////////////////////////////////////////////////////////
	WriteStringBool(defaultValue, clipView);
	regValue = AfxGetApp()->GetProfileString("Settings", "clipView", defaultValue);
	ReadStringBool(clipView, LPCSTR(regValue));

	/////////////////////////////////////////////////////////////////////////
	// The clipSurface flag
	/////////////////////////////////////////////////////////////////////////
	WriteStringBool(defaultValue, clipSurface);
	regValue = AfxGetApp()->GetProfileString("Settings", "clipSurface", defaultValue);
	ReadStringBool(clipSurface, LPCSTR(regValue));

	/////////////////////////////////////////////////////////////////////////
	// The displayClipPlane flag
	/////////////////////////////////////////////////////////////////////////
	WriteStringBool(defaultValue, displayClipPlane);
	regValue = AfxGetApp()->GetProfileString("Settings", "displayClipPlane", defaultValue);
	ReadStringBool(displayClipPlane, LPCSTR(regValue));
	
	/////////////////////////////////////////////////////////////////////////
	// The clipPlaneClr color
	/////////////////////////////////////////////////////////////////////////
	WriteStringRGBA(defaultValue, clipPlaneClr);
	regValue = AfxGetApp()->GetProfileString("Settings", "clipPlaneColor", defaultValue);
	ReadStringRGBA(clipPlaneClr, LPCSTR(regValue));
	
	clipper.SetColor(clipPlaneClr);
	
	/////////////////////////////////////////////////////////////////////////
	// The resolution value
	/////////////////////////////////////////////////////////////////////////
	WriteStringInt(defaultValue, resolution);
	regValue = AfxGetApp()->GetProfileString("Settings", "resolution", defaultValue);
	ReadStringInt(tmpInt, LPCSTR(regValue));

	if((tmpInt >= HIGH_RES) && (tmpInt <= CUSTOM_RES)){
		resolution = tmpInt;
	}
	
	/////////////////////////////////////////////////////////////////////////
	// The resolution value
	/////////////////////////////////////////////////////////////////////////
	WriteStringInt(defaultValue, resolution);
	regValue = AfxGetApp()->GetProfileString("Settings", "resolution", defaultValue);
	ReadStringInt(tmpInt, LPCSTR(regValue));

	if((tmpInt >= HIGH_RES) && (tmpInt <= CUSTOM_RES)){
		resolution = tmpInt;
	}
	
	////////////////////////////////////////////////////////////////////////
	// The atomRes value
	/////////////////////////////////////////////////////////////////////////
	WriteStringInt(defaultValue, atomRes);
	regValue = AfxGetApp()->GetProfileString("Settings", "atomRes", defaultValue);
	ReadStringInt(tmpInt, LPCSTR(regValue));

	if((tmpInt >= 2) && (tmpInt <= 75)){
		atomRes = tmpInt;
	}

	////////////////////////////////////////////////////////////////////////
	// The bondRes value
	/////////////////////////////////////////////////////////////////////////
	WriteStringInt(defaultValue, bondRes);
	regValue = AfxGetApp()->GetProfileString("Settings", "bondRes", defaultValue);
	ReadStringInt(tmpInt, LPCSTR(regValue));

	if((tmpInt >= 2) && (tmpInt <= 75)){
		bondRes = tmpInt;
	}

	////////////////////////////////////////////////////////////////////////
	// The selectionRes value
	/////////////////////////////////////////////////////////////////////////
	WriteStringInt(defaultValue, selectionRes);
	regValue = AfxGetApp()->GetProfileString("Settings", "selectionRes", defaultValue);
	ReadStringInt(tmpInt, LPCSTR(regValue));

	if((tmpInt >= 3) && (tmpInt <= 75)){
		selectionRes = tmpInt;
	}

	////////////////////////////////////////////////////////////////////////
	// The splineRes value
	/////////////////////////////////////////////////////////////////////////
	WriteStringInt(defaultValue, splineRes);
	regValue = AfxGetApp()->GetProfileString("Settings", "splineRes", defaultValue);
	ReadStringInt(tmpInt, LPCSTR(regValue));

	if((tmpInt >= 0) && (tmpInt <= 50)){
		splineRes = tmpInt;
	}

	////////////////////////////////////////////////////////////////////////
	// The ribbonRes value
	/////////////////////////////////////////////////////////////////////////
	WriteStringInt(defaultValue, ribbonRes);
	regValue = AfxGetApp()->GetProfileString("Settings", "ribbonRes", defaultValue);
	ReadStringInt(tmpInt, LPCSTR(regValue));

	if((tmpInt >= 3) && (tmpInt <= 50)){
		ribbonRes = tmpInt;
	}

	////////////////////////////////////////////////////////////////////////
	// The splineTension value
	/////////////////////////////////////////////////////////////////////////
	WriteStringFloat(defaultValue, splineTension);
	regValue = AfxGetApp()->GetProfileString("Settings", "splineTension", defaultValue);
	ReadStringFloat(tmpFloat, LPCSTR(regValue));

	if((tmpFloat >= 1.0f) && (tmpFloat <= 2.0f)){
		splineTension = tmpFloat;
	}

	////////////////////////////////////////////////////////////////////////
	// The line_width value
	/////////////////////////////////////////////////////////////////////////
	WriteStringFloat(defaultValue, line_width);
	regValue = AfxGetApp()->GetProfileString("Settings", "lineWidth", defaultValue);
	ReadStringFloat(tmpFloat, LPCSTR(regValue));

	if((tmpFloat >= 1.0f) && (tmpFloat <= 10.0f)){
		line_width = tmpFloat;
	}
	
	float tmpRGBColor[3];

	/////////////////////////////////////////////////////////////////////////
	// The scaleToFit flag
	/////////////////////////////////////////////////////////////////////////
	WriteStringBool(defaultValue, scaleToFit);
	regValue = AfxGetApp()->GetProfileString("Settings", "scaleToFit", defaultValue);
	ReadStringBool(scaleToFit, LPCSTR(regValue));
	

	/////////////////////////////////////////////////////////////////////////
	// The Solid object light model [0]
	/////////////////////////////////////////////////////////////////////////
	lb_solid.GetSpecular(0, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringRGB(defaultValue, tmpRGBColor);
	regValue = AfxGetApp()->GetProfileString("Settings", "SolidSpecular0", defaultValue);
	ReadStringRGB(tmpRGBColor, LPCSTR(regValue));
	lb_solid.SetSpecular(0, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	lb_solid.GetDiffuse(0, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringRGB(defaultValue, tmpRGBColor);
	regValue = AfxGetApp()->GetProfileString("Settings", "SolidDiffuse0", defaultValue);
	ReadStringRGB(tmpRGBColor, LPCSTR(regValue));
	lb_solid.SetDiffuse(0, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringBool(defaultValue, lb_solid.Status(0));
	regValue = AfxGetApp()->GetProfileString("Settings", "TurnOnSolid0", defaultValue);
	ReadStringBool(tmpInt, LPCSTR(regValue));
	
	if(tmpInt)
		lb_solid.TurnOn(0);
	else
		lb_solid.TurnOff(0);

	/////////////////////////////////////////////////////////////////////////
	// The Solid object light model [1]
	/////////////////////////////////////////////////////////////////////////
	lb_solid.GetSpecular(1, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringRGB(defaultValue, tmpRGBColor);
	regValue = AfxGetApp()->GetProfileString("Settings", "SolidSpecular1", defaultValue);
	ReadStringRGB(tmpRGBColor, LPCSTR(regValue));
	lb_solid.SetSpecular(1, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	lb_solid.GetDiffuse(1, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringRGB(defaultValue, tmpRGBColor);
	regValue = AfxGetApp()->GetProfileString("Settings", "SolidDiffuse1", defaultValue);
	ReadStringRGB(tmpRGBColor, LPCSTR(regValue));
	lb_solid.SetDiffuse(1, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringBool(defaultValue, lb_solid.Status(1));
	regValue = AfxGetApp()->GetProfileString("Settings", "TurnOnSolid1", defaultValue);
	ReadStringBool(tmpInt, LPCSTR(regValue));
	
	if(tmpInt)
		lb_solid.TurnOn(1);
	else
		lb_solid.TurnOff(1);

	/////////////////////////////////////////////////////////////////////////
	// The Solid object light model [2]
	/////////////////////////////////////////////////////////////////////////
	lb_solid.GetSpecular(2, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringRGB(defaultValue, tmpRGBColor);
	regValue = AfxGetApp()->GetProfileString("Settings", "SolidSpecular2", defaultValue);
	ReadStringRGB(tmpRGBColor, LPCSTR(regValue));
	lb_solid.SetSpecular(2, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	lb_solid.GetDiffuse(2, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringRGB(defaultValue, tmpRGBColor);
	regValue = AfxGetApp()->GetProfileString("Settings", "SolidDiffuse1", defaultValue);
	ReadStringRGB(tmpRGBColor, LPCSTR(regValue));
	lb_solid.SetDiffuse(2, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringBool(defaultValue, lb_solid.Status(2));
	regValue = AfxGetApp()->GetProfileString("Settings", "TurnOnSolid2", defaultValue);
	ReadStringBool(tmpInt, LPCSTR(regValue));
	
	if(tmpInt)
		lb_solid.TurnOn(2);
	else
		lb_solid.TurnOff(2);

	/////////////////////////////////////////////////////////////////////////
	// The Solid object light model [3]
	/////////////////////////////////////////////////////////////////////////
	lb_solid.GetSpecular(3, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringRGB(defaultValue, tmpRGBColor);
	regValue = AfxGetApp()->GetProfileString("Settings", "SolidSpecular3", defaultValue);
	ReadStringRGB(tmpRGBColor, LPCSTR(regValue));
	lb_solid.SetSpecular(3, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	lb_solid.GetDiffuse(3, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringRGB(defaultValue, tmpRGBColor);
	regValue = AfxGetApp()->GetProfileString("Settings", "SolidDiffuse3", defaultValue);
	ReadStringRGB(tmpRGBColor, LPCSTR(regValue));
	lb_solid.SetDiffuse(3, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringBool(defaultValue, lb_solid.Status(3));
	regValue = AfxGetApp()->GetProfileString("Settings", "TurnOnSolid3", defaultValue);
	ReadStringBool(tmpInt, LPCSTR(regValue));
	
	if(tmpInt)
		lb_solid.TurnOn(3);
	else
		lb_solid.TurnOff(3);

	/////////////////////////////////////////////////////////////////////////
	// The Solid object light model [4]
	/////////////////////////////////////////////////////////////////////////
	lb_solid.GetSpecular(4, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringRGB(defaultValue, tmpRGBColor);
	regValue = AfxGetApp()->GetProfileString("Settings", "SolidSpecular4", defaultValue);
	ReadStringRGB(tmpRGBColor, LPCSTR(regValue));
	lb_solid.SetSpecular(4, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	lb_solid.GetDiffuse(4, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringRGB(defaultValue, tmpRGBColor);
	regValue = AfxGetApp()->GetProfileString("Settings", "SolidDiffuse4", defaultValue);
	ReadStringRGB(tmpRGBColor, LPCSTR(regValue));
	lb_solid.SetDiffuse(4, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringBool(defaultValue, lb_solid.Status(4));
	regValue = AfxGetApp()->GetProfileString("Settings", "TurnOnSolid4", defaultValue);
	ReadStringBool(tmpInt, LPCSTR(regValue));
	
	if(tmpInt)
		lb_solid.TurnOn(4);
	else
		lb_solid.TurnOff(4);

	/////////////////////////////////////////////////////////////////////////
	// The Solid object light model [5]
	/////////////////////////////////////////////////////////////////////////
	lb_solid.GetSpecular(5, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringRGB(defaultValue, tmpRGBColor);
	regValue = AfxGetApp()->GetProfileString("Settings", "SolidSpecular5", defaultValue);
	ReadStringRGB(tmpRGBColor, LPCSTR(regValue));
	lb_solid.SetSpecular(5, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	lb_solid.GetDiffuse(5, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringRGB(defaultValue, tmpRGBColor);
	regValue = AfxGetApp()->GetProfileString("Settings", "SolidDiffuse5", defaultValue);
	ReadStringRGB(tmpRGBColor, LPCSTR(regValue));
	lb_solid.SetDiffuse(5, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringBool(defaultValue, lb_solid.Status(5));
	regValue = AfxGetApp()->GetProfileString("Settings", "TurnOnSolid5", defaultValue);
	ReadStringBool(tmpInt, LPCSTR(regValue));
	
	if(tmpInt)
		lb_solid.TurnOn(5);
	else
		lb_solid.TurnOff(5);

	/////////////////////////////////////////////////////////////////////////
	// The Solid object light model Ambient
	/////////////////////////////////////////////////////////////////////////
	lb_solid.GetAmbient(QM_AMBIENT, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringRGB(defaultValue, tmpRGBColor);
	regValue = AfxGetApp()->GetProfileString("Settings", "SolidAmbient", defaultValue);
	ReadStringRGB(tmpRGBColor, LPCSTR(regValue));
	lb_solid.SetAmbient(QM_AMBIENT, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringBool(defaultValue, lb_solid.Status(QM_AMBIENT));
	regValue = AfxGetApp()->GetProfileString("Settings", "TurnOnAmbient", defaultValue);
	ReadStringBool(tmpInt, LPCSTR(regValue));
	
	if(tmpInt)
		lb_solid.TurnOn(QM_AMBIENT);
	else
		lb_solid.TurnOff(QM_AMBIENT);
}

void CQmolView::SaveRegistry()
{
	char defaultValue[1024];

	/////////////////////////////////////////////////////////////////////////
	// int print_color_depth
	/////////////////////////////////////////////////////////////////////////
	WriteStringInt(defaultValue, print_color_depth);
	AfxGetApp()->WriteProfileString("Settings", "printColorDepth", defaultValue);

	/////////////////////////////////////////////////////////////////////////
	// The specular material settings
	/////////////////////////////////////////////////////////////////////////
	WriteStringRGBA(defaultValue, mat_specular);
	AfxGetApp()->WriteProfileString("Settings", "matSpecular", defaultValue);

	/////////////////////////////////////////////////////////////////////////
	// float mat_shininess[0]
	/////////////////////////////////////////////////////////////////////////
	WriteStringFloat(defaultValue, mat_shininess[0]);
	AfxGetApp()->WriteProfileString("Settings", "matShininess", defaultValue);
	
	/////////////////////////////////////////////////////////////////////////
	// The background color
	/////////////////////////////////////////////////////////////////////////
	WriteStringRGBA(defaultValue, background);
	AfxGetApp()->WriteProfileString("Settings", "backgroundColor", defaultValue);

	/////////////////////////////////////////////////////////////////////////
	// The selection color
	/////////////////////////////////////////////////////////////////////////
	WriteStringRGBA(defaultValue, selection_color);
	AfxGetApp()->WriteProfileString("Settings", "selectionColor", defaultValue);
	
	/////////////////////////////////////////////////////////////////////////
	// The H bond color
	/////////////////////////////////////////////////////////////////////////
	WriteStringRGB(defaultValue, hbond_color);
	AfxGetApp()->WriteProfileString("Settings", "hBondColor", defaultValue);
	
	/////////////////////////////////////////////////////////////////////////
	// The depthCue flag
	/////////////////////////////////////////////////////////////////////////
	WriteStringBool(defaultValue, depthCue);
	AfxGetApp()->WriteProfileString("Settings", "depthCue", defaultValue);

	/////////////////////////////////////////////////////////////////////////
	// The lineAlias flag
	/////////////////////////////////////////////////////////////////////////
	WriteStringBool(defaultValue, lineAlias);
	AfxGetApp()->WriteProfileString("Settings", "lineAlias", defaultValue);

	/////////////////////////////////////////////////////////////////////////
	// The smooth3D flag
	/////////////////////////////////////////////////////////////////////////
	WriteStringBool(defaultValue, smooth3D);
	AfxGetApp()->WriteProfileString("Settings", "smooth3D", defaultValue);

	/////////////////////////////////////////////////////////////////////////
	// The displayStereo value
	/////////////////////////////////////////////////////////////////////////
	WriteStringInt(defaultValue, displayStereo);
	AfxGetApp()->WriteProfileString("Settings", "displayStereo", defaultValue);
	
	/////////////////////////////////////////////////////////////////////////
	// The clipView flag
	/////////////////////////////////////////////////////////////////////////
	WriteStringBool(defaultValue, clipView);
	AfxGetApp()->WriteProfileString("Settings", "clipView", defaultValue);

	/////////////////////////////////////////////////////////////////////////
	// The clipSurface flag
	/////////////////////////////////////////////////////////////////////////
	WriteStringBool(defaultValue, clipSurface);
	AfxGetApp()->WriteProfileString("Settings", "clipSurface", defaultValue);

	/////////////////////////////////////////////////////////////////////////
	// The displayClipPlane flag
	/////////////////////////////////////////////////////////////////////////
	WriteStringBool(defaultValue, displayClipPlane);
	AfxGetApp()->WriteProfileString("Settings", "displayClipPlane", defaultValue);
	
	/////////////////////////////////////////////////////////////////////////
	// The clipPlaneClr color
	/////////////////////////////////////////////////////////////////////////
	WriteStringRGBA(defaultValue, clipPlaneClr);
	AfxGetApp()->WriteProfileString("Settings", "clipPlaneColor", defaultValue);
	
	/////////////////////////////////////////////////////////////////////////
	// The resolution value
	/////////////////////////////////////////////////////////////////////////
	WriteStringInt(defaultValue, resolution);
	AfxGetApp()->WriteProfileString("Settings", "resolution", defaultValue);
	
	/////////////////////////////////////////////////////////////////////////
	// The resolution value
	/////////////////////////////////////////////////////////////////////////
	WriteStringInt(defaultValue, resolution);
	AfxGetApp()->WriteProfileString("Settings", "resolution", defaultValue);
	
	////////////////////////////////////////////////////////////////////////
	// The atomRes value
	/////////////////////////////////////////////////////////////////////////
	WriteStringInt(defaultValue, atomRes);
	AfxGetApp()->WriteProfileString("Settings", "atomRes", defaultValue);

	////////////////////////////////////////////////////////////////////////
	// The bondRes value
	/////////////////////////////////////////////////////////////////////////
	WriteStringInt(defaultValue, bondRes);
	AfxGetApp()->WriteProfileString("Settings", "bondRes", defaultValue);

	////////////////////////////////////////////////////////////////////////
	// The selectionRes value
	/////////////////////////////////////////////////////////////////////////
	WriteStringInt(defaultValue, selectionRes);
	AfxGetApp()->WriteProfileString("Settings", "selectionRes", defaultValue);

	////////////////////////////////////////////////////////////////////////
	// The splineRes value
	/////////////////////////////////////////////////////////////////////////
	WriteStringInt(defaultValue, splineRes);
	AfxGetApp()->WriteProfileString("Settings", "splineRes", defaultValue);

	////////////////////////////////////////////////////////////////////////
	// The ribbonRes value
	/////////////////////////////////////////////////////////////////////////
	WriteStringInt(defaultValue, ribbonRes);
	AfxGetApp()->WriteProfileString("Settings", "ribbonRes", defaultValue);

	////////////////////////////////////////////////////////////////////////
	// The splineTension value
	/////////////////////////////////////////////////////////////////////////
	WriteStringFloat(defaultValue, splineTension);
	AfxGetApp()->WriteProfileString("Settings", "splineTension", defaultValue);

	////////////////////////////////////////////////////////////////////////
	// The line_width value
	/////////////////////////////////////////////////////////////////////////
	WriteStringFloat(defaultValue, line_width);
	AfxGetApp()->WriteProfileString("Settings", "lineWidth", defaultValue);
	
	/////////////////////////////////////////////////////////////////////////
	// The displayClipPlane flag
	/////////////////////////////////////////////////////////////////////////
	WriteStringBool(defaultValue, scaleToFit);
	AfxGetApp()->WriteProfileString("Settings", "scaleToFit", defaultValue);

	float tmpRGBColor[3];

	/////////////////////////////////////////////////////////////////////////
	// The Solid object light model [0]
	/////////////////////////////////////////////////////////////////////////
	lb_solid.GetSpecular(0, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringRGB(defaultValue, tmpRGBColor);
	AfxGetApp()->WriteProfileString("Settings", "SolidSpecular0", defaultValue);

	lb_solid.GetDiffuse(0, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringRGB(defaultValue, tmpRGBColor);
	AfxGetApp()->WriteProfileString("Settings", "SolidDiffuse0", defaultValue);

	WriteStringBool(defaultValue, lb_solid.Status(0));
	AfxGetApp()->WriteProfileString("Settings", "TurnOnSolid0", defaultValue);

	/////////////////////////////////////////////////////////////////////////
	// The Solid object light model [1]
	/////////////////////////////////////////////////////////////////////////
	lb_solid.GetSpecular(1, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringRGB(defaultValue, tmpRGBColor);
	AfxGetApp()->WriteProfileString("Settings", "SolidSpecular1", defaultValue);

	lb_solid.GetDiffuse(1, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringRGB(defaultValue, tmpRGBColor);
	AfxGetApp()->WriteProfileString("Settings", "SolidDiffuse1", defaultValue);

	WriteStringBool(defaultValue, lb_solid.Status(1));
	AfxGetApp()->WriteProfileString("Settings", "TurnOnSolid1", defaultValue);

	/////////////////////////////////////////////////////////////////////////
	// The Solid object light model [2]
	/////////////////////////////////////////////////////////////////////////
	lb_solid.GetSpecular(2, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringRGB(defaultValue, tmpRGBColor);
	AfxGetApp()->WriteProfileString("Settings", "SolidSpecular2", defaultValue);

	lb_solid.GetDiffuse(2, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringRGB(defaultValue, tmpRGBColor);
	AfxGetApp()->WriteProfileString("Settings", "SolidDiffuse1", defaultValue);

	WriteStringBool(defaultValue, lb_solid.Status(2));
	AfxGetApp()->WriteProfileString("Settings", "TurnOnSolid2", defaultValue);

	/////////////////////////////////////////////////////////////////////////
	// The Solid object light model [3]
	/////////////////////////////////////////////////////////////////////////
	lb_solid.GetSpecular(3, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringRGB(defaultValue, tmpRGBColor);
	AfxGetApp()->WriteProfileString("Settings", "SolidSpecular3", defaultValue);

	lb_solid.GetDiffuse(3, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringRGB(defaultValue, tmpRGBColor);
	AfxGetApp()->WriteProfileString("Settings", "SolidDiffuse3", defaultValue);

	WriteStringBool(defaultValue, lb_solid.Status(3));
	AfxGetApp()->WriteProfileString("Settings", "TurnOnSolid3", defaultValue);

	/////////////////////////////////////////////////////////////////////////
	// The Solid object light model [4]
	/////////////////////////////////////////////////////////////////////////
	lb_solid.GetSpecular(4, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringRGB(defaultValue, tmpRGBColor);
	AfxGetApp()->WriteProfileString("Settings", "SolidSpecular4", defaultValue);

	lb_solid.GetDiffuse(4, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringRGB(defaultValue, tmpRGBColor);
	AfxGetApp()->WriteProfileString("Settings", "SolidDiffuse4", defaultValue);

	WriteStringBool(defaultValue, lb_solid.Status(4));
	AfxGetApp()->WriteProfileString("Settings", "TurnOnSolid4", defaultValue);

	/////////////////////////////////////////////////////////////////////////
	// The Solid object light model [5]
	/////////////////////////////////////////////////////////////////////////
	lb_solid.GetSpecular(5, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringRGB(defaultValue, tmpRGBColor);
	AfxGetApp()->WriteProfileString("Settings", "SolidSpecular5", defaultValue);

	lb_solid.GetDiffuse(5, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringRGB(defaultValue, tmpRGBColor);
	AfxGetApp()->WriteProfileString("Settings", "SolidDiffuse5", defaultValue);

	WriteStringBool(defaultValue, lb_solid.Status(5));
	AfxGetApp()->WriteProfileString("Settings", "TurnOnSolid5", defaultValue);

	/////////////////////////////////////////////////////////////////////////
	// The Solid object light model Ambient
	/////////////////////////////////////////////////////////////////////////
	lb_solid.GetAmbient(QM_AMBIENT, tmpRGBColor[0], tmpRGBColor[1], tmpRGBColor[2]);

	WriteStringRGB(defaultValue, tmpRGBColor);
	AfxGetApp()->WriteProfileString("Settings", "SolidAmbient", defaultValue);

	WriteStringBool(defaultValue, lb_solid.Status(QM_AMBIENT));
	AfxGetApp()->WriteProfileString("Settings", "TurnOnAmbient", defaultValue);

}

// Remove all registry entries
void CQmolView::CleanRegistry()
{
	// Tell the app to flush the MRU file list
	g_purge_mru = TRUE;

	/////////////////////////////////////////////////////////////////////////
	// int print_color_depth
	/////////////////////////////////////////////////////////////////////////
	AfxGetApp()->WriteProfileString("Settings", "printColorDepth", NULL);

	/////////////////////////////////////////////////////////////////////////
	// The specular material settings
	/////////////////////////////////////////////////////////////////////////
	AfxGetApp()->WriteProfileString("Settings", "matSpecular", NULL);

	/////////////////////////////////////////////////////////////////////////
	// float mat_shininess[0]
	/////////////////////////////////////////////////////////////////////////
	AfxGetApp()->WriteProfileString("Settings", "matShininess", NULL);
	
	/////////////////////////////////////////////////////////////////////////
	// The background color
	/////////////////////////////////////////////////////////////////////////
	AfxGetApp()->WriteProfileString("Settings", "backgroundColor", NULL);

	/////////////////////////////////////////////////////////////////////////
	// The selection color
	/////////////////////////////////////////////////////////////////////////
	AfxGetApp()->WriteProfileString("Settings", "selectionColor", NULL);
	
	/////////////////////////////////////////////////////////////////////////
	// The H bond color
	/////////////////////////////////////////////////////////////////////////
	AfxGetApp()->WriteProfileString("Settings", "hBondColor", NULL);
	
	/////////////////////////////////////////////////////////////////////////
	// The depthCue flag
	/////////////////////////////////////////////////////////////////////////
	AfxGetApp()->WriteProfileString("Settings", "depthCue", NULL);

	/////////////////////////////////////////////////////////////////////////
	// The lineAlias flag
	/////////////////////////////////////////////////////////////////////////
	AfxGetApp()->WriteProfileString("Settings", "lineAlias", NULL);

	/////////////////////////////////////////////////////////////////////////
	// The smooth3D flag
	/////////////////////////////////////////////////////////////////////////
	AfxGetApp()->WriteProfileString("Settings", "smooth3D", NULL);

	/////////////////////////////////////////////////////////////////////////
	// The displayStereo value
	/////////////////////////////////////////////////////////////////////////
	AfxGetApp()->WriteProfileString("Settings", "displayStereo", NULL);
	
	/////////////////////////////////////////////////////////////////////////
	// The clipView flag
	/////////////////////////////////////////////////////////////////////////
	AfxGetApp()->WriteProfileString("Settings", "clipView", NULL);

	/////////////////////////////////////////////////////////////////////////
	// The clipSurface flag
	/////////////////////////////////////////////////////////////////////////
	AfxGetApp()->WriteProfileString("Settings", "clipSurface", NULL);

	/////////////////////////////////////////////////////////////////////////
	// The displayClipPlane flag
	/////////////////////////////////////////////////////////////////////////
	AfxGetApp()->WriteProfileString("Settings", "displayClipPlane", NULL);
	
	/////////////////////////////////////////////////////////////////////////
	// The clipPlaneClr color
	/////////////////////////////////////////////////////////////////////////
	AfxGetApp()->WriteProfileString("Settings", "clipPlaneColor", NULL);
	
	/////////////////////////////////////////////////////////////////////////
	// The resolution value
	/////////////////////////////////////////////////////////////////////////
	AfxGetApp()->WriteProfileString("Settings", "resolution", NULL);
	
	/////////////////////////////////////////////////////////////////////////
	// The resolution value
	/////////////////////////////////////////////////////////////////////////
	AfxGetApp()->WriteProfileString("Settings", "resolution", NULL);
	
	////////////////////////////////////////////////////////////////////////
	// The atomRes value
	/////////////////////////////////////////////////////////////////////////
	AfxGetApp()->WriteProfileString("Settings", "atomRes", NULL);

	////////////////////////////////////////////////////////////////////////
	// The bondRes value
	/////////////////////////////////////////////////////////////////////////
	AfxGetApp()->WriteProfileString("Settings", "bondRes", NULL);

	////////////////////////////////////////////////////////////////////////
	// The selectionRes value
	/////////////////////////////////////////////////////////////////////////
	AfxGetApp()->WriteProfileString("Settings", "selectionRes", NULL);

	////////////////////////////////////////////////////////////////////////
	// The splineRes value
	/////////////////////////////////////////////////////////////////////////
	AfxGetApp()->WriteProfileString("Settings", "splineRes", NULL);

	////////////////////////////////////////////////////////////////////////
	// The ribbonRes value
	/////////////////////////////////////////////////////////////////////////
	AfxGetApp()->WriteProfileString("Settings", "ribbonRes", NULL);

	////////////////////////////////////////////////////////////////////////
	// The splineTension value
	/////////////////////////////////////////////////////////////////////////
	AfxGetApp()->WriteProfileString("Settings", "splineTension", NULL);

	////////////////////////////////////////////////////////////////////////
	// The line_width value
	/////////////////////////////////////////////////////////////////////////
	AfxGetApp()->WriteProfileString("Settings", "lineWidth", NULL);

	/////////////////////////////////////////////////////////////////////////
	// The Solid object light model [0]
	/////////////////////////////////////////////////////////////////////////
	AfxGetApp()->WriteProfileString("Settings", "SolidSpecular0", NULL);

	AfxGetApp()->WriteProfileString("Settings", "SolidDiffuse0", NULL);

	AfxGetApp()->WriteProfileString("Settings", "TurnOnSolid0", NULL);

	/////////////////////////////////////////////////////////////////////////
	// The Solid object light model [1]
	/////////////////////////////////////////////////////////////////////////
	AfxGetApp()->WriteProfileString("Settings", "SolidSpecular1", NULL);

	AfxGetApp()->WriteProfileString("Settings", "SolidDiffuse1", NULL);

	AfxGetApp()->WriteProfileString("Settings", "TurnOnSolid1", NULL);

	/////////////////////////////////////////////////////////////////////////
	// The Solid object light model [2]
	/////////////////////////////////////////////////////////////////////////
	AfxGetApp()->WriteProfileString("Settings", "SolidSpecular2", NULL);

	AfxGetApp()->WriteProfileString("Settings", "SolidDiffuse1", NULL);

	AfxGetApp()->WriteProfileString("Settings", "TurnOnSolid2", NULL);

	/////////////////////////////////////////////////////////////////////////
	// The Solid object light model [3]
	/////////////////////////////////////////////////////////////////////////
	AfxGetApp()->WriteProfileString("Settings", "SolidSpecular3", NULL);

	AfxGetApp()->WriteProfileString("Settings", "SolidDiffuse3", NULL);

	AfxGetApp()->WriteProfileString("Settings", "TurnOnSolid3", NULL);

	/////////////////////////////////////////////////////////////////////////
	// The Solid object light model [4]
	/////////////////////////////////////////////////////////////////////////
	AfxGetApp()->WriteProfileString("Settings", "SolidSpecular4", NULL);

	AfxGetApp()->WriteProfileString("Settings", "SolidDiffuse4", NULL);

	AfxGetApp()->WriteProfileString("Settings", "TurnOnSolid4", NULL);

	/////////////////////////////////////////////////////////////////////////
	// The Solid object light model [5]
	/////////////////////////////////////////////////////////////////////////
	AfxGetApp()->WriteProfileString("Settings", "SolidSpecular5", NULL);

	AfxGetApp()->WriteProfileString("Settings", "SolidDiffuse5", NULL);

	AfxGetApp()->WriteProfileString("Settings", "TurnOnSolid5", NULL);

	/////////////////////////////////////////////////////////////////////////
	// The Solid object light model Ambient
	/////////////////////////////////////////////////////////////////////////
	AfxGetApp()->WriteProfileString("Settings", "SolidAmbient", NULL);

	AfxGetApp()->WriteProfileString("Settings", "TurnOnAmbient", NULL);
}

void WriteStringInt(char *defaultValue, int m_int)
{
	sprintf(defaultValue, "%d", m_int);
}

void WriteStringBool(char *defaultValue, int m_bool)
{
	if(m_bool == TRUE)
		sprintf(defaultValue, "TRUE");
	else
		sprintf(defaultValue, "FALSE");
}

void WriteStringFloat(char *defaultValue, float m_float)
{
	sprintf(defaultValue, "%.3f", m_float);
}

void WriteStringRGB(char *defaultValue, float m_color[3])
{
	sprintf(defaultValue, "%.3f %.3f %.3f", m_color[0], m_color[1], m_color[2]);
}

void WriteStringRGBA(char *defaultValue, float m_color[4])
{
	sprintf(defaultValue, "%.3f %.3f %.3f %.3f", 
		m_color[0], m_color[1], m_color[2], m_color[3]);
}


int ReadStringInt(int &m_int, const char *m_value)
{
	m_int = atoi(m_value);

	return TRUE;
}

int ReadStringBool(int &m_bool, const char *m_value)
{
	if(strcmpi(m_value, "TRUE") == 0){
		m_bool = TRUE;
		return TRUE;
	}

	if(strcmpi(m_value, "FALSE") == 0){
		m_bool = FALSE;
		return TRUE;
	}

	return FALSE;
}

int ReadStringFloat(float &m_float, const char *m_value)
{
	m_float = float(atof(m_value));

	return TRUE;
}

int ReadStringRGB(float m_color[3], const char *m_value)
{
	if(sscanf(m_value, "%f %f %f", &(m_color[0]), &(m_color[1]), &(m_color[2])) != 3)
		return FALSE;

	return TRUE;
}

int ReadStringRGBA(float m_color[4], const char *m_value)
{
	if(sscanf(m_value, "%f %f %f %f", 
		&(m_color[0]), &(m_color[1]), &(m_color[2]), &(m_color[3])) != 4)
		return FALSE;

	return TRUE;
}
