//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.

// Much of the code below is based on the 
// "Marching Cubes Example Program 
// by Cory Bloyd (corysama@yahoo.com)
//
// A simple, portable and complete implementation of the Marching Cubes
// and Marching Tetrahedrons algorithms in a single source file.
// There are many ways that this code could be made faster, but the 
// intent is for the code to be easy to understand.
//
// For a description of the algorithm go to
// http://astronomy.swin.edu.au/pbourke/modelling/polygonise/
//
// This code is public domain."
//

#include "stdafx.h"
#include <math.h>
#include "QmolView.h"
#include "MainFrm.h"

// ******* Atom box parameters *******
// These values obtained by finding minimum box length that didn't cause
// surface artifacts. They still need some fine tuning
//#define	BOX_LENGTH		1.9f
//#define	MAX_NUM_BOX		100.0f
//#define	MAX_LENGTH		(BOX_LENGTH*MAX_NUM_BOX)

#define	VERTEX_BOX_LENGTH		0.5f
#define	VERTEX_MAX_NUM_BOX		200.0f
#define	VERTEX_MAX_LENGTH		(VERTEX_BOX_LENGTH*VERTEX_MAX_NUM_BOX)

#define	MAX_MESH_SIZE	1.75f
#define	MIN_MESH_SIZE	0.1f

// Which parameterization of the molecular surface should we use?
#define	USE_R6		// A "softer" surface
//#define	USE_R8	// A "harder" surface

//a2fVertexOffset lists the positions, relative to vertex0, of each of the 8 vertices of a cube
static const GLfloat a2fVertexOffset[8][3] =
{
        {0.0, 0.0, 0.0},{1.0, 0.0, 0.0},{1.0, 1.0, 0.0},{0.0, 1.0, 0.0},
        {0.0, 0.0, 1.0},{1.0, 0.0, 1.0},{1.0, 1.0, 1.0},{0.0, 1.0, 1.0}
};

//a2iEdgeConnection lists the index of the endpoint vertices for each of the 12 edges of the cube
static const GLint a2iEdgeConnection[12][2] = 
{
        {0,1}, {1,2}, {2,3}, {3,0},
        {4,5}, {5,6}, {6,7}, {7,4},
        {0,4}, {1,5}, {2,6}, {3,7}
};

//a2fEdgeDirection lists the direction vector (vertex1-vertex0) for each edge in the cube
static const GLfloat a2fEdgeDirection[12][3] =
{
        {1.0, 0.0, 0.0},{0.0, 1.0, 0.0},{-1.0, 0.0, 0.0},{0.0, -1.0, 0.0},
        {1.0, 0.0, 0.0},{0.0, 1.0, 0.0},{-1.0, 0.0, 0.0},{0.0, -1.0, 0.0},
        {0.0, 0.0, 1.0},{0.0, 0.0, 1.0},{ 0.0, 0.0, 1.0},{0.0,  0.0, 1.0}
};

//  For each of the possible vertex states listed in aiCubeEdgeFlags there is a specific triangulation
//  of the edge intersection points.  a2iTriangleConnectionTable lists all of them in the form of
//  0-5 edge triples with the list terminated by the invalid value -1.
//  For example: a2iTriangleConnectionTable[3] list the 2 triangles formed when corner[0] 
//  and corner[1] are inside of the surface, but the rest of the cube is not.
//
//  I found this table in an example program someone wrote long ago.  It was probably generated by hand

GLint a2iTriangleConnectionTable[256][16] =  
{
        {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 8, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 1, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 8, 3, 9, 8, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 2, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 8, 3, 1, 2, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {9, 2, 10, 0, 2, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {2, 8, 3, 2, 10, 8, 10, 9, 8, -1, -1, -1, -1, -1, -1, -1},
        {3, 11, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 11, 2, 8, 11, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 9, 0, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 11, 2, 1, 9, 11, 9, 8, 11, -1, -1, -1, -1, -1, -1, -1},
        {3, 10, 1, 11, 10, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 10, 1, 0, 8, 10, 8, 11, 10, -1, -1, -1, -1, -1, -1, -1},
        {3, 9, 0, 3, 11, 9, 11, 10, 9, -1, -1, -1, -1, -1, -1, -1},
        {9, 8, 10, 10, 8, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {4, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {4, 3, 0, 7, 3, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 1, 9, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {4, 1, 9, 4, 7, 1, 7, 3, 1, -1, -1, -1, -1, -1, -1, -1},
        {1, 2, 10, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {3, 4, 7, 3, 0, 4, 1, 2, 10, -1, -1, -1, -1, -1, -1, -1},
        {9, 2, 10, 9, 0, 2, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1},
        {2, 10, 9, 2, 9, 7, 2, 7, 3, 7, 9, 4, -1, -1, -1, -1},
        {8, 4, 7, 3, 11, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {11, 4, 7, 11, 2, 4, 2, 0, 4, -1, -1, -1, -1, -1, -1, -1},
        {9, 0, 1, 8, 4, 7, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1},
        {4, 7, 11, 9, 4, 11, 9, 11, 2, 9, 2, 1, -1, -1, -1, -1},
        {3, 10, 1, 3, 11, 10, 7, 8, 4, -1, -1, -1, -1, -1, -1, -1},
        {1, 11, 10, 1, 4, 11, 1, 0, 4, 7, 11, 4, -1, -1, -1, -1},
        {4, 7, 8, 9, 0, 11, 9, 11, 10, 11, 0, 3, -1, -1, -1, -1},
        {4, 7, 11, 4, 11, 9, 9, 11, 10, -1, -1, -1, -1, -1, -1, -1},
        {9, 5, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {9, 5, 4, 0, 8, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 5, 4, 1, 5, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {8, 5, 4, 8, 3, 5, 3, 1, 5, -1, -1, -1, -1, -1, -1, -1},
        {1, 2, 10, 9, 5, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {3, 0, 8, 1, 2, 10, 4, 9, 5, -1, -1, -1, -1, -1, -1, -1},
        {5, 2, 10, 5, 4, 2, 4, 0, 2, -1, -1, -1, -1, -1, -1, -1},
        {2, 10, 5, 3, 2, 5, 3, 5, 4, 3, 4, 8, -1, -1, -1, -1},
        {9, 5, 4, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 11, 2, 0, 8, 11, 4, 9, 5, -1, -1, -1, -1, -1, -1, -1},
        {0, 5, 4, 0, 1, 5, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1},
        {2, 1, 5, 2, 5, 8, 2, 8, 11, 4, 8, 5, -1, -1, -1, -1},
        {10, 3, 11, 10, 1, 3, 9, 5, 4, -1, -1, -1, -1, -1, -1, -1},
        {4, 9, 5, 0, 8, 1, 8, 10, 1, 8, 11, 10, -1, -1, -1, -1},
        {5, 4, 0, 5, 0, 11, 5, 11, 10, 11, 0, 3, -1, -1, -1, -1},
        {5, 4, 8, 5, 8, 10, 10, 8, 11, -1, -1, -1, -1, -1, -1, -1},
        {9, 7, 8, 5, 7, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {9, 3, 0, 9, 5, 3, 5, 7, 3, -1, -1, -1, -1, -1, -1, -1},
        {0, 7, 8, 0, 1, 7, 1, 5, 7, -1, -1, -1, -1, -1, -1, -1},
        {1, 5, 3, 3, 5, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {9, 7, 8, 9, 5, 7, 10, 1, 2, -1, -1, -1, -1, -1, -1, -1},
        {10, 1, 2, 9, 5, 0, 5, 3, 0, 5, 7, 3, -1, -1, -1, -1},
        {8, 0, 2, 8, 2, 5, 8, 5, 7, 10, 5, 2, -1, -1, -1, -1},
        {2, 10, 5, 2, 5, 3, 3, 5, 7, -1, -1, -1, -1, -1, -1, -1},
        {7, 9, 5, 7, 8, 9, 3, 11, 2, -1, -1, -1, -1, -1, -1, -1},
        {9, 5, 7, 9, 7, 2, 9, 2, 0, 2, 7, 11, -1, -1, -1, -1},
        {2, 3, 11, 0, 1, 8, 1, 7, 8, 1, 5, 7, -1, -1, -1, -1},
        {11, 2, 1, 11, 1, 7, 7, 1, 5, -1, -1, -1, -1, -1, -1, -1},
        {9, 5, 8, 8, 5, 7, 10, 1, 3, 10, 3, 11, -1, -1, -1, -1},
        {5, 7, 0, 5, 0, 9, 7, 11, 0, 1, 0, 10, 11, 10, 0, -1},
        {11, 10, 0, 11, 0, 3, 10, 5, 0, 8, 0, 7, 5, 7, 0, -1},
        {11, 10, 5, 7, 11, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {10, 6, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 8, 3, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {9, 0, 1, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 8, 3, 1, 9, 8, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1},
        {1, 6, 5, 2, 6, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 6, 5, 1, 2, 6, 3, 0, 8, -1, -1, -1, -1, -1, -1, -1},
        {9, 6, 5, 9, 0, 6, 0, 2, 6, -1, -1, -1, -1, -1, -1, -1},
        {5, 9, 8, 5, 8, 2, 5, 2, 6, 3, 2, 8, -1, -1, -1, -1},
        {2, 3, 11, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {11, 0, 8, 11, 2, 0, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1},
        {0, 1, 9, 2, 3, 11, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1},
        {5, 10, 6, 1, 9, 2, 9, 11, 2, 9, 8, 11, -1, -1, -1, -1},
        {6, 3, 11, 6, 5, 3, 5, 1, 3, -1, -1, -1, -1, -1, -1, -1},
        {0, 8, 11, 0, 11, 5, 0, 5, 1, 5, 11, 6, -1, -1, -1, -1},
        {3, 11, 6, 0, 3, 6, 0, 6, 5, 0, 5, 9, -1, -1, -1, -1},
        {6, 5, 9, 6, 9, 11, 11, 9, 8, -1, -1, -1, -1, -1, -1, -1},
        {5, 10, 6, 4, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {4, 3, 0, 4, 7, 3, 6, 5, 10, -1, -1, -1, -1, -1, -1, -1},
        {1, 9, 0, 5, 10, 6, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1},
        {10, 6, 5, 1, 9, 7, 1, 7, 3, 7, 9, 4, -1, -1, -1, -1},
        {6, 1, 2, 6, 5, 1, 4, 7, 8, -1, -1, -1, -1, -1, -1, -1},
        {1, 2, 5, 5, 2, 6, 3, 0, 4, 3, 4, 7, -1, -1, -1, -1},
        {8, 4, 7, 9, 0, 5, 0, 6, 5, 0, 2, 6, -1, -1, -1, -1},
        {7, 3, 9, 7, 9, 4, 3, 2, 9, 5, 9, 6, 2, 6, 9, -1},
        {3, 11, 2, 7, 8, 4, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1},
        {5, 10, 6, 4, 7, 2, 4, 2, 0, 2, 7, 11, -1, -1, -1, -1},
        {0, 1, 9, 4, 7, 8, 2, 3, 11, 5, 10, 6, -1, -1, -1, -1},
        {9, 2, 1, 9, 11, 2, 9, 4, 11, 7, 11, 4, 5, 10, 6, -1},
        {8, 4, 7, 3, 11, 5, 3, 5, 1, 5, 11, 6, -1, -1, -1, -1},
        {5, 1, 11, 5, 11, 6, 1, 0, 11, 7, 11, 4, 0, 4, 11, -1},
        {0, 5, 9, 0, 6, 5, 0, 3, 6, 11, 6, 3, 8, 4, 7, -1},
        {6, 5, 9, 6, 9, 11, 4, 7, 9, 7, 11, 9, -1, -1, -1, -1},
        {10, 4, 9, 6, 4, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {4, 10, 6, 4, 9, 10, 0, 8, 3, -1, -1, -1, -1, -1, -1, -1},
        {10, 0, 1, 10, 6, 0, 6, 4, 0, -1, -1, -1, -1, -1, -1, -1},
        {8, 3, 1, 8, 1, 6, 8, 6, 4, 6, 1, 10, -1, -1, -1, -1},
        {1, 4, 9, 1, 2, 4, 2, 6, 4, -1, -1, -1, -1, -1, -1, -1},
        {3, 0, 8, 1, 2, 9, 2, 4, 9, 2, 6, 4, -1, -1, -1, -1},
        {0, 2, 4, 4, 2, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {8, 3, 2, 8, 2, 4, 4, 2, 6, -1, -1, -1, -1, -1, -1, -1},
        {10, 4, 9, 10, 6, 4, 11, 2, 3, -1, -1, -1, -1, -1, -1, -1},
        {0, 8, 2, 2, 8, 11, 4, 9, 10, 4, 10, 6, -1, -1, -1, -1},
        {3, 11, 2, 0, 1, 6, 0, 6, 4, 6, 1, 10, -1, -1, -1, -1},
        {6, 4, 1, 6, 1, 10, 4, 8, 1, 2, 1, 11, 8, 11, 1, -1},
        {9, 6, 4, 9, 3, 6, 9, 1, 3, 11, 6, 3, -1, -1, -1, -1},
        {8, 11, 1, 8, 1, 0, 11, 6, 1, 9, 1, 4, 6, 4, 1, -1},
        {3, 11, 6, 3, 6, 0, 0, 6, 4, -1, -1, -1, -1, -1, -1, -1},
        {6, 4, 8, 11, 6, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {7, 10, 6, 7, 8, 10, 8, 9, 10, -1, -1, -1, -1, -1, -1, -1},
        {0, 7, 3, 0, 10, 7, 0, 9, 10, 6, 7, 10, -1, -1, -1, -1},
        {10, 6, 7, 1, 10, 7, 1, 7, 8, 1, 8, 0, -1, -1, -1, -1},
        {10, 6, 7, 10, 7, 1, 1, 7, 3, -1, -1, -1, -1, -1, -1, -1},
        {1, 2, 6, 1, 6, 8, 1, 8, 9, 8, 6, 7, -1, -1, -1, -1},
        {2, 6, 9, 2, 9, 1, 6, 7, 9, 0, 9, 3, 7, 3, 9, -1},
        {7, 8, 0, 7, 0, 6, 6, 0, 2, -1, -1, -1, -1, -1, -1, -1},
        {7, 3, 2, 6, 7, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {2, 3, 11, 10, 6, 8, 10, 8, 9, 8, 6, 7, -1, -1, -1, -1},
        {2, 0, 7, 2, 7, 11, 0, 9, 7, 6, 7, 10, 9, 10, 7, -1},
        {1, 8, 0, 1, 7, 8, 1, 10, 7, 6, 7, 10, 2, 3, 11, -1},
        {11, 2, 1, 11, 1, 7, 10, 6, 1, 6, 7, 1, -1, -1, -1, -1},
        {8, 9, 6, 8, 6, 7, 9, 1, 6, 11, 6, 3, 1, 3, 6, -1},
        {0, 9, 1, 11, 6, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {7, 8, 0, 7, 0, 6, 3, 11, 0, 11, 6, 0, -1, -1, -1, -1},
        {7, 11, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {7, 6, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {3, 0, 8, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 1, 9, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {8, 1, 9, 8, 3, 1, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1},
        {10, 1, 2, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 2, 10, 3, 0, 8, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1},
        {2, 9, 0, 2, 10, 9, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1},
        {6, 11, 7, 2, 10, 3, 10, 8, 3, 10, 9, 8, -1, -1, -1, -1},
        {7, 2, 3, 6, 2, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {7, 0, 8, 7, 6, 0, 6, 2, 0, -1, -1, -1, -1, -1, -1, -1},
        {2, 7, 6, 2, 3, 7, 0, 1, 9, -1, -1, -1, -1, -1, -1, -1},
        {1, 6, 2, 1, 8, 6, 1, 9, 8, 8, 7, 6, -1, -1, -1, -1},
        {10, 7, 6, 10, 1, 7, 1, 3, 7, -1, -1, -1, -1, -1, -1, -1},
        {10, 7, 6, 1, 7, 10, 1, 8, 7, 1, 0, 8, -1, -1, -1, -1},
        {0, 3, 7, 0, 7, 10, 0, 10, 9, 6, 10, 7, -1, -1, -1, -1},
        {7, 6, 10, 7, 10, 8, 8, 10, 9, -1, -1, -1, -1, -1, -1, -1},
        {6, 8, 4, 11, 8, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {3, 6, 11, 3, 0, 6, 0, 4, 6, -1, -1, -1, -1, -1, -1, -1},
        {8, 6, 11, 8, 4, 6, 9, 0, 1, -1, -1, -1, -1, -1, -1, -1},
        {9, 4, 6, 9, 6, 3, 9, 3, 1, 11, 3, 6, -1, -1, -1, -1},
        {6, 8, 4, 6, 11, 8, 2, 10, 1, -1, -1, -1, -1, -1, -1, -1},
        {1, 2, 10, 3, 0, 11, 0, 6, 11, 0, 4, 6, -1, -1, -1, -1},
        {4, 11, 8, 4, 6, 11, 0, 2, 9, 2, 10, 9, -1, -1, -1, -1},
        {10, 9, 3, 10, 3, 2, 9, 4, 3, 11, 3, 6, 4, 6, 3, -1},
        {8, 2, 3, 8, 4, 2, 4, 6, 2, -1, -1, -1, -1, -1, -1, -1},
        {0, 4, 2, 4, 6, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 9, 0, 2, 3, 4, 2, 4, 6, 4, 3, 8, -1, -1, -1, -1},
        {1, 9, 4, 1, 4, 2, 2, 4, 6, -1, -1, -1, -1, -1, -1, -1},
        {8, 1, 3, 8, 6, 1, 8, 4, 6, 6, 10, 1, -1, -1, -1, -1},
        {10, 1, 0, 10, 0, 6, 6, 0, 4, -1, -1, -1, -1, -1, -1, -1},
        {4, 6, 3, 4, 3, 8, 6, 10, 3, 0, 3, 9, 10, 9, 3, -1},
        {10, 9, 4, 6, 10, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {4, 9, 5, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 8, 3, 4, 9, 5, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1},
        {5, 0, 1, 5, 4, 0, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1},
        {11, 7, 6, 8, 3, 4, 3, 5, 4, 3, 1, 5, -1, -1, -1, -1},
        {9, 5, 4, 10, 1, 2, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1},
        {6, 11, 7, 1, 2, 10, 0, 8, 3, 4, 9, 5, -1, -1, -1, -1},
        {7, 6, 11, 5, 4, 10, 4, 2, 10, 4, 0, 2, -1, -1, -1, -1},
        {3, 4, 8, 3, 5, 4, 3, 2, 5, 10, 5, 2, 11, 7, 6, -1},
        {7, 2, 3, 7, 6, 2, 5, 4, 9, -1, -1, -1, -1, -1, -1, -1},
        {9, 5, 4, 0, 8, 6, 0, 6, 2, 6, 8, 7, -1, -1, -1, -1},
        {3, 6, 2, 3, 7, 6, 1, 5, 0, 5, 4, 0, -1, -1, -1, -1},
        {6, 2, 8, 6, 8, 7, 2, 1, 8, 4, 8, 5, 1, 5, 8, -1},
        {9, 5, 4, 10, 1, 6, 1, 7, 6, 1, 3, 7, -1, -1, -1, -1},
        {1, 6, 10, 1, 7, 6, 1, 0, 7, 8, 7, 0, 9, 5, 4, -1},
        {4, 0, 10, 4, 10, 5, 0, 3, 10, 6, 10, 7, 3, 7, 10, -1},
        {7, 6, 10, 7, 10, 8, 5, 4, 10, 4, 8, 10, -1, -1, -1, -1},
        {6, 9, 5, 6, 11, 9, 11, 8, 9, -1, -1, -1, -1, -1, -1, -1},
        {3, 6, 11, 0, 6, 3, 0, 5, 6, 0, 9, 5, -1, -1, -1, -1},
        {0, 11, 8, 0, 5, 11, 0, 1, 5, 5, 6, 11, -1, -1, -1, -1},
        {6, 11, 3, 6, 3, 5, 5, 3, 1, -1, -1, -1, -1, -1, -1, -1},
        {1, 2, 10, 9, 5, 11, 9, 11, 8, 11, 5, 6, -1, -1, -1, -1},
        {0, 11, 3, 0, 6, 11, 0, 9, 6, 5, 6, 9, 1, 2, 10, -1},
        {11, 8, 5, 11, 5, 6, 8, 0, 5, 10, 5, 2, 0, 2, 5, -1},
        {6, 11, 3, 6, 3, 5, 2, 10, 3, 10, 5, 3, -1, -1, -1, -1},
        {5, 8, 9, 5, 2, 8, 5, 6, 2, 3, 8, 2, -1, -1, -1, -1},
        {9, 5, 6, 9, 6, 0, 0, 6, 2, -1, -1, -1, -1, -1, -1, -1},
        {1, 5, 8, 1, 8, 0, 5, 6, 8, 3, 8, 2, 6, 2, 8, -1},
        {1, 5, 6, 2, 1, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 3, 6, 1, 6, 10, 3, 8, 6, 5, 6, 9, 8, 9, 6, -1},
        {10, 1, 0, 10, 0, 6, 9, 5, 0, 5, 6, 0, -1, -1, -1, -1},
        {0, 3, 8, 5, 6, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {10, 5, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {11, 5, 10, 7, 5, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {11, 5, 10, 11, 7, 5, 8, 3, 0, -1, -1, -1, -1, -1, -1, -1},
        {5, 11, 7, 5, 10, 11, 1, 9, 0, -1, -1, -1, -1, -1, -1, -1},
        {10, 7, 5, 10, 11, 7, 9, 8, 1, 8, 3, 1, -1, -1, -1, -1},
        {11, 1, 2, 11, 7, 1, 7, 5, 1, -1, -1, -1, -1, -1, -1, -1},
        {0, 8, 3, 1, 2, 7, 1, 7, 5, 7, 2, 11, -1, -1, -1, -1},
        {9, 7, 5, 9, 2, 7, 9, 0, 2, 2, 11, 7, -1, -1, -1, -1},
        {7, 5, 2, 7, 2, 11, 5, 9, 2, 3, 2, 8, 9, 8, 2, -1},
        {2, 5, 10, 2, 3, 5, 3, 7, 5, -1, -1, -1, -1, -1, -1, -1},
        {8, 2, 0, 8, 5, 2, 8, 7, 5, 10, 2, 5, -1, -1, -1, -1},
        {9, 0, 1, 5, 10, 3, 5, 3, 7, 3, 10, 2, -1, -1, -1, -1},
        {9, 8, 2, 9, 2, 1, 8, 7, 2, 10, 2, 5, 7, 5, 2, -1},
        {1, 3, 5, 3, 7, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 8, 7, 0, 7, 1, 1, 7, 5, -1, -1, -1, -1, -1, -1, -1},
        {9, 0, 3, 9, 3, 5, 5, 3, 7, -1, -1, -1, -1, -1, -1, -1},
        {9, 8, 7, 5, 9, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {5, 8, 4, 5, 10, 8, 10, 11, 8, -1, -1, -1, -1, -1, -1, -1},
        {5, 0, 4, 5, 11, 0, 5, 10, 11, 11, 3, 0, -1, -1, -1, -1},
        {0, 1, 9, 8, 4, 10, 8, 10, 11, 10, 4, 5, -1, -1, -1, -1},
        {10, 11, 4, 10, 4, 5, 11, 3, 4, 9, 4, 1, 3, 1, 4, -1},
        {2, 5, 1, 2, 8, 5, 2, 11, 8, 4, 5, 8, -1, -1, -1, -1},
        {0, 4, 11, 0, 11, 3, 4, 5, 11, 2, 11, 1, 5, 1, 11, -1},
        {0, 2, 5, 0, 5, 9, 2, 11, 5, 4, 5, 8, 11, 8, 5, -1},
        {9, 4, 5, 2, 11, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {2, 5, 10, 3, 5, 2, 3, 4, 5, 3, 8, 4, -1, -1, -1, -1},
        {5, 10, 2, 5, 2, 4, 4, 2, 0, -1, -1, -1, -1, -1, -1, -1},
        {3, 10, 2, 3, 5, 10, 3, 8, 5, 4, 5, 8, 0, 1, 9, -1},
        {5, 10, 2, 5, 2, 4, 1, 9, 2, 9, 4, 2, -1, -1, -1, -1},
        {8, 4, 5, 8, 5, 3, 3, 5, 1, -1, -1, -1, -1, -1, -1, -1},
        {0, 4, 5, 1, 0, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {8, 4, 5, 8, 5, 3, 9, 0, 5, 0, 3, 5, -1, -1, -1, -1},
        {9, 4, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {4, 11, 7, 4, 9, 11, 9, 10, 11, -1, -1, -1, -1, -1, -1, -1},
        {0, 8, 3, 4, 9, 7, 9, 11, 7, 9, 10, 11, -1, -1, -1, -1},
        {1, 10, 11, 1, 11, 4, 1, 4, 0, 7, 4, 11, -1, -1, -1, -1},
        {3, 1, 4, 3, 4, 8, 1, 10, 4, 7, 4, 11, 10, 11, 4, -1},
        {4, 11, 7, 9, 11, 4, 9, 2, 11, 9, 1, 2, -1, -1, -1, -1},
        {9, 7, 4, 9, 11, 7, 9, 1, 11, 2, 11, 1, 0, 8, 3, -1},
        {11, 7, 4, 11, 4, 2, 2, 4, 0, -1, -1, -1, -1, -1, -1, -1},
        {11, 7, 4, 11, 4, 2, 8, 3, 4, 3, 2, 4, -1, -1, -1, -1},
        {2, 9, 10, 2, 7, 9, 2, 3, 7, 7, 4, 9, -1, -1, -1, -1},
        {9, 10, 7, 9, 7, 4, 10, 2, 7, 8, 7, 0, 2, 0, 7, -1},
        {3, 7, 10, 3, 10, 2, 7, 4, 10, 1, 10, 0, 4, 0, 10, -1},
        {1, 10, 2, 8, 7, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {4, 9, 1, 4, 1, 7, 7, 1, 3, -1, -1, -1, -1, -1, -1, -1},
        {4, 9, 1, 4, 1, 7, 0, 8, 1, 8, 7, 1, -1, -1, -1, -1},
        {4, 0, 3, 7, 4, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {4, 8, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {9, 10, 8, 10, 11, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {3, 0, 9, 3, 9, 11, 11, 9, 10, -1, -1, -1, -1, -1, -1, -1},
        {0, 1, 10, 0, 10, 8, 8, 10, 11, -1, -1, -1, -1, -1, -1, -1},
        {3, 1, 10, 11, 3, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 2, 11, 1, 11, 9, 9, 11, 8, -1, -1, -1, -1, -1, -1, -1},
        {3, 0, 9, 3, 9, 11, 1, 2, 9, 2, 11, 9, -1, -1, -1, -1},
        {0, 2, 11, 8, 0, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {3, 2, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {2, 3, 8, 2, 8, 10, 10, 8, 9, -1, -1, -1, -1, -1, -1, -1},
        {9, 10, 2, 0, 9, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {2, 3, 8, 2, 8, 10, 0, 1, 8, 1, 10, 8, -1, -1, -1, -1},
        {1, 10, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {1, 3, 8, 9, 1, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 9, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {0, 3, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
        {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}
};

// For any edge, if one vertex is inside of the surface and the other is 
// outside of the surface then the edge intersects the surface
// For each of the 8 vertices of the cube can be two possible states : 
// either inside or outside of the surface
// For any cube the are 2^8=256 possible sets of vertex states
// This table lists the edges intersected by the surface for all 256 
// possible vertex states. There are 12 edges.  For each entry in the table, 
// if edge #n is intersected, then bit #n is set to 1

GLint aiCubeEdgeFlags[256]=
{
        0x000, 0x109, 0x203, 0x30a, 0x406, 0x50f, 0x605, 0x70c, 0x80c, 0x905, 0xa0f, 0xb06, 0xc0a, 0xd03, 0xe09, 0xf00, 
        0x190, 0x099, 0x393, 0x29a, 0x596, 0x49f, 0x795, 0x69c, 0x99c, 0x895, 0xb9f, 0xa96, 0xd9a, 0xc93, 0xf99, 0xe90, 
        0x230, 0x339, 0x033, 0x13a, 0x636, 0x73f, 0x435, 0x53c, 0xa3c, 0xb35, 0x83f, 0x936, 0xe3a, 0xf33, 0xc39, 0xd30, 
        0x3a0, 0x2a9, 0x1a3, 0x0aa, 0x7a6, 0x6af, 0x5a5, 0x4ac, 0xbac, 0xaa5, 0x9af, 0x8a6, 0xfaa, 0xea3, 0xda9, 0xca0, 
        0x460, 0x569, 0x663, 0x76a, 0x066, 0x16f, 0x265, 0x36c, 0xc6c, 0xd65, 0xe6f, 0xf66, 0x86a, 0x963, 0xa69, 0xb60, 
        0x5f0, 0x4f9, 0x7f3, 0x6fa, 0x1f6, 0x0ff, 0x3f5, 0x2fc, 0xdfc, 0xcf5, 0xfff, 0xef6, 0x9fa, 0x8f3, 0xbf9, 0xaf0, 
        0x650, 0x759, 0x453, 0x55a, 0x256, 0x35f, 0x055, 0x15c, 0xe5c, 0xf55, 0xc5f, 0xd56, 0xa5a, 0xb53, 0x859, 0x950, 
        0x7c0, 0x6c9, 0x5c3, 0x4ca, 0x3c6, 0x2cf, 0x1c5, 0x0cc, 0xfcc, 0xec5, 0xdcf, 0xcc6, 0xbca, 0xac3, 0x9c9, 0x8c0, 
        0x8c0, 0x9c9, 0xac3, 0xbca, 0xcc6, 0xdcf, 0xec5, 0xfcc, 0x0cc, 0x1c5, 0x2cf, 0x3c6, 0x4ca, 0x5c3, 0x6c9, 0x7c0, 
        0x950, 0x859, 0xb53, 0xa5a, 0xd56, 0xc5f, 0xf55, 0xe5c, 0x15c, 0x055, 0x35f, 0x256, 0x55a, 0x453, 0x759, 0x650, 
        0xaf0, 0xbf9, 0x8f3, 0x9fa, 0xef6, 0xfff, 0xcf5, 0xdfc, 0x2fc, 0x3f5, 0x0ff, 0x1f6, 0x6fa, 0x7f3, 0x4f9, 0x5f0, 
        0xb60, 0xa69, 0x963, 0x86a, 0xf66, 0xe6f, 0xd65, 0xc6c, 0x36c, 0x265, 0x16f, 0x066, 0x76a, 0x663, 0x569, 0x460, 
        0xca0, 0xda9, 0xea3, 0xfaa, 0x8a6, 0x9af, 0xaa5, 0xbac, 0x4ac, 0x5a5, 0x6af, 0x7a6, 0x0aa, 0x1a3, 0x2a9, 0x3a0, 
        0xd30, 0xc39, 0xf33, 0xe3a, 0x936, 0x83f, 0xb35, 0xa3c, 0x53c, 0x435, 0x73f, 0x636, 0x13a, 0x033, 0x339, 0x230, 
        0xe90, 0xf99, 0xc93, 0xd9a, 0xa96, 0xb9f, 0x895, 0x99c, 0x69c, 0x795, 0x49f, 0x596, 0x29a, 0x393, 0x099, 0x190, 
        0xf00, 0xe09, 0xd03, 0xc0a, 0xb06, 0xa0f, 0x905, 0x80c, 0x70c, 0x605, 0x50f, 0x406, 0x30a, 0x203, 0x109, 0x000
};

MolSurface::MolSurface()
{
	atomBox = NULL;
	validSurface = FALSE;

	// The raidus of the probe sphere is used in the heuristic 
	// solvent accessible surface calculation.
	probeSphere = 0.0f;

	// Try different iso-surface contour levels
	fTargetValue = 0.475f;
	
	surfaceType = HEURISTIC;
	renderStyle = SOLID;

	targetValueCutOff = 2.0f*fTargetValue;

	// When a solvent probe sphere is used, 
	// the boxLength must be increased
	boxLength = 3.0f + probeSphere*probeSphere; 

	maxNumBox = 200;
	maxLength = boxLength*maxNumBox;

	SetSurfQuality(70);

	include_solvent_atoms = FALSE;

	color_by_atom = TRUE;
	
	// Make the color-by-atom surface opaque by default
	transparency = 255;

	// Make the usered defined surface color grey
	color[0] = color[1] = color[2] = 128;
	color[3] = 255;
}

MolSurface::~MolSurface()
{
	DeleteAtomBox();
}

BOOL MolSurface::BuildSurface(array<atom> *m_ptrAtoms, 
		list <array <float> > *m_ptrTraj, int m_currentStep)
{
	GLfloat x, y, z;
	int i, j, k;

	// Which structure are we generating a surface for?
	traj_index = m_currentStep - 1;

	// Store a local copy of the molecule pointer
	ptrAtoms = m_ptrAtoms;
	ptrTraj = m_ptrTraj;

	BuildAtomBox();

	GLfloat xmin = molMin[X], 
			ymin = molMin[Y],
			zmin = molMin[Z];

	// Build the cube box
	cubeSize[X] = (int)((molMax[X] - xmin)/meshSize) + 1;
	cubeSize[Y] = (int)((molMax[Y] - ymin)/meshSize) + 1;
	cubeSize[Z] = (int)((molMax[Z] - zmin)/meshSize) + 1;
	
	cubeXY = cubeSize[X]*cubeSize[Y];

	cubeBox.Alloc(cubeSize[X]*cubeSize[Y]*cubeSize[Z]);

	// Keep the user informed
	CMainFrame *pFrame = (CMainFrame*)AfxGetApp()->m_pMainWnd;
	const int step_every = 10;
	int max_count = 2*cubeSize[Z] - 1;
	int progress_counter = 0;
	if(pFrame){
		// 100% for the marching cubes and 100% for the vertex and triangle processing
		pFrame->InitProgress(200);
	}

	// Clean up the triangle list
	while(!triangleList.IsEmpty()){
		triangleList.Pop();
	}

	for(z = zmin, i = 0; i < cubeSize[Z]; z += meshSize, i++){
		for(y = ymin, j = 0; j < cubeSize[Y]; y += meshSize, j++){
			for(x = xmin, k = 0; k < cubeSize[X]; x += meshSize, k++){
				cubeBox[i*cubeXY + j*cubeSize[X] + k] = ScalarField(x, y, z);
			}
		}

		progress_counter ++;
		if((progress_counter%step_every == 0) && pFrame){
			// Advance the progress bar
			pFrame->SetProgress((100*progress_counter)/max_count);
		}
	}

	// Don't check the outer edges of the "box" of scalar field values.
	xmin += meshSize;
	ymin += meshSize;
	zmin += meshSize;

	int max_x_index = cubeSize[X] - 1;
	int max_y_index = cubeSize[Y] - 1;
	int max_z_index = cubeSize[Z] - 1;

	for(z = zmin, i = 1; i < max_z_index ; z += meshSize, i++){
		for(y = ymin, j = 1; j < max_y_index; y += meshSize, j++){
			for(x = xmin, k = 1; k < max_x_index; x += meshSize, k++){
				vMarchCube(x, y, z, k, j, i, meshSize);
			}
		}

		progress_counter ++;
		if((progress_counter%step_every == 0) && pFrame){
			// Advance the progress bar
			pFrame->SetProgress((100*progress_counter)/max_count);
		}
	}

	//  ** We now have a list of triangle verticies. **

	// Convert this list into a non-redundant array of verticies 
	// (and remove "runt" verticies). After we compute this list, 
	// we will compute the color index of each vertex.
	BuildVertexArray();

	DeleteAtomBox();

	cubeBox.Alloc(0);	

	if(pFrame){
		pFrame->CloseProgress();
	}

	// Set the valid surface flag
	validSurface = TRUE;

	return TRUE;
}

void MolSurface::RenderSurface(array<atom> &m_atoms)
{
	if(validSurface){

		int num_tri_vert = indexArray.Length();
		int i, last_color_index = -1;
		int index;
		int colorIndex;

		glVertexPointer(3, GL_FLOAT, 0, vertexArray.Pointer());
		glNormalPointer(GL_FLOAT, 0, normalArray.Pointer());

		glEnableClientState(GL_VERTEX_ARRAY);
		glEnableClientState(GL_NORMAL_ARRAY);
		
		switch(renderStyle){
			case MESH:
				// Only render polygone outlines in mesh mode
				glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
				break;
			case POINT:
				// Only render polygone outlines in mesh mode
				glPolygonMode(GL_FRONT_AND_BACK, GL_POINT);
				break;
		};

		if(color_by_atom){

			// Are we using a transparent surface?
			if(transparency < 255){
				// Implement surface transparency by setting the alpha component of
				// the atomic color vectors.
				glBegin(GL_TRIANGLES );

				for(i = 0;i < num_tri_vert;i++){
					index = indexArray[i];

					colorIndex = colorArray[index];
					
					if(last_color_index != colorIndex){
						GLubyte *tmpColor = m_atoms[colorIndex].rgb_color;
						GLubyte tmpAlpha = tmpColor[3];
						
						tmpColor[3] = transparency;

						glColor4ubv(tmpColor);
						
						tmpColor[3] = tmpAlpha;
						
						last_color_index = colorIndex;
					}
					
					glArrayElement(index);
				}
				
				glEnd();
			}
			else{ 
				// No surface transparency -- just use the atomic colors
				glBegin(GL_TRIANGLES );
				
				for(i = 0;i < num_tri_vert;i++){
					index = indexArray[i];

					colorIndex = colorArray[index];
					
					if(last_color_index != colorIndex){
						glColor4ubv(m_atoms[colorIndex].rgb_color);
						
						last_color_index = colorIndex;
					}
					
					glArrayElement(index);
				}
				
				glEnd();
			}
		}
		else{
			// Apply user defined color
			glColor4ubv(color);

			// Note that the static color allows use to use the much faster
			// glDrawElements function instead of the glArrayElement function.
			glDrawElements(GL_TRIANGLES, num_tri_vert, GL_UNSIGNED_INT, indexArray.Pointer());
		}

		glDisableClientState(GL_VERTEX_ARRAY);
		glDisableClientState(GL_NORMAL_ARRAY);

		if(renderStyle != SOLID){
			// Restore the solid fill for polygones
			glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		}
	}
}

//vMarchCube1 performs the Marching Cubes algorithm on a single cube
GLvoid MolSurface::vMarchCube(GLfloat fX, GLfloat fY, GLfloat fZ, 
							  int xIndex, int yIndex, int zIndex, GLfloat fScale)
{
	GLint iFlagIndex, iEdgeFlags;
	GLfloat fOffset;
	GLfloat afCubeValue[8];
	GLfloat asEdgeVertex[12][3];

	//Make a local copy of the values at the cube's corners
	afCubeValue[0] = cubeBox[zIndex*cubeXY + yIndex*cubeSize[X] + xIndex];
	afCubeValue[1] = cubeBox[zIndex*cubeXY + yIndex*cubeSize[X] + xIndex + 1];
	afCubeValue[2] = cubeBox[zIndex*cubeXY + (yIndex + 1)*cubeSize[X] + xIndex + 1];
	afCubeValue[3] = cubeBox[zIndex*cubeXY + (yIndex + 1)*cubeSize[X] + xIndex];
	afCubeValue[4] = cubeBox[(zIndex + 1)*cubeXY + yIndex*cubeSize[X] + xIndex];
	afCubeValue[5] = cubeBox[(zIndex + 1)*cubeXY + yIndex*cubeSize[X] + xIndex + 1];
	afCubeValue[6] = cubeBox[(zIndex + 1)*cubeXY + (yIndex + 1)*cubeSize[X] + xIndex + 1];
	afCubeValue[7] = cubeBox[(zIndex + 1)*cubeXY + (yIndex + 1)*cubeSize[X] + xIndex];

	//Find which vertices are inside of the surface and which are outside
	iFlagIndex = 0;

	if(afCubeValue[0] <= fTargetValue)
		// No need to use a bitwise operator for the first
		// bit.
		iFlagIndex = 1;

	if(afCubeValue[1] <= fTargetValue) 
		iFlagIndex |= 2;

	if(afCubeValue[2] <= fTargetValue) 
		iFlagIndex |= 4;

	if(afCubeValue[3] <= fTargetValue) 
		iFlagIndex |= 8;

	if(afCubeValue[4] <= fTargetValue) 
		iFlagIndex |= 16;

	if(afCubeValue[5] <= fTargetValue) 
		iFlagIndex |= 32;

	if(afCubeValue[6] <= fTargetValue) 
		iFlagIndex |= 64;

	if(afCubeValue[7] <= fTargetValue) 
		iFlagIndex |= 128;

	//Find which edges are intersected by the surface
	iEdgeFlags = aiCubeEdgeFlags[iFlagIndex];

	//If the cube is entirely inside or outside of the surface, then there will be no intersections
	if(iEdgeFlags == 0) {
		return;
	}

	// Find the point of intersection of the surface with each edge.
	// Then find the normal to the surface at those points
	// if there is an intersection on the given edge.
	if(iEdgeFlags & 1){
		fOffset = fGetOffset(afCubeValue[ a2iEdgeConnection[0][0] ], 
			afCubeValue[ a2iEdgeConnection[0][1] ], 
			fTargetValue);
		
		asEdgeVertex[0][X] = fX + (a2fVertexOffset[ a2iEdgeConnection[0][0] ][0]  +  
			fOffset * a2fEdgeDirection[0][0]) * fScale;
		asEdgeVertex[0][Y] = fY + (a2fVertexOffset[ a2iEdgeConnection[0][0] ][1]  +  
			fOffset * a2fEdgeDirection[0][1]) * fScale;
		asEdgeVertex[0][Z] = fZ + (a2fVertexOffset[ a2iEdgeConnection[0][0] ][2]  +  
			fOffset * a2fEdgeDirection[0][2]) * fScale;
	}

	if(iEdgeFlags & 2){
		fOffset = fGetOffset(afCubeValue[ a2iEdgeConnection[1][0] ], 
			afCubeValue[ a2iEdgeConnection[1][1] ], 
			fTargetValue);
		
		asEdgeVertex[1][X] = fX + (a2fVertexOffset[ a2iEdgeConnection[1][0] ][0]  +  
			fOffset * a2fEdgeDirection[1][0]) * fScale;
		asEdgeVertex[1][Y] = fY + (a2fVertexOffset[ a2iEdgeConnection[1][0] ][1]  +  
			fOffset * a2fEdgeDirection[1][1]) * fScale;
		asEdgeVertex[1][Z] = fZ + (a2fVertexOffset[ a2iEdgeConnection[1][0] ][2]  +  
			fOffset * a2fEdgeDirection[1][2]) * fScale;
	}

	if(iEdgeFlags & 4){
		fOffset = fGetOffset(afCubeValue[ a2iEdgeConnection[2][0] ], 
			afCubeValue[ a2iEdgeConnection[2][1] ], 
			fTargetValue);
		
		asEdgeVertex[2][X] = fX + (a2fVertexOffset[ a2iEdgeConnection[2][0] ][0]  +  
			fOffset * a2fEdgeDirection[2][0]) * fScale;
		asEdgeVertex[2][Y] = fY + (a2fVertexOffset[ a2iEdgeConnection[2][0] ][1]  +  
			fOffset * a2fEdgeDirection[2][1]) * fScale;
		asEdgeVertex[2][Z] = fZ + (a2fVertexOffset[ a2iEdgeConnection[2][0] ][2]  +  
			fOffset * a2fEdgeDirection[2][2]) * fScale;
	}

	if(iEdgeFlags & 8){
		fOffset = fGetOffset(afCubeValue[ a2iEdgeConnection[3][0] ], 
			afCubeValue[ a2iEdgeConnection[3][1] ], 
			fTargetValue);
		
		asEdgeVertex[3][X] = fX + (a2fVertexOffset[ a2iEdgeConnection[3][0] ][0]  +  
			fOffset * a2fEdgeDirection[3][0]) * fScale;
		asEdgeVertex[3][Y] = fY + (a2fVertexOffset[ a2iEdgeConnection[3][0] ][1]  +  
			fOffset * a2fEdgeDirection[3][1]) * fScale;
		asEdgeVertex[3][Z] = fZ + (a2fVertexOffset[ a2iEdgeConnection[3][0] ][2]  +  
			fOffset * a2fEdgeDirection[3][2]) * fScale;
	}

	if(iEdgeFlags & 16){
		fOffset = fGetOffset(afCubeValue[ a2iEdgeConnection[4][0] ], 
			afCubeValue[ a2iEdgeConnection[4][1] ], 
			fTargetValue);
		
		asEdgeVertex[4][X] = fX + (a2fVertexOffset[ a2iEdgeConnection[4][0] ][0]  +  
			fOffset * a2fEdgeDirection[4][0]) * fScale;
		asEdgeVertex[4][Y] = fY + (a2fVertexOffset[ a2iEdgeConnection[4][0] ][1]  +  
			fOffset * a2fEdgeDirection[4][1]) * fScale;
		asEdgeVertex[4][Z] = fZ + (a2fVertexOffset[ a2iEdgeConnection[4][0] ][2]  +  
			fOffset * a2fEdgeDirection[4][2]) * fScale;
	}

	if(iEdgeFlags & 32){
		fOffset = fGetOffset(afCubeValue[ a2iEdgeConnection[5][0] ], 
			afCubeValue[ a2iEdgeConnection[5][1] ], 
			fTargetValue);
		
		asEdgeVertex[5][X] = fX + (a2fVertexOffset[ a2iEdgeConnection[5][0] ][0]  +  
			fOffset * a2fEdgeDirection[5][0]) * fScale;
		asEdgeVertex[5][Y] = fY + (a2fVertexOffset[ a2iEdgeConnection[5][0] ][1]  +  
			fOffset * a2fEdgeDirection[5][1]) * fScale;
		asEdgeVertex[5][Z] = fZ + (a2fVertexOffset[ a2iEdgeConnection[5][0] ][2]  +  
			fOffset * a2fEdgeDirection[5][2]) * fScale;
	}

	if(iEdgeFlags & 64){
		fOffset = fGetOffset(afCubeValue[ a2iEdgeConnection[6][0] ], 
			afCubeValue[ a2iEdgeConnection[6][1] ], 
			fTargetValue);
		
		asEdgeVertex[6][X] = fX + (a2fVertexOffset[ a2iEdgeConnection[6][0] ][0]  +  
			fOffset * a2fEdgeDirection[6][0]) * fScale;
		asEdgeVertex[6][Y] = fY + (a2fVertexOffset[ a2iEdgeConnection[6][0] ][1]  +  
			fOffset * a2fEdgeDirection[6][1]) * fScale;
		asEdgeVertex[6][Z] = fZ + (a2fVertexOffset[ a2iEdgeConnection[6][0] ][2]  +  
			fOffset * a2fEdgeDirection[6][2]) * fScale;
	}

	if(iEdgeFlags & 128){
		fOffset = fGetOffset(afCubeValue[ a2iEdgeConnection[7][0] ], 
			afCubeValue[ a2iEdgeConnection[7][1] ], 
			fTargetValue);
		
		asEdgeVertex[7][X] = fX + (a2fVertexOffset[ a2iEdgeConnection[7][0] ][0]  +  
			fOffset * a2fEdgeDirection[7][0]) * fScale;
		asEdgeVertex[7][Y] = fY + (a2fVertexOffset[ a2iEdgeConnection[7][0] ][1]  +  
			fOffset * a2fEdgeDirection[7][1]) * fScale;
		asEdgeVertex[7][Z] = fZ + (a2fVertexOffset[ a2iEdgeConnection[7][0] ][2]  +  
			fOffset * a2fEdgeDirection[7][2]) * fScale;
	}

	if(iEdgeFlags & 256){
		fOffset = fGetOffset(afCubeValue[ a2iEdgeConnection[8][0] ], 
			afCubeValue[ a2iEdgeConnection[8][1] ], 
			fTargetValue);
		
		asEdgeVertex[8][X] = fX + (a2fVertexOffset[ a2iEdgeConnection[8][0] ][0]  +  
			fOffset * a2fEdgeDirection[8][0]) * fScale;
		asEdgeVertex[8][Y] = fY + (a2fVertexOffset[ a2iEdgeConnection[8][0] ][1]  +  
			fOffset * a2fEdgeDirection[8][1]) * fScale;
		asEdgeVertex[8][Z] = fZ + (a2fVertexOffset[ a2iEdgeConnection[8][0] ][2]  +  
			fOffset * a2fEdgeDirection[8][2]) * fScale;
	}

	if(iEdgeFlags & 512){
		fOffset = fGetOffset(afCubeValue[ a2iEdgeConnection[9][0] ], 
			afCubeValue[ a2iEdgeConnection[9][1] ], 
			fTargetValue);
		
		asEdgeVertex[9][X] = fX + (a2fVertexOffset[ a2iEdgeConnection[9][0] ][0]  +  
			fOffset * a2fEdgeDirection[9][0]) * fScale;
		asEdgeVertex[9][Y] = fY + (a2fVertexOffset[ a2iEdgeConnection[9][0] ][1]  +  
			fOffset * a2fEdgeDirection[9][1]) * fScale;
		asEdgeVertex[9][Z] = fZ + (a2fVertexOffset[ a2iEdgeConnection[9][0] ][2]  +  
			fOffset * a2fEdgeDirection[9][2]) * fScale;
	}

	if(iEdgeFlags & 1024){
		fOffset = fGetOffset(afCubeValue[ a2iEdgeConnection[10][0] ], 
			afCubeValue[ a2iEdgeConnection[10][1] ], 
			fTargetValue);
		
		asEdgeVertex[10][X] = fX + (a2fVertexOffset[ a2iEdgeConnection[10][0] ][0]  +  
			fOffset * a2fEdgeDirection[10][0]) * fScale;
		asEdgeVertex[10][Y] = fY + (a2fVertexOffset[ a2iEdgeConnection[10][0] ][1]  +  
			fOffset * a2fEdgeDirection[10][1]) * fScale;
		asEdgeVertex[10][Z] = fZ + (a2fVertexOffset[ a2iEdgeConnection[10][0] ][2]  +  
			fOffset * a2fEdgeDirection[10][2]) * fScale;
	}

	if(iEdgeFlags & 2048){
		fOffset = fGetOffset(afCubeValue[ a2iEdgeConnection[11][0] ], 
			afCubeValue[ a2iEdgeConnection[11][1] ], 
			fTargetValue);
		
		asEdgeVertex[11][X] = fX + (a2fVertexOffset[ a2iEdgeConnection[11][0] ][0]  +  
			fOffset * a2fEdgeDirection[11][0]) * fScale;
		asEdgeVertex[11][Y] = fY + (a2fVertexOffset[ a2iEdgeConnection[11][0] ][1]  +  
			fOffset * a2fEdgeDirection[11][1]) * fScale;
		asEdgeVertex[11][Z] = fZ + (a2fVertexOffset[ a2iEdgeConnection[11][0] ][2]  +  
			fOffset * a2fEdgeDirection[11][2]) * fScale;
	}

	// Save the triangles that were found.  There can be up to five per cube
	Triangle tmpTriangle;

	if(a2iTriangleConnectionTable[iFlagIndex][0] >= 0){
		tmpTriangle.SetVertex(0, asEdgeVertex[a2iTriangleConnectionTable[iFlagIndex][0]]);
		tmpTriangle.SetVertex(1, asEdgeVertex[a2iTriangleConnectionTable[iFlagIndex][1]]);
		tmpTriangle.SetVertex(2, asEdgeVertex[a2iTriangleConnectionTable[iFlagIndex][2]]);

		// Save the triangle
		triangleList.PushTop(tmpTriangle);
	}

	if(a2iTriangleConnectionTable[iFlagIndex][3] >= 0){
		tmpTriangle.SetVertex(0, asEdgeVertex[a2iTriangleConnectionTable[iFlagIndex][3]]);
		tmpTriangle.SetVertex(1, asEdgeVertex[a2iTriangleConnectionTable[iFlagIndex][4]]);
		tmpTriangle.SetVertex(2, asEdgeVertex[a2iTriangleConnectionTable[iFlagIndex][5]]);

		// Save the triangle
		triangleList.PushTop(tmpTriangle);
	}

	if(a2iTriangleConnectionTable[iFlagIndex][6] >= 0){
		tmpTriangle.SetVertex(0, asEdgeVertex[a2iTriangleConnectionTable[iFlagIndex][6]]);
		tmpTriangle.SetVertex(1, asEdgeVertex[a2iTriangleConnectionTable[iFlagIndex][7]]);
		tmpTriangle.SetVertex(2, asEdgeVertex[a2iTriangleConnectionTable[iFlagIndex][8]]);

		// Save the triangle
		triangleList.PushTop(tmpTriangle);
	}

	if(a2iTriangleConnectionTable[iFlagIndex][9] >= 0){
		tmpTriangle.SetVertex(0, asEdgeVertex[a2iTriangleConnectionTable[iFlagIndex][9]]);
		tmpTriangle.SetVertex(1, asEdgeVertex[a2iTriangleConnectionTable[iFlagIndex][10]]);
		tmpTriangle.SetVertex(2, asEdgeVertex[a2iTriangleConnectionTable[iFlagIndex][11]]);

		// Save the triangle
		triangleList.PushTop(tmpTriangle);
	}

	if(a2iTriangleConnectionTable[iFlagIndex][12] >= 0){
		tmpTriangle.SetVertex(0, asEdgeVertex[a2iTriangleConnectionTable[iFlagIndex][12]]);
		tmpTriangle.SetVertex(1, asEdgeVertex[a2iTriangleConnectionTable[iFlagIndex][13]]);
		tmpTriangle.SetVertex(2, asEdgeVertex[a2iTriangleConnectionTable[iFlagIndex][14]]);

		// Save the triangle
		triangleList.PushTop(tmpTriangle);
	}
}

GLfloat MolSurface::ScalarFieldHeuristic(GLfloat fX, GLfloat fY, GLfloat fZ)
{
	GLfloat fResult = 0.0f;

	int boxIndex = numBox[X]*((int)((fZ - molMin[Z]) * invBoxSize[Z])*numBox[Y] + 
							  (int)((fY - molMin[Y]) * invBoxSize[Y]))
				   + (int)((fX - molMin[X]) * invBoxSize[X]);

	// Don't evaluate points outside the box!
	if((boxIndex < 0) || (boxIndex >= totb)){
		return fResult;
	}

	int tmpIndex;
	GLfloat r2;
	int index_x, index_y, index_z;

	// Check the current box and all neighbooring boxes too
	for(index_z = -1;index_z <= 1;index_z ++){
		for(index_y = -1;index_y <= 1;index_y ++){
			for(index_x = -1;index_x <= 1;index_x ++){
				tmpIndex = boxIndex + index_z*xytotb + index_y*numBox[X] + index_x;

				// Are we out of bounds?
				if((tmpIndex < 0) || (tmpIndex >= totb))
					continue;

				// Get the list of atoms in this box
				list<int>* currBox = atomBox[tmpIndex];
				
				// Are there any atms in this box?
				if(currBox == NULL)
					// No atoms in this box!
					continue;
				
				ListIterator<int> next((*currBox));
				int *ptr;

				while(ptr = next()){
					atom &currAtom = (*ptrAtoms)[*ptr];

					// Don't include solvent atoms in the surface 
					// calculation
					if(!include_solvent_atoms && currAtom.isSolvent)
						continue;

					r2 = currAtom.radius + probeSphere;

					if(traj_index == -1){
						r2 = currAtom.GetDist2(fX, fY, fZ)/(r2*r2);
					}
					else{
						float tmpR[3];

						memcpy(tmpR, &((*ptrTraj)[traj_index][3*(*ptr)]), 3*sizeof(float));
						
						tmpR[X] -= fX;
						tmpR[Y] -= fY;
						tmpR[Z] -= fZ;

						r2 = DOT(tmpR, tmpR)/(r2*r2);
					}
					
#ifdef USE_R6
					fResult += 1.0f/(1.0f + r2*r2*r2);
#else // USE_R8
					r2 *= r2;

					fResult += 1.0f/(1.0f + r2*r2);
#endif

					// If fResult is large enough (i.e. fResult > targetValueCutOff), 
					// don't bother summing over more atoms in this box
					// (i.e. this grid element will not longer contribute to the shape of 
					// the iso-surface).
					if(fResult > targetValueCutOff)
						return fResult;
				}
			}
		}
	}

	return fResult;
}

// The gradient and the desired color of the scalar field at the point {fX, fY, fZ}.
void MolSurface::GradientAndColorIndexHeuristic(GLfloat r[3], GLfloat grad[3], int &colorIndex)
{
	int boxIndex = numBox[X]*((int)((r[Z] - molMin[Z]) * invBoxSize[Z])*numBox[Y] + 
							  (int)((r[Y] - molMin[Y]) * invBoxSize[Y]))
				   + (int)((r[X] - molMin[X]) * invBoxSize[X]);

	int tmpIndex;
	GLfloat  coord[3];
	GLfloat r2, tmp, inv_radius;
	int index_x, index_y, index_z;
	GLfloat min_r2;
	int range = 1;
	const int maxRange = 5;

	// Zero the gradient
	grad[X] = grad[Y] = grad[Z] = 0.0f;
	
	// Set colorIndex to -1 to initialize min_r2.
	colorIndex = -1;

	while(colorIndex == -1){
		// Check the current box and all neighbooring boxes too
		for(index_z = -range;index_z <= range;index_z ++){
			for(index_y = -range;index_y <= range;index_y ++){
				for(index_x = -range;index_x <= range;index_x ++){
					tmpIndex = boxIndex + index_z*xytotb + index_y*numBox[X] + index_x;

					// Are we out of bounds?
					if((tmpIndex < 0) || (tmpIndex >= totb))
						continue;

					// Get the list of atoms in this box
					list<int>* currBox = atomBox[tmpIndex];
					
					// Are there any atms in this box?
					if(currBox == NULL)
						// No atoms in this box!
						continue;
					
					ListIterator<int> next((*currBox));
					int *ptr;

					while(ptr = next()){
						atom &currAtom = (*ptrAtoms)[*ptr];
						
						// Don't include solvent atoms in the surface 
						// calculation
						if(!include_solvent_atoms && currAtom.isSolvent)
							continue;

						if(traj_index == -1){
							currAtom.GetCoord(coord);
						}
						else{
							memcpy(coord, &((*ptrTraj)[traj_index][3*(*ptr)]), 3*sizeof(float));
						}

						inv_radius = 1.0f/(currAtom.radius + probeSphere);

						// Compute \Delta R
						coord[X] = (r[X] - coord[X])*inv_radius;
						coord[Y] = (r[Y] - coord[Y])*inv_radius;
						coord[Z] = (r[Z] - coord[Z])*inv_radius;

						r2 = coord[X]*coord[X] + 
							 coord[Y]*coord[Y] + 
							 coord[Z]*coord[Z];
						
						// Track the index of the closest atom to assign
						// a color to this pactch of the surface
						if(colorIndex == -1){
							min_r2 = r2;
							colorIndex = *ptr;
						}
						else{
							if(min_r2 > r2){
								colorIndex = *ptr;
								min_r2 = r2;
							}
						}
#ifdef USE_R6
						tmp = 1.0f/(r2*r2 + 1.0f/r2);
						tmp = tmp*tmp*inv_radius;
#else // USE_R8
						tmp = 1.0f/(r2*r2*r2 + 1.0f/r2);
						tmp = r2*tmp*tmp*inv_radius;

#endif
						grad[X] += tmp*coord[X];
						grad[Y] += tmp*coord[Y];
						grad[Z] += tmp*coord[Z];
					}
				}
			}
		}

		if(colorIndex != -1)
			return;

		// Increase the range until we find the atom
		range ++;

		if(range >= maxRange)
			throw QmolError("Invalid color index in GradientAndColorIndex", 
				__FILE__, __LINE__);
	}
}

// Build a 3D grid and collect the atom ids that fall into each grid cell.
void MolSurface::BuildAtomBox()
{
	if((ptrAtoms == NULL) || (ptrTraj == NULL))
		throw QmolError("NULL ptr in BuildAtomBox", __FILE__, __LINE__);

	float r[3];
  	int i;
  	int axb, ayb, azb, aindex;
	float tmp_xsize, tmp_ysize, tmp_zsize;
	int num_atom;

	num_atom = ptrAtoms->Length();

	if(num_atom == 0)
		return;

	// Find min/max bounds of molecule's coordinates.
	if(traj_index == -1){
		(*ptrAtoms)[0].GetCoord(r);
	}
	else{
		memcpy(r, &((*ptrTraj)[traj_index][0]), 3*sizeof(float));
	}


	molMin[X] = molMax[X] = r[X];
	molMin[Y] = molMax[Y] = r[Y];
	molMin[Z] = molMax[Z] = r[Z];

  	for(i = 1;i < num_atom; i++){
		if(traj_index == -1){
			(*ptrAtoms)[i].GetCoord(r);
		}
		else{
			memcpy(r, &((*ptrTraj)[traj_index][3*i]), 3*sizeof(float));
		}
		
		if(r[X] < molMin[X])
			molMin[X] = r[X];
		else if(r[X] > molMax[X])
        	molMax[X] = r[X];

		if(r[Y] < molMin[Y])
        	molMin[Y] = r[Y];
		else if(r[Y] > molMax[Y])
        	molMax[Y] = r[Y];

		if(r[Z] < molMin[Z])
        	molMin[Z] = r[Z];
		else if(r[Z] > molMax[Z])
			molMax[Z] = r[Z];
	}

	molMin[X] -= boxLength + 2.0f*probeSphere;
	molMin[Y] -= boxLength + 2.0f*probeSphere;
	molMin[Z] -= boxLength + 2.0f*probeSphere;

	molMax[X] += boxLength + 2.0f*probeSphere;
	molMax[Y] += boxLength + 2.0f*probeSphere;
	molMax[Z] += boxLength + 2.0f*probeSphere;


	// From size of molecule, break up space into boxes of side length=BOX_LENGTH
  	// (or, if the molecule size is too large in any direction, the number of boxes is
  	// increased so that a maximum of MAX_NUM_BOX boxes are used in any direction).
	tmp_xsize = molMax[X] - molMin[X];
	tmp_ysize = molMax[Y] - molMin[Y];
	tmp_zsize = molMax[Z] - molMin[Z];

  	boxSize[X] = tmp_xsize >= maxLength ? tmp_xsize/maxNumBox : boxLength;
  	boxSize[Y] = tmp_ysize >= maxLength ? tmp_ysize/maxNumBox : boxLength;
  	boxSize[Z] = tmp_zsize >= maxLength ? tmp_zsize/maxNumBox : boxLength;

	invBoxSize[X] = 1.0f/boxSize[X];
	invBoxSize[Y] = 1.0f/boxSize[Y];
	invBoxSize[Z] = 1.0f/boxSize[Z];

  	numBox[X] = (int)(tmp_xsize * invBoxSize[X]) + 1;
  	numBox[Y] = (int)(tmp_ysize * invBoxSize[Y]) + 1;
  	numBox[Z] = (int)(tmp_zsize * invBoxSize[Z]) + 1;

  	xytotb = numBox[Y]*numBox[X];
  	totb = xytotb*numBox[Z];
	
	DeleteAtomBox();

	atomBox.Alloc(totb);

	for(i = 0; i < totb;i++)
		atomBox[i] = NULL;

  	// For all the atoms, put them in their particular box
  	for(i = 0; i < num_atom; i++) {
		if(traj_index == -1){
			(*ptrAtoms)[i].GetCoord(r);
		}
		else{
			memcpy(r, &((*ptrTraj)[traj_index][3*i]), 3*sizeof(float));
		}

    	axb = (int)((r[X] - molMin[X]) * invBoxSize[X]);
    	ayb = (int)((r[Y] - molMin[Y]) * invBoxSize[Y]);
    	azb = (int)((r[Z] - molMin[Z]) * invBoxSize[Z]);
		
    	aindex = azb*xytotb + ayb*numBox[X] + axb;

		if(atomBox[aindex] == NULL)
			// Only allocate memory for a list if there is at least one
			// atom in the box
			atomBox[aindex] = new list<int>;

		// Add the current record to box aindex.
		atomBox[aindex]->PushTop(i);
	}
}

void MolSurface::DeleteAtomBox()
{
	// free up the storage space allocted for the grid search
	for(int i = 0;i < atomBox.Length(); i++)
    	if(atomBox[i])  
			delete atomBox[i];

	atomBox.Alloc(0);
}

void MolSurface::BuildVertexArray()
{
	float triMax[3], triMin[3], triSize[3], invTriSize[3];
	int numTriBox[3];
  	int i, index_x, index_y, index_z, tmpIndex;
  	int axb, ayb, azb, aindex;
	float tmp_xsize, tmp_ysize, tmp_zsize;
	int tri_xytot, tri_tot;

	float vertex_cut_off = (meshSize*5.0e-1f)*(meshSize*5.0e-1f);

	int num_vertex = 0;

	if(triangleList.Length() == 0)
		return;

	ListIterator<Triangle> next(triangleList);
	Triangle *ptr;

	// Get the address of the first triangle
	ptr = next();

	// Find min/max bounds of all the triangle verticies
	triMin[X] = triMax[X] = ptr->vertexData[0][X];
	triMin[Y] = triMax[Y] = ptr->vertexData[0][Y];
	triMin[Z] = triMax[Z] = ptr->vertexData[0][Z];

	do{
		for(i = 0;i < 3;i++){
			if(ptr->vertexData[i][X] < triMin[X])
				triMin[X] = ptr->vertexData[i][X];
			else if(ptr->vertexData[i][X] > triMax[X])
				triMax[X] = ptr->vertexData[i][X];
			
			if(ptr->vertexData[i][Y] < triMin[Y])
				triMin[Y] = ptr->vertexData[i][Y];
			else if(ptr->vertexData[i][Y] > triMax[Y])
				triMax[Y] = ptr->vertexData[i][Y];
			
			if(ptr->vertexData[i][Z] < triMin[Z])
				triMin[Z] = ptr->vertexData[i][Z];
			else if(ptr->vertexData[i][Z] > triMax[Z])
				triMax[Z] = ptr->vertexData[i][Z];
		}
	}
	while(ptr = next());


	// From size of molecule, break up space into boxes of side length=VERTEX_BOX_LENGTH
  	// (or, if the molecule size is too large in any direction, the number of boxes is
  	// increased so that a maximum of VERTEX_MAX_NUM_BOX boxes are used in any direction).
	tmp_xsize = triMax[X] - triMin[X];
	tmp_ysize = triMax[Y] - triMin[Y];
	tmp_zsize = triMax[Z] - triMin[Z];

  	triSize[X] = tmp_xsize >= VERTEX_MAX_LENGTH ? tmp_xsize/VERTEX_MAX_NUM_BOX : VERTEX_BOX_LENGTH;
  	triSize[Y] = tmp_ysize >= VERTEX_MAX_LENGTH ? tmp_ysize/VERTEX_MAX_NUM_BOX : VERTEX_BOX_LENGTH;
  	triSize[Z] = tmp_zsize >= VERTEX_MAX_LENGTH ? tmp_zsize/VERTEX_MAX_NUM_BOX : VERTEX_BOX_LENGTH;

	invTriSize[X] = 1.0f/triSize[X];
	invTriSize[Y] = 1.0f/triSize[Y];
	invTriSize[Z] = 1.0f/triSize[Z];

  	numTriBox[X] = (int)(tmp_xsize * invTriSize[X]) + 1;
  	numTriBox[Y] = (int)(tmp_ysize * invTriSize[Y]) + 1;
  	numTriBox[Z] = (int)(tmp_zsize * invTriSize[Z]) + 1;

  	tri_xytot = numTriBox[Y]*numTriBox[X];
  	tri_tot = tri_xytot*numTriBox[Z];

	vertexBox.Alloc(tri_tot);

	for(i = 0; i < tri_tot;i++)
		vertexBox[i] = NULL;

  	// For all the verticies, put them in their particular box
	next.Init(triangleList);

	while(ptr = next()){
		for(i = 0;i < 3;i++){
    		axb = (int)((ptr->vertexData[i][X] - triMin[X]) * invTriSize[X]);
    		ayb = (int)((ptr->vertexData[i][Y] - triMin[Y]) * invTriSize[Y]);
    		azb = (int)((ptr->vertexData[i][Z] - triMin[Z]) * invTriSize[Z]);
				
    		aindex = azb*tri_xytot + ayb*numTriBox[X] + axb;
			
#ifdef _DEBUG
			// Are we out of bounds?
			if((aindex < 0) || (aindex >= tri_tot))
				throw QmolError("Vertex index out of bounds", __FILE__, __LINE__);
#endif // _DEBUG

			if(vertexBox[aindex] == NULL)
				// Only allocate memory for a list if there is at least one
				// atom in the box
				vertexBox[aindex] = new list<VertexInfo>;
			
			// Vertex is distinct. Save the vertex
			VertexInfo tmpInfo;

			tmpInfo.index = i;
			tmpInfo.parent = ptr;
			vertexBox[aindex]->PushTop(tmpInfo);				
		}
	}

	// For all the verticies, locate the nearest neightboor vertex
	next.Init(triangleList);

	while(ptr = next()){
		for(i = 0;i < 3;i++){

			// Skip assigned verticies
			if(ptr->vertexIndex[i] != -1)
				continue;

    		axb = (int)((ptr->vertexData[i][X] - triMin[X]) * invTriSize[X]);
    		ayb = (int)((ptr->vertexData[i][Y] - triMin[Y]) * invTriSize[Y]);
    		azb = (int)((ptr->vertexData[i][Z] - triMin[Z]) * invTriSize[Z]);
				
    		aindex = azb*tri_xytot + ayb*numTriBox[X] + axb;
			
#ifdef _DEBUG
			// Are we out of bounds?
			if((aindex < 0) || (aindex >= tri_tot))
				throw QmolError("Vertex index out of bounds", __FILE__, __LINE__);

			if(vertexBox[aindex] == NULL)
				// There must be at least one vertex in this box
				throw QmolError("Empty Vertex Box", __FILE__, __LINE__);
#endif // _DEBUG

			// Check the current box and all neighbooring boxes too
			for(index_z = -1;index_z <= 1;index_z ++){
				for(index_y = -1;index_y <= 1;index_y ++){
					for(index_x = -1;index_x <= 1;index_x ++){
						tmpIndex = aindex + index_z*tri_xytot + index_y*numTriBox[X] + index_x;

						// Are we out of bounds?
						if((tmpIndex < 0) || (tmpIndex >= tri_tot))
							continue;

						// Get the list of verticies in this box
						list<VertexInfo>* currBox = vertexBox[tmpIndex];
							
						// Are there any verticies in this box?
						if(currBox == NULL)
							// No verticies in this box!
							continue;
							
						ListIterator<VertexInfo> nextVertex(*currBox);
						VertexInfo* tmpPtr;
						
						while(tmpPtr = nextVertex()){
							// Are the verticies in ptr distinct from the verticies
							// in tmpPtr?
							GLfloat dr2, tmp_dr2;

							dr2 = ptr->vertexPtr[i][X] - 
								tmpPtr->parent->vertexPtr[tmpPtr->index][X];

							dr2 *= dr2;

							if(dr2 > vertex_cut_off)
								continue;

							tmp_dr2 = ptr->vertexPtr[i][Y] - 
								tmpPtr->parent->vertexPtr[tmpPtr->index][Y];

							dr2 += tmp_dr2*tmp_dr2;

							if(dr2 > vertex_cut_off)
								continue;

							tmp_dr2 = ptr->vertexPtr[i][Z] - 
								tmpPtr->parent->vertexPtr[tmpPtr->index][Z];

							dr2 += tmp_dr2*tmp_dr2;

							if(dr2 > vertex_cut_off)
								continue;

							if(tmpPtr->parent->vertexIndex[tmpPtr->index] == -1){
								if(ptr->vertexIndex[i] == -1){
									ptr->vertexIndex[i] = num_vertex;
									num_vertex++;
								}
								
								tmpPtr->parent->vertexIndex[tmpPtr->index] = 
									ptr->vertexIndex[i];
								
								tmpPtr->parent->vertexPtr[tmpPtr->index] = 
									ptr->vertexPtr[i];
							}
							else{
								if(ptr->vertexIndex[i] == -1){
									ptr->vertexIndex[i] = 
										tmpPtr->parent->vertexIndex[tmpPtr->index];
									ptr->vertexPtr[i] = 
										tmpPtr->parent->vertexPtr[tmpPtr->index];
								}
							}
						}
					}
				}
			}
		}
	}

	// Free box memory
	for(i = 0; i < tri_tot;i++)
		if(vertexBox[i] != NULL)
			delete vertexBox[i];
	
	vertexBox.Alloc(0);

	// Whew! That was a lot of work just to determine which vertex goes where ...
	// Now build an array of verticies
	vertexArray.Alloc(3*num_vertex);
	normalArray.Alloc(3*num_vertex);
	colorArray.Alloc(num_vertex);
	int triIndex[3];

	// Keep the user informed
	CMainFrame *pFrame = (CMainFrame*)AfxGetApp()->m_pMainWnd;

	const int step_every = 10;
	int max_count = triangleList.Length() + num_vertex;
	int progress_counter = 0;

	next.Init(triangleList);

	int count = 0;

	while(ptr = next()){

		// The indicies of each vertex in the triangle
		triIndex[0] = 3*ptr->vertexIndex[0];
		triIndex[1] = 3*ptr->vertexIndex[1];
		triIndex[2] = 3*ptr->vertexIndex[2];

		// Copy the triangle coordinates into the vertex 
		// array (this is the array that we will render from).
		vertexArray[triIndex[0]] = ptr->vertexPtr[0][X];
		vertexArray[triIndex[0] + 1] = ptr->vertexPtr[0][Y];
		vertexArray[triIndex[0] + 2] = ptr->vertexPtr[0][Z];
		
		vertexArray[triIndex[1]] = ptr->vertexPtr[1][X];
		vertexArray[triIndex[1] + 1] = ptr->vertexPtr[1][Y];
		vertexArray[triIndex[1] + 2] = ptr->vertexPtr[1][Z];
		
		vertexArray[triIndex[2]] = ptr->vertexPtr[2][X];
		vertexArray[triIndex[2] + 1] = ptr->vertexPtr[2][Y];
		vertexArray[triIndex[2] + 2] = ptr->vertexPtr[2][Z];

		if((triIndex[0] == triIndex[1]) || (triIndex[0] == triIndex[2]) ||
			(triIndex[1] == triIndex[2])){

			// This triangle is NOT rendered
			ptr->vertexPtr[0] = NULL;
		}
		else{
			// Count the number of valid triangles
			count++;
		}

		progress_counter ++;

		if((progress_counter%step_every == 0) && pFrame)
			// Advance the progress bar
			pFrame->SetProgress((100*progress_counter)/max_count + 100);
	}

	// Save the vertex indicies of triangles with non-zero area
	indexArray.Alloc(3*count);

	count = 0;
	next.Init(triangleList);

	while(ptr = next()){
		if(ptr->vertexPtr[0] != NULL){
			for(i = 0;i < 3;i++){	
				// Save the index
				indexArray[count] = ptr->vertexIndex[i];
				count++;
			}
		}
	}

	// Clean up the triangle list
	while(!triangleList.IsEmpty())
		triangleList.Pop();

	// Compute the normal and color
	for(i = 0;i < num_vertex;i++){
		GradientAndColorIndex(vertexArray.Pointer() + i*3, 
			normalArray.Pointer() + i*3, colorArray[i]);

		progress_counter ++;

		if((progress_counter%step_every == 0) && pFrame)
			// Advance the progress bar
			pFrame->SetProgress((100*progress_counter)/max_count + 100);
	}

	// Make sure that all of the triangles have the correct winding order
	for(i = 0;i < indexArray.Length();i += 3){
		int a, b, c;
		float AB[3], CB[3], N[3], tmpN[3];
		
		a = 3*indexArray[i];
		b = 3*indexArray[i + 1];
		c = 3*indexArray[i + 2];

		AB[X] = vertexArray[a + X] - vertexArray[b + X];
		AB[Y] = vertexArray[a + Y] - vertexArray[b + Y];
		AB[Z] = vertexArray[a + Z] - vertexArray[b + Z];
		
		CB[X] = vertexArray[c + X] - vertexArray[b + X];
		CB[Y] = vertexArray[c + Y] - vertexArray[b + Y];
		CB[Z] = vertexArray[c + Z] - vertexArray[b + Z];

		// Compute the cross product
		CROSS(N, AB, CB);

		//const float eps = 1.0e-7f;
		//const float eps = 1.0e-2f;
		const float eps = 5.0e-1f;

		// Normalize N
		float mag = 1.0f/(float)sqrt(DOT(N, N));
		
		N[X] *= mag;
		N[Y] *= mag;
		N[Z] *= mag;

		memcpy(tmpN, normalArray.Pointer() + a, 3*sizeof(float));

		mag = (float)sqrt(DOT(tmpN, tmpN));

		if(DOT(N, tmpN)/mag > eps){			

			memcpy(tmpN, normalArray.Pointer() + b, 3*sizeof(float));

			mag = (float)sqrt(DOT(tmpN, tmpN));
			
			if(DOT(N, tmpN)/mag > eps){

				memcpy(tmpN, normalArray.Pointer() + c, 3*sizeof(float));

				mag = (float)sqrt(DOT(tmpN, tmpN));
			
				if(DOT(N, tmpN)/mag > eps){
					// Swap the a and c indicies to reverse the winding order
					indexArray[i] = c/3;
					indexArray[i + 2] = a/3;
				}
			}
		}
	}
}

int MolSurface::GetSurfQuality()
{
	// "Quality" is inversely proportional to mesh size
	return (int)(100.0f*(1.0f - (meshSize - MIN_MESH_SIZE)/
		(MAX_MESH_SIZE - MIN_MESH_SIZE)));
}

void MolSurface::SetSurfQuality(int m_quality)
{
	// "Quality" is inversely proportional to mesh size
	if((m_quality >= 0) && (m_quality <= 100)){
		meshSize = (100 - m_quality)*(MAX_MESH_SIZE - MIN_MESH_SIZE)/100.0f + 
			MIN_MESH_SIZE;
	}
}

void MolSurface::GetAtomSurfaceArea(array<float> &m_area)
{
	if(!validSurface)
		throw QmolError("Can not compute surface area for invalid surface",
			__FILE__, __LINE__);

	int i;

	// Zero the area arrat
	for(i = 0;i < m_area.Length();i++){
		m_area[i] = 0.0f;
	}

	int num_tri_vertex = indexArray.Length();

	for(i = 0;i < num_tri_vertex;i += 3){
		float triArea = TriangleArea(indexArray[i], 
									 indexArray[i+1], 
									 indexArray[i+2]);
		
		// Each vertex will be assigned 1/3 of the triangle area
		triArea /= 3.0f;

		if(colorArray[indexArray[i]] >= m_area.Length()){
			throw QmolError("Invalid atom index in GetAtomSurfaceArea", 
				__FILE__, __LINE__);
		}
		else{
			m_area[colorArray[indexArray[i]]] += triArea;
		}

		if(colorArray[indexArray[i+1]] >= m_area.Length()){
			throw QmolError("Invalid atom index in GetAtomSurfaceArea", 
				__FILE__, __LINE__);
		}
		else{
			m_area[colorArray[indexArray[i+1]]] += triArea;
		}

		if(colorArray[indexArray[i+2]] >= m_area.Length()){
			throw QmolError("Invalid atom index in GetAtomSurfaceArea", 
				__FILE__, __LINE__);
		}
		else{
			m_area[colorArray[indexArray[i+2]]] += triArea;
		}
	}
}

// Traingle area is (1/2) base * hieght == 1/2 |A X B|
float MolSurface::TriangleArea(int a, int b, int c)
{
	float A[3], B[3], C[3];

	A[X] = vertexArray[3*a + X] - vertexArray[3*b + X];
	A[Y] = vertexArray[3*a + Y] - vertexArray[3*b + Y];
	A[Z] = vertexArray[3*a + Z] - vertexArray[3*b + Z];

	B[X] = vertexArray[3*c + X] - vertexArray[3*b + X];
	B[Y] = vertexArray[3*c + Y] - vertexArray[3*b + Y];
	B[Z] = vertexArray[3*c + Z] - vertexArray[3*b + Z];

	CROSS(C, A, B);

	return 0.5f*(float)sqrt(DOT(C, C));
}
