//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.
//
//	Look-and-feel inspired by Xmol 
//	(http://www.networkcs.com/msc/docs/xmol/).
//
//	OpenGL selection code and quaterion rotation matrix
//	formulation from molview demo program by Mark Kilgard
//	(http://reality.sgi.com/opengl/OpenGLforX.html). PostScript
//	rendering also from Mark Kilgard.
//
//	The printing routines are based on code from
//	Craig Fahrnbach <craignan@home.com> via
//	Uwe Kotyczka <kotyczka@bnm-gmbh.de>.
//
//	Covalent bond determination code based on VMD
//	http://www.ks.uiuc.edu/Research/vmd/
//
//	Stride (http://www.embl-heidelberg.de/stride/stride.html)
//	is used for secondary structure determination and has been 
//	included in Qmol with the kind permission of 
//	Dmitrij Frishman, PhD
//	Institute for Bioinformatics
//	GSF - Forschungszentrum fr Umwelt und Gesundheit, GmbH
//	Ingolstdter Landstrae 1,
//	D-85764 Neuherberg, Germany
//	Telephone: +49-89-3187-4201
//	Fax: +49-89-31873585
//	e-mail: d.frishman@gsf.de
//	WWW: http://mips.gsf.de/mips/staff/frishman/

#include "stdafx.h"

// Include the OpenGL headers
#include <GL\gl.h>
#include <GL\glu.h>
#include <math.h>
#include "Excepts.h"

// Need a definition of PI
#include "Qmol.h"

// Partial sphere code based on Mesa gluSphere implementation.
// See http://www.mesa3d.org/ for details.
void gluPartialSphere(GLfloat radius, GLint slices, GLint stacks)
{
	GLfloat rho, drho, theta, dtheta;
	GLfloat x, y, z;
	GLint i, j, imin, imax;

	drho = PI / (GLfloat) stacks;
	dtheta = 2.0f * PI / (GLfloat) slices;

	imin = 1;
	imax = stacks/2;

	// draw +Z end as a triangle fan
	glBegin(GL_TRIANGLE_FAN);
	glNormal3f(0.0f, 0.0f, 1.0f);
	glVertex3f(0.0f, 0.0f, radius);
	for (j = 0; j <= slices; j++) {
		theta = (j == slices) ? 0.0f : j * dtheta;
	    x = (GLfloat)(-sin(theta) * sin(drho));
	    y = (GLfloat)(cos(theta) * sin(drho));
	    z = (GLfloat)(cos(drho));
		glNormal3f(x, y, z);
	    glVertex3f(x * radius, y * radius, z * radius);
	}
	glEnd();

	// draw intermediate stacks as quad strips
	for (i = imin; i < imax; i++) {
		rho = i * drho;
		glBegin(GL_QUAD_STRIP);
		for (j = 0; j <= slices; j++) {
			theta = (j == slices) ? 0.0f : j * dtheta;
			x = (GLfloat)(-sin(theta) * sin(rho));
			y = (GLfloat)(cos(theta) * sin(rho));
			z = (GLfloat)(cos(rho));
			glNormal3f(x, y, z);
			glVertex3f(x * radius, y * radius, z * radius);
			x = (GLfloat)(-sin(theta) * sin(rho + drho));
			y = (GLfloat)(cos(theta) * sin(rho + drho));
			z = (GLfloat)(cos(rho + drho));
			glNormal3f(x, y, z);
			glVertex3f(x * radius, y * radius, z * radius);
		}
		glEnd();
	}
}

void MakeSphere(int m_listID, int m_resolution)
{
	// Use fewer stacks than slices to improve rendering performance.
	int num_stack = max(m_resolution - 6, 2);

	glNewList(m_listID, GL_COMPILE);
	gluPartialSphere(1.0f, m_resolution, num_stack);
	glEndList();
}

void MakeDisk(int m_listID, int m_resolution)
{
	GLUquadricObj *quadObj;

	quadObj = gluNewQuadric();

	if(!quadObj)
		throw QMOL_ERROR("Can't allocate memory for disk!");

	gluQuadricDrawStyle(quadObj, GLU_FILL);
  	gluQuadricOrientation(quadObj, GLU_OUTSIDE);
  	gluQuadricNormals(quadObj, GLU_SMOOTH);
	
  	glNewList(m_listID, GL_COMPILE);
  	gluDisk(quadObj, 0.0, 1.0, m_resolution, 1);
  	glEndList();

  	gluDeleteQuadric(quadObj);
}

void MakeCylinder(int m_listID, GLfloat height, int m_resolution)
{
	GLUquadricObj *quadObj;

	quadObj = gluNewQuadric();

	if(!quadObj)
		throw QMOL_ERROR("Can't allocate memory for cylinder!");

	gluQuadricDrawStyle(quadObj, GLU_FILL);
  	gluQuadricOrientation(quadObj, GLU_OUTSIDE);
  	gluQuadricNormals(quadObj, GLU_SMOOTH);
	
  	glNewList(m_listID, GL_COMPILE);
  	gluCylinder(quadObj, 1.0, 1.0, height, m_resolution, 1);
  	glEndList();

  	gluDeleteQuadric(quadObj);
}

void MakeCone(int m_listID, GLfloat height, int m_resolution)
{
	GLUquadricObj *quadObj;

	quadObj = gluNewQuadric();

	if(!quadObj)
		throw QMOL_ERROR("Can't allocate memory for cone!");

	gluQuadricDrawStyle(quadObj, GLU_FILL);
  	gluQuadricOrientation(quadObj, GLU_OUTSIDE);
  	gluQuadricNormals(quadObj, GLU_SMOOTH);
	
  	glNewList(m_listID, GL_COMPILE);
  	gluCylinder(quadObj, 1.0, 0.0, height, m_resolution, 1);
  	glEndList();

  	gluDeleteQuadric(quadObj);
}

double Det(double m[][3])
{
	return m[0][0]*(m[1][1]*m[2][2] - m[1][2]*m[2][1]) - 
		   m[0][1]*(m[1][0]*m[2][2] - m[1][2]*m[2][0]) + 
		   m[0][2]*(m[1][0]*m[2][1] - m[1][1]*m[2][0]);
}

void InvertMatrix(GLfloat m[][4])
{
	int i, j;
 
	GLfloat c[3][3];

	c[0][0] = (m[1][1]*m[2][2] - m[2][1]*m[1][2]);
	c[0][1] = -(m[1][0]*m[2][2] - m[2][0]*m[1][2]);
	c[0][2] = (m[1][0]*m[2][1] - m[2][0]*m[1][1]);

	c[1][0] = -(m[0][1]*m[2][2] - m[0][2]*m[2][1]);
	c[1][1] = (m[0][0]*m[2][2] - m[0][2]*m[2][0]);
	c[1][2] = -(m[0][0]*m[2][1] - m[0][1]*m[2][0]);

	c[2][0] = (m[0][1]*m[1][2] - m[0][2]*m[1][1]);
	c[2][1] = -(m[0][0]*m[1][2] - m[0][2]*m[1][0]);
	c[2][2] = (m[0][0]*m[1][1] - m[0][1]*m[1][0]);

	GLfloat det_m = m[0][0]*(m[1][1]*m[2][2] - m[1][2]*m[2][1]) - 
					m[0][1]*(m[1][0]*m[2][2] - m[1][2]*m[2][0]) + 
					m[0][2]*(m[1][0]*m[2][1] - m[1][1]*m[2][0]);

	det_m = 1.0f/det_m;

	for(i = 0;i < 3;i++)
		for(j = 0;j < 3;j++)
			// m equals the transpose of cofactor matrix
			m[i][j] = det_m*c[j][i];
}

void InvertMatrix(GLfloat dest[][4], GLfloat src[][4])
{
	int i, j;
 
	GLfloat c[3][3];

	c[0][0] = (src[1][1]*src[2][2] - src[2][1]*src[1][2]);
	c[0][1] = -(src[1][0]*src[2][2] - src[2][0]*src[1][2]);
	c[0][2] = (src[1][0]*src[2][1] - src[2][0]*src[1][1]);

	c[1][0] = -(src[0][1]*src[2][2] - src[0][2]*src[2][1]);
	c[1][1] = (src[0][0]*src[2][2] - src[0][2]*src[2][0]);
	c[1][2] = -(src[0][0]*src[2][1] - src[0][1]*src[2][0]);

	c[2][0] = (src[0][1]*src[1][2] - src[0][2]*src[1][1]);
	c[2][1] = -(src[0][0]*src[1][2] - src[0][2]*src[1][0]);
	c[2][2] = (src[0][0]*src[1][1] - src[0][1]*src[1][0]);

	GLfloat det_m = src[0][0]*(src[1][1]*src[2][2] - src[1][2]*src[2][1]) - 
					src[0][1]*(src[1][0]*src[2][2] - src[1][2]*src[2][0]) + 
					src[0][2]*(src[1][0]*src[2][1] - src[1][1]*src[2][0]);

	det_m = 1.0f/det_m;

	for(i = 0;i < 3;i++)
		for(j = 0;j < 3;j++)
			// m equals the transpose of cofactor matrix
			dest[i][j] = det_m*c[j][i];
}

void InvertMatrix(double m[][3])
{
	int i, j;
 
	double c[3][3];

	c[0][0] = (m[1][1]*m[2][2] - m[2][1]*m[1][2]);
	c[0][1] = -(m[1][0]*m[2][2] - m[2][0]*m[1][2]);
	c[0][2] = (m[1][0]*m[2][1] - m[2][0]*m[1][1]);

	c[1][0] = -(m[0][1]*m[2][2] - m[0][2]*m[2][1]);
	c[1][1] = (m[0][0]*m[2][2] - m[0][2]*m[2][0]);
	c[1][2] = -(m[0][0]*m[2][1] - m[0][1]*m[2][0]);

	c[2][0] = (m[0][1]*m[1][2] - m[0][2]*m[1][1]);
	c[2][1] = -(m[0][0]*m[1][2] - m[0][2]*m[1][0]);
	c[2][2] = (m[0][0]*m[1][1] - m[0][1]*m[1][0]);

	double det_m = m[0][0]*(m[1][1]*m[2][2] - m[1][2]*m[2][1]) - 
				   m[0][1]*(m[1][0]*m[2][2] - m[1][2]*m[2][0]) + 
				   m[0][2]*(m[1][0]*m[2][1] - m[1][1]*m[2][0]);

	det_m = 1.0/det_m;

	for(i = 0;i < 3;i++)
		for(j = 0;j < 3;j++)
			// m equals the transpose of cofactor matrix
			m[i][j] = det_m*c[j][i];
}

float NormalizeByRange(float x, float min, float max)
{
	// Don't divide by zero
	if(fabs(max - min) < 1.0e-6)
		return 1.0f;

	return (x - min)/(max - min);
}

float ScaleByRange(float x, float min, float max)
{
	// Don't divide by zero
	if(fabs(max - min) < 1.0e-6)
		return 1.0f;

	return x/(max - min);
}
