//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.
//
//	Look-and-feel inspired by Xmol 
//	(http://www.networkcs.com/msc/docs/xmol/).
//
//	OpenGL selection code and quaterion rotation matrix
//	formulation from molview demo program by Mark Kilgard
//	(http://reality.sgi.com/opengl/OpenGLforX.html). PostScript
//	rendering also from Mark Kilgard.
//
//	The printing routines are based on code from
//	Craig Fahrnbach <craignan@home.com> via
//	Uwe Kotyczka <kotyczka@bnm-gmbh.de>.
//
//	Covalent bond determination code based on VMD
//	http://www.ks.uiuc.edu/Research/vmd/
//
//	Stride (http://www.embl-heidelberg.de/stride/stride.html)
//	is used for secondary structure determination and has been 
//	included in Qmol with the kind permission of 
//	Dmitrij Frishman, PhD
//	Institute for Bioinformatics
//	GSF - Forschungszentrum fr Umwelt und Gesundheit, GmbH
//	Ingolstdter Landstrae 1,
//	D-85764 Neuherberg, Germany
//	Telephone: +49-89-3187-4201
//	Fax: +49-89-31873585
//	e-mail: d.frishman@gsf.de
//	WWW: http://mips.gsf.de/mips/staff/frishman/
//
// Implement the kabsch rotation for aligning a vector (or set of vectors) to
// a reference vector such that the RMSD (i.e. root-mean-square-deviation) is
// minimized. See Acta Cryst. (1978). A34, 827-828 W.K. Kabsch

#include "stdafx.h"
#include "kabsch.h"
#include "LinearAlgebra3D.h"
#include "Excepts.h"
#include <complex>

int EigenSystem3D(double a[3][3], double vect[3][3], double val[3]);
void EigenValues3D(double a[3][3], double val[3]);

#define		ALIGNMENT_EPS	1.0e-7

// ref ... The reference vector
// vec ... The vector to be aligned to ref
// w   ... A vector of weights
void kabsch(array<atom> &ref, array<float> &vec, array<float> &w)
{
	// The number of atoms (i.e. the number of 3 space vectors)
	int num = ref.Length();
	int i, j, k;

	int small = X;
	int med = X;
	int large = X;

	double R[3][3], RtR[3][3], U[3][3];
	double A[3][3], B[3][3];
	double eigenVal[3];
	double mag;

	for(i = X;i <= Z;i++)
		for(j = X;j <= Z;j++)
			R[i][j] = RtR[i][j] = 0.0;

	for(i = 0;i < num;i++){
		R[X][X] += w[i]*ref[i].coord[X]*vec[3*i + X];
		R[X][Y] += w[i]*ref[i].coord[X]*vec[3*i + Y];
		R[X][Z] += w[i]*ref[i].coord[X]*vec[3*i + Z];

		R[Y][X] += w[i]*ref[i].coord[Y]*vec[3*i + X];
		R[Y][Y] += w[i]*ref[i].coord[Y]*vec[3*i + Y];
		R[Y][Z] += w[i]*ref[i].coord[Y]*vec[3*i + Z];

		R[Z][X] += w[i]*ref[i].coord[Z]*vec[3*i + X];
		R[Z][Y] += w[i]*ref[i].coord[Z]*vec[3*i + Y];
		R[Z][Z] += w[i]*ref[i].coord[Z]*vec[3*i + Z];
	}

	// RtR holds R^{transpose).R
	for(i = X;i <= Z;i++)
		for(j = X;j <= Z;j++)
			for(k = X;k <= Z;k++)
				RtR[i][j] += R[k][i]*R[k][j];

	// Compute the eigenvalues and eigen vectors of RtR.
	//RtR.Jacobi(A, eigenVal);
	if(EigenSystem3D(RtR, A, eigenVal) == FALSE)
		throw QMOL_ERROR("kabsch:: Unable to diagonalize matrix");

	// Test all of the eigen values to make sure that we have a valid
	// solution.Only skip structures where all of the atoms are colinear
	// (i.e. 2 zero eigenvalues).
	int nullsubspace = 0;

	for(i = X;i <= Z;i++){
		if(eigenVal[i] < ALIGNMENT_EPS){
			//throw QMOL_ERROR("Zero eigen value in kabsch: Colinear and "
			//	"coplanar molecules are not aligned.");
			nullsubspace ++;
		}
	}

	if(nullsubspace > 1){
		throw QMOL_ERROR("Zero eigen values in kabsch: Colinear molecules "
			"are not currently allowed.");
	}

	// Sort the eigen values (in case we need to remove improper
	// an rotation).
	for(i = Y;i <= Z;i++){
		if(fabs(eigenVal[i]) < fabs(eigenVal[small]))
			small = i;

		if(fabs(eigenVal[i]) > fabs(eigenVal[large]))
			large = i;
	}

	if((large + 1)%3 == small)
		med = (large + 2)%3;
	else
		med = (large + 1)%3;


	double smallVec[3];
	double medVec[3];
	double largeVec[3];

	for(i = X;i <= Z;i++){
		medVec[i] = A[i][med];
		largeVec[i] = A[i][large];
	}

	// *************************************************************************
	// If you get this error message, you need to get the most recent
	// patch for the MS compiler!
	// fatal error C1001: INTERNAL COMPILER ERROR
    // (compiler file 'E:\8168\vc98\p2\src\P2\main.c', line 494)
    // Please choose the Technical Support command on the Visual C++
    // Help menu, or open the Technical Support help file for more information
	// *************************************************************************

	// The cross product defines the eigen vector cooresponding to
	// the smallest eigen value.
	CROSS(smallVec, largeVec, medVec);

	for(i = X;i <= Z;i++)
		A[i][small] = smallVec[i];
	
	/* Invert the eigenvalues as called for in kabsch */
	if(nullsubspace){
		eigenVal[large] = 1.0/sqrt(eigenVal[large]);
		eigenVal[med] = 1.0/sqrt(eigenVal[med]);

		if(eigenVal[large]*eigenVal[med] < 0.0)
			eigenVal[med] *= -1.0;
	}
	else{
		for(i = X;i <= Z;i++)
			eigenVal[i] = 1.0/sqrt(eigenVal[i]);

		if(eigenVal[X]*eigenVal[Y]*eigenVal[Z] < 0.0)
			eigenVal[small] *= -1.0;
	}

    for(i = X;i <= Z;i++)
		for(j = X;j <= Z;j++){
			B[i][j] = 0.0;
			for(k = X;k <= Z;k++)
    			B[i][j] += R[j][k]*A[k][i]*eigenVal[i];
		}
	
	for(i = X;i <= Z;i++){
		medVec[i] = B[med][i];
		largeVec[i] = B[large][i];
	}

	mag = 1.0/sqrt(DOT(medVec,medVec));

	medVec[X] *= mag;
	medVec[Y] *= mag;
	medVec[Z] *= mag;

	mag = 1.0/sqrt(DOT(largeVec,largeVec));

	largeVec[X] *= mag;
	largeVec[Y] *= mag;
	largeVec[Z] *= mag;

	// The cross product defines the eigen vector cooresponding to
	// the smallest eigen value.
	//smallVec = largeVec | medVec;
	CROSS(smallVec, largeVec, medVec);

	for(i = X;i <= Z;i++)
		B[small][i] = smallVec[i];

    for(i = X;i <= Z;i++)
		for(j = X;j <= Z;j++){
			U[i][j] = 0.0;
			for(k = X;k <= Z;k++)
				U[i][j] += A[j][k]*B[k][i];
		}

	for(i = 0;i < num;i++){
		double tmp[3];

		tmp[X] = U[X][X]*vec[3*i + X] + U[X][Y]*vec[3*i + Y] + 
			U[X][Z]*vec[3*i + Z];
		tmp[Y] = U[Y][X]*vec[3*i + X] + U[Y][Y]*vec[3*i + Y] + 
			U[Y][Z]*vec[3*i + Z];
		tmp[Z] = U[Z][X]*vec[3*i + X] + U[Z][Y]*vec[3*i + Y] + 
			U[Z][Z]*vec[3*i + Z];

		vec[3*i + X] = (float)tmp[X];
		vec[3*i + Y] = (float)tmp[Y];
		vec[3*i + Z] = (float)tmp[Z];
    }
}

float ComputeRmsd(molecule *molA, molecule *molB, 
				  list<int> &selectA, list<int> &selectB,
				  BOOL mass_weight)
{
	ASSERT(selectA.Length() == selectB.Length());
	
	double rmsd = 0.0;
	double total_mass = 0.0;

	float mA[4][4], mB[4][4];

	molA->BuildRotationMatrix(mA);
	molB->BuildRotationMatrix(mB);

	ListIterator<int> nextA(selectA);
	ListIterator<int> nextB(selectB);
	int *indexA, *indexB;

	while((indexA = nextA()) && (indexB = nextB())){
		atom &refA = molA->atoms[*indexA];
		atom &refB = molB->atoms[*indexB];

		float tmpA[3], tmpB[3];
		float finalA[3], finalB[3];

		if(molA->current_step == 0){	
			memcpy(tmpA, refA.coord, 3*sizeof(float));
		}
		else{
			memcpy(tmpA, &(molA->traj[molA->current_step - 1][3*(*indexA)]), 
				3*sizeof(float));
		}

		TRANSLATE(tmpA, molA->R);
		ROTATE(finalA, mA, tmpA);

		if(molB->current_step == 0){
			memcpy(tmpB, refB.coord, 3*sizeof(float));
		}
		else{
			memcpy(tmpB, &(molB->traj[molB->current_step - 1][3*(*indexB)]), 
				3*sizeof(float));
		}

		TRANSLATE(tmpB, molB->R);
		ROTATE(finalB, mB, tmpB);

		float w;
		float tmp;
		float tmpRmsd;

		if(mass_weight){
			// The masses of each atom may be different, so use
			// a symeterized mass weighting
			w = 0.5f*(refA.mass + refB.mass);
		}
		else{
			w = 1.0f;
		}
		
		tmp = finalA[X] - finalB[X];
		tmpRmsd = tmp*tmp;

		tmp = finalA[Y] - finalB[Y];
		tmpRmsd += tmp*tmp;

		tmp = finalA[Z] - finalB[Z];
		tmpRmsd += tmp*tmp;

		rmsd += w*tmpRmsd;

		total_mass += w;
	}

	return (float)(sqrt(rmsd/total_mass));
}


// Align molecule B onto molecule A by a ridgid body
// translation and rotation.
void AlignMolecules(molecule *molA, molecule *molB, 
				    list<int> &selectA, list<int> &selectB, 
				    BOOL mass_weight)
{
	ASSERT(selectA.Length() == selectB.Length());

	// First align the center of mass of the selected atoms
	float mA[4][4], mB[4][4], mBInv[4][4];
	
	double total_mass = 0.0;

	double comA[3];
	double comB[3];
	double tmp[3];

	memset(comA, 0, 3*sizeof(double));
	memset(comB, 0, 3*sizeof(double));

	molA->BuildRotationMatrix(mA);
	molB->BuildRotationMatrix(mB);

	ListIterator<int> nextA(selectA);
	ListIterator<int> nextB(selectB);
	int *indexA, *indexB;

	while((indexA = nextA()) && (indexB = nextB())){
		atom &refA = molA->atoms[*indexA];
		atom &refB = molB->atoms[*indexB];

		double w;

		if(mass_weight){
			// The masses of each atom may be different, so use
			// a symeterized mass weighting
			w = 0.5*(refA.mass + refB.mass);
		}
		else{
			w = 1.0;
		}
		
		total_mass += w;

		double final[3];

		if(molA->current_step == 0){	
			tmp[X] = refA.coord[X];
			tmp[Y] = refA.coord[Y];
			tmp[Z] = refA.coord[Z];
		}
		else{
			float *traj = &(molA->traj[molA->current_step - 1][3*(*indexA)]);
			tmp[X] = traj[X];
			tmp[Y] = traj[Y];
			tmp[Z] = traj[Z];
		}

		ROTATE(final, mA, tmp);

		comA[X] += w*final[X];
		comA[Y] += w*final[Y];
		comA[Z] += w*final[Z];

		if(molB->current_step == 0){	
			tmp[X] = refB.coord[X];
			tmp[Y] = refB.coord[Y];
			tmp[Z] = refB.coord[Z];
		}
		else{
			float *traj = &(molB->traj[molB->current_step - 1][3*(*indexB)]);
			tmp[X] = traj[X];
			tmp[Y] = traj[Y];
			tmp[Z] = traj[Z];
		}

		ROTATE(final, mB, tmp);

		comB[X] += w*final[X];
		comB[Y] += w*final[Y];
		comB[Z] += w*final[Z];
	}
	
	comA[X] /= total_mass;
	comA[Y] /= total_mass;
	comA[Z] /= total_mass;
	
	ROTATE(tmp, mA, molA->R);
	TRANSLATE(comA, tmp);

	comB[X] /= total_mass;
	comB[Y] /= total_mass;
	comB[Z] /= total_mass;

	ROTATE(tmp, mB, molB->R);
	TRANSLATE(comB, tmp);

	nextA.Init(selectA);
	nextB.Init(selectB);

	int i, j, k;

	int small = X;
	int med = X;
	int large = X;

	double R[3][3], RtR[3][3], U[3][3];
	double A[3][3], B[3][3];
	double eigenVal[3];
	double mag;

	for(i = X;i <= Z;i++)
		for(j = X;j <= Z;j++)
			R[i][j] = RtR[i][j] = 0.0;

	while((indexA = nextA()) && (indexB = nextB())){
		atom &refA = molA->atoms[*indexA];
		atom &refB = molB->atoms[*indexB];

		double finalA[3], finalB[3];

		if(molA->current_step == 0){	
			tmp[X] = refA.coord[X];
			tmp[Y] = refA.coord[Y];
			tmp[Z] = refA.coord[Z];
		}
		else{
			float *traj = &(molA->traj[molA->current_step - 1][3*(*indexA)]);
			tmp[X] = traj[X];
			tmp[Y] = traj[Y];
			tmp[Z] = traj[Z];
		}

		TRANSLATE(tmp, molA->R);
		ROTATE(finalA, mA, tmp);
		
		// Remove the center of mass
		finalA[X] -= comA[X];
		finalA[Y] -= comA[Y];
		finalA[Z] -= comA[Z];

		if(molB->current_step == 0){	
			tmp[X] = refB.coord[X];
			tmp[Y] = refB.coord[Y];
			tmp[Z] = refB.coord[Z];
		}
		else{
			float *traj = &(molB->traj[molB->current_step - 1][3*(*indexB)]);
			tmp[X] = traj[X];
			tmp[Y] = traj[Y];
			tmp[Z] = traj[Z];
		}

		TRANSLATE(tmp, molB->R);
		ROTATE(finalB, mB, tmp);
		
		// Remove the center of mass
		finalB[X] -= comB[X];
		finalB[Y] -= comB[Y];
		finalB[Z] -= comB[Z];

		double w;

		if(mass_weight){
			// The masses of each atom may be different, so use
			// a symeterized mass weighting
			w = 0.5*(refA.mass + refB.mass);
		}
		else{
			w = 1.0;
		}

		R[X][X] += w*finalA[X]*finalB[X];
		R[X][Y] += w*finalA[X]*finalB[Y];
		R[X][Z] += w*finalA[X]*finalB[Z];

		R[Y][X] += w*finalA[Y]*finalB[X];
		R[Y][Y] += w*finalA[Y]*finalB[Y];
		R[Y][Z] += w*finalA[Y]*finalB[Z];

		R[Z][X] += w*finalA[Z]*finalB[X];
		R[Z][Y] += w*finalA[Z]*finalB[Y];
		R[Z][Z] += w*finalA[Z]*finalB[Z];
	}

	// RtR holds R^{transpose).R
	for(i = X;i <= Z;i++)
		for(j = X;j <= Z;j++)
			for(k = X;k <= Z;k++)
				RtR[i][j] += R[k][i]*R[k][j];

	// Compute the eigenvalues and eigen vectors of RtR.
	//RtR.Jacobi(A, eigenVal);
	if(EigenSystem3D(RtR, A, eigenVal) == FALSE)
		throw QMOL_ERROR("kabsch:: Unable to diagonalize matrix");

	// Test all of the eigen values to make sure that we have a valid
	// solution.Only skip structures where all of the atoms are colinear
	// (i.e. 2 zero eigenvalues).
	int nullsubspace = 0;

	for(i = X;i <= Z;i++){
		if(eigenVal[i] < ALIGNMENT_EPS){
			//throw QMOL_ERROR("Zero eigen value in kabsch: Colinear and "
			//	"coplanar molecules are not aligned.");
			nullsubspace ++;
		}
	}

	if(nullsubspace > 1){
		throw QMOL_ERROR("Zero eigen values in kabsch: Colinear molecules "
			"are not currently allowed.");
	}

	// Sort the eigen values (in case we need to remove an improper
	// rotation).
	for(i = Y;i <= Z;i++){
		if(fabs(eigenVal[i]) < fabs(eigenVal[small]))
			small = i;

		if(fabs(eigenVal[i]) > fabs(eigenVal[large]))
			large = i;
	}

	if((large + 1)%3 == small)
		med = (large + 2)%3;
	else
		med = (large + 1)%3;


	double smallVec[3];
	double medVec[3];
	double largeVec[3];

	for(i = X;i <= Z;i++){
		medVec[i] = A[i][med];
		largeVec[i] = A[i][large];
	}

	// *************************************************************************
	// If you get this error message, you need to get the most recent
	// patch for the MS compiler!
	// fatal error C1001: INTERNAL COMPILER ERROR
    // (compiler file 'E:\8168\vc98\p2\src\P2\main.c', line 494)
    // Please choose the Technical Support command on the Visual C++
    // Help menu, or open the Technical Support help file for more information
	// *************************************************************************

	// The cross product defines the eigen vector cooresponding to
	// the smallest eigen value.
	CROSS(smallVec, largeVec, medVec);

	for(i = X;i <= Z;i++)
		A[i][small] = smallVec[i];

	/* Invert the eigenvalues as called for in kabsch */
	if(nullsubspace){
		eigenVal[large] = 1.0/sqrt(eigenVal[large]);
		eigenVal[med] = 1.0/sqrt(eigenVal[med]);

		if(eigenVal[large]*eigenVal[med] < 0.0)
			eigenVal[med] *= -1.0;
	}
	else{
		for(i = X;i <= Z;i++)
			eigenVal[i] = 1.0/sqrt(eigenVal[i]);

		if(eigenVal[X]*eigenVal[Y]*eigenVal[Z] < 0.0)
			eigenVal[small] *= -1.0;
	}

    for(i = X;i <= Z;i++){
		for(j = X;j <= Z;j++){
			B[i][j] = R[j][X]*A[X][i]*eigenVal[i];

			for(k = Y;k <= Z;k++)
    			B[i][j] += R[j][k]*A[k][i]*eigenVal[i];
		}
	}

	for(i = X;i <= Z;i++){
		medVec[i] = B[med][i];
		largeVec[i] = B[large][i];
	}

	mag = 1.0/sqrt(DOT(medVec,medVec));

	medVec[X] *= mag;
	medVec[Y] *= mag;
	medVec[Z] *= mag;

	mag = 1.0/sqrt(DOT(largeVec,largeVec));

	largeVec[X] *= mag;
	largeVec[Y] *= mag;
	largeVec[Z] *= mag;

	// The cross product defines the eigen vector cooresponding to
	// the smallest eigen value.
	//smallVec = largeVec | medVec;
	CROSS(smallVec, largeVec, medVec);

	for(i = X;i <= Z;i++)
		B[small][i] = smallVec[i];

    for(i = X;i <= Z;i++){
		for(j = X;j <= Z;j++){
			U[i][j] = A[j][X]*B[X][i];

			for(k = Y;k <= Z;k++){
				U[i][j] += A[j][k]*B[k][i];
			}
		}
	}

	// Adjust the rotation matrix for molecule B. Note that we
	// must use the transpose of U since the above code was derived
	// for column, not row, vectors.
	// [(V + R)*M - Com_B]*U^T + Com_A = (V*U_p + R_p)*M
	//
	// Solving for U_p and R_p:
	// U_p = M*U^T*M^(-1)
	// R_p = [R - Com_B*M^(-1)]*U_p + Com_A*M^(-1)

	// Use mA as temporary storage
	for(i = X;i <= Z;i++){
		for(j = X;j <= Z;j++){
			mA[i][j] = mB[i][X]*U[j][X];

			for(k = Y;k <= Z;k++)
				mA[i][j] += mB[i][k]*U[j][k];
		}
	}

	double Up[3][3];

	InvertMatrix(mBInv, mB);

	for(i = X;i <= Z;i++){
		for(j = X;j <= Z;j++){
			Up[i][j] = mA[i][X]*mBInv[X][j];

			for(k = Y;k <= Z;k++)
				Up[i][j] += mA[i][k]*mBInv[k][j];
		}
	}
	
	// Adjust the coordinates of ALL the atoms in molecule B
	int num = molB->atoms.Length();

	for(i = 0;i < num;i++){
		float tmpCoord[3];
		float *coord;

		if(molB->current_step == 0){	
			coord = molB->atoms[i].coord;
		}
		else{
			coord = &(molB->traj[molB->current_step - 1][3*i]);
		}

		memcpy(tmpCoord, coord, 3*sizeof(float));

		ROTATE(coord, Up, tmpCoord);
    }

	// Compute the vector that will align the center of
	// mass of molecule B with the center of mass of molecule A.
	ROTATE(tmp, mBInv, comB);
	
	tmp[X] = molB->R[X] - tmp[X];
	tmp[Y] = molB->R[Y] - tmp[Y];
	tmp[Z] = molB->R[Z] - tmp[Z];
	
	ROTATE(molB->R, Up, tmp);

	ROTATE(tmp, mBInv, comA);

	TRANSLATE(molB->R, tmp);
}

using namespace std;

// Compute the eigenvectors and eigenvales for a 3X3 symmetric matrix.
// Return TRUE (i.e. 1) if the matrix has been successfully diagonalized
// or FALSE (i.e. 0) if matrix is singular.
int EigenSystem3D(double a[3][3], double vect[3][3], double val[3])
{
	// A small number
	const double eps = 1.0e-7;

	// The eigen vectors (returned as columns of the matrix)
	if(fabs(a[X][Z]) < eps){ // Don't divide by zero!
		
		if((fabs(a[X][Y]) < eps) || (fabs(a[Y][Z]) < eps)){
			if((fabs(a[X][Y]) < eps) && (fabs(a[Y][Z]) < eps)){
				// Matrix is diagonal!

				// Compute the eigenvalues
				val[X] = a[X][X];
				val[Y] = a[Y][Y];
				val[Z] = a[Z][Z];

				vect[X][X] = 1.0;
				vect[Y][X] = 0.0;
				vect[Z][X] = 0.0;

				vect[X][Y] = 0.0;
				vect[Y][Y] = 1.0;
				vect[Z][Y] = 0.0;

				vect[X][Z] = 0.0;
				vect[Y][Z] = 0.0;
				vect[Z][Z] = 1.0;

				// Success!
				return TRUE;
			}
			else{
				if(fabs(a[X][Y]) < eps){
					// Compute the eigenvalues
					val[X] = a[X][X];
					val[Y] = 0.5*(a[Y][Y] + a[Z][Z] - 
						sqrt(a[Y][Y]*a[Y][Y] + 4.0*a[Y][Z]*a[Y][Z] - 
						2.0*a[Y][Y]*a[Z][Z] + a[Z][Z]*a[Z][Z]));
					val[Z] =  0.5*(a[Y][Y] + a[Z][Z] +
						sqrt(a[Y][Y]*a[Y][Y] + 4.0*a[Y][Z]*a[Y][Z] - 
						2.0*a[Y][Y]*a[Z][Z] + a[Z][Z]*a[Z][Z]));

					vect[X][X] = 1.0;
					vect[Y][X] = 0.0;
					vect[Z][X] = 0.0;

					vect[X][Y] = 0.0;
					vect[Y][Y] = -0.5*(-a[Y][Y] + a[Z][Z] +
						sqrt(a[Y][Y]*a[Y][Y] + 4.0*a[Y][Z]*a[Y][Z] - 
						2.0*a[Y][Y]*a[Z][Z] + a[Z][Z]*a[Z][Z]))/a[Y][Z];
					vect[Z][Y] = 1.0;

					vect[X][Z] = 0.0;
					vect[Y][Z] = -0.5*(-a[Y][Y] + a[Z][Z] -
						sqrt(a[Y][Y]*a[Y][Y] + 4.0*a[Y][Z]*a[Y][Z] - 
						2.0*a[Y][Y]*a[Z][Z] + a[Z][Z]*a[Z][Z]))/a[Y][Z];
					vect[Z][Z] = 1.0;
				}
				else{ //(fabs(a[Y][Z]) < eps)
				
					// Compute the eigenvalues
					val[X] = 0.5*(a[X][X] + a[Y][Y] - sqrt(a[X][X]*a[X][X] + 4.0*a[X][Y]*a[X][Y] - 
						2.0*a[X][X]*a[Y][Y] + a[Y][Y]*a[Y][Y]));
					val[Y] = 0.5*(a[X][X] + a[Y][Y] + sqrt(a[X][X]*a[X][X] + 4.0*a[X][Y]*a[X][Y] - 
						2.0*a[X][X]*a[Y][Y] + a[Y][Y]*a[Y][Y]));
					val[Z] = a[Z][Z];

					vect[X][X] = -0.5*(-a[X][X] + a[Y][Y] + sqrt(a[X][X]*a[X][X] + 4.0*a[X][Y]*a[X][Y] - 
						2.0*a[X][X]*a[Y][Y] + a[Y][Y]*a[Y][Y]))/a[X][Y];
					vect[Y][X] = 1.0;
					vect[Z][X] = 0.0;

					vect[X][Y] = -0.5*(-a[X][X] + a[Y][Y] - sqrt(a[X][X]*a[X][X] + 4.0*a[X][Y]*a[X][Y] - 
						2.0*a[X][X]*a[Y][Y] + a[Y][Y]*a[Y][Y]))/a[X][Y];
					vect[Y][Y] = 1.0;
					vect[Z][Y] = 0.0;

					vect[X][Z] = 0.0;
					vect[Y][Z] = 0.0;
					vect[Z][Z] = 1.0;
				}
			}
		}
		else{
			// Compute eigenvalues
			EigenValues3D(a, val);

			vect[X][X] = -(a[Y][Z])/(a[X][Y]) + ((a[Y][Y]-val[X])*(-val[X]+a[Z][Z]))/(a[X][Y]*a[Y][Z]);
			vect[Y][X] = -(-val[X]+a[Z][Z])/(a[Y][Z]);
			vect[Z][X] = 1.0;

			vect[X][Y] =  -(a[Y][Z])/(a[X][Y]) + ((a[Y][Y]-val[Y])*(-val[Y]+a[Z][Z]))/(a[X][Y]*a[Y][Z]);
			vect[Y][Y] = -(-val[Y]+a[Z][Z])/(a[Y][Z]);
			vect[Z][Y] = 1.0;

			vect[X][Z] =  -(a[Y][Z])/(a[X][Y]) + ((a[Y][Y]-val[Z])*(-val[Z]+a[Z][Z]))/(a[X][Y]*a[Y][Z]);
			vect[Y][Z] = -(-val[Z]+a[Z][Z])/(a[Y][Z]);
			vect[Z][Z] = 1.0;
		}
	}
	else{
		// Compute eigenvalues
		EigenValues3D(a, val);

		vect[X][X] = -(-val[X] + a[Z][Z])/(a[X][Z]) + 
					  (a[Y][Z]*(a[X][Z]*a[Y][Z]-a[X][Y]*(-val[X]+a[Z][Z])))/
					  (a[X][Z]*(-a[X][Y]*a[Y][Z]+a[X][Z]*(a[Y][Y]-val[X])));
		vect[Y][X] = -(a[X][Z]*a[Y][Z]-a[X][Y]*(-val[X]+a[Z][Z]))/
					  (-a[X][Y]*a[Y][Z]+a[X][Z]*(a[Y][Y]-val[X]));
		vect[Z][X] = 1.0;

		vect[X][Y] = -(-val[Y]+a[Z][Z])/(a[X][Z]) + 
					  (a[Y][Z]*(a[X][Z]*a[Y][Z]-a[X][Y]*(-val[Y]+a[Z][Z])))/
					  (a[X][Z]*(-a[X][Y]*a[Y][Z]+a[X][Z]*(a[Y][Y]-val[Y])));
		vect[Y][Y] = -(a[X][Z]*a[Y][Z]-a[X][Y]*(-val[Y]+a[Z][Z]))/
					  (-a[X][Y]*a[Y][Z]+a[X][Z]*(a[Y][Y]-val[Y]));
		vect[Z][Y] = 1.0;

		vect[X][Z] = -(-val[Z]+a[Z][Z])/(a[X][Z]) + 
					  (a[Y][Z]*(a[X][Z]*a[Y][Z]-a[X][Y]*(-val[Z]+a[Z][Z])))/
					  (a[X][Z]*(-a[X][Y]*a[Y][Z]+a[X][Z]*(a[Y][Y]-val[Z])));
		vect[Y][Z] = -(a[X][Z]*a[Y][Z]-a[X][Y]*(-val[Z]+a[Z][Z]))/
					  (-a[X][Y]*a[Y][Z]+a[X][Z]*(a[Y][Y]-val[Z]));
		vect[Z][Z] = 1.0;
	}

	// Normalize the eigenvectors
	double mag;

	mag = vect[X][X]*vect[X][X] + vect[Y][X]*vect[Y][X] + vect[Z][X]*vect[Z][X];
	
	mag = sqrt(mag);

	if(mag < eps)
		return FALSE;

	mag = 1.0/mag;

	vect[X][X] *= mag;
	vect[Y][X] *= mag;
	vect[Z][X] *= mag;

	mag = vect[X][Y]*vect[X][Y] + vect[Y][Y]*vect[Y][Y] + vect[Z][Y]*vect[Z][Y];

	mag = sqrt(mag);

	if(mag < eps)
		return FALSE;

	mag = 1.0/mag;
	
	vect[X][Y] *= mag;
	vect[Y][Y] *= mag;
	vect[Z][Y] *= mag;

	mag = vect[X][Z]*vect[X][Z] + vect[Y][Z]*vect[Y][Z] + vect[Z][Z]*vect[Z][Z];

	mag = sqrt(mag);

	if(mag < eps)
		return FALSE;

	mag = 1.0/mag;
	
	vect[X][Z] *= mag;
	vect[Y][Z] *= mag;
	vect[Z][Z] *= mag;

	return TRUE;
}

// Compute the eigenvalues for a 3x3 symmetric real matrix
void EigenValues3D(double a[3][3], double val[3])
{
	// Trace of the matrix a
	double trace_a = a[X][X] + a[Y][Y] + a[Z][Z];

	// Determinant of the matrix a
	double det_a = -a[X][Z]*a[X][Z]*a[Y][Y] + 2*a[X][Y]*a[X][Z]*a[Y][Z] - 
					a[X][X]*a[Y][Z]*a[Y][Z] - a[X][Y]*a[X][Y]*a[Z][Z] + 
					a[X][X]*a[Y][Y]*a[Z][Z];

	// Determinant of the cofactors of the matrix a
	double det_c = -a[X][Y]*a[X][Y] - a[X][Z]*a[X][Z] + a[X][X]*a[Y][Y] - 
					a[Y][Z]*a[Y][Z] + a[X][X]*a[Z][Z] + a[Y][Y]*a[Z][Z];

	// Intermediate constants
	complex<double> foo = 3.0*det_c - trace_a*trace_a;
	complex<double> bar = 27.0*det_a - 9.0*det_c*trace_a + 2.0*trace_a*trace_a*trace_a;

	complex<double> z = bar + sqrt(bar*bar + 4.0*foo*foo*foo);
	complex<double> z_1_3 = pow(z, 1.0/3.0);

	const complex<double> I(0, 1);
	const double two_1_3 = pow(2.0, 1.0/3.0);
	const double two_2_3 = pow(2.0, 2.0/3.0);
	const double three_1_2 = sqrt(3.0);

	complex<double> tmp;
	
	// The eigenvalues
	tmp = trace_a/3.0 - two_1_3*foo/(3.0*z_1_3) + z_1_3/(3.0*two_1_3);
	val[X] = tmp.real();

	tmp = trace_a/3.0 + (1.0 + I*three_1_2)*foo/(3.0*two_2_3*z_1_3) - 
		(1.0 - I*three_1_2)*z_1_3/(6.0*two_1_3);
	val[Y] = tmp.real();

	tmp = trace_a/3.0 + (1.0 - I*three_1_2)*foo/(3.0*two_2_3*z_1_3) - 
		(1.0 + I*three_1_2)*z_1_3/(6.0*two_1_3);
	val[Z] = tmp.real();

}