//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.
//
//	Look-and-feel inspired by Xmol 
//	(http://www.networkcs.com/msc/docs/xmol/).
//
//	OpenGL selection code and quaterion rotation matrix
//	formulation from molview demo program by Mark Kilgard
//	(http://reality.sgi.com/opengl/OpenGLforX.html). PostScript
//	rendering also from Mark Kilgard.
//
//	The printing routines are based on code from
//	Craig Fahrnbach <craignan@home.com> via
//	Uwe Kotyczka <kotyczka@bnm-gmbh.de>.
//
//	Covalent bond determination code based on VMD
//	http://www.ks.uiuc.edu/Research/vmd/
//
//	Stride (http://www.embl-heidelberg.de/stride/stride.html)
//	is used for secondary structure determination and has been 
//	included in Qmol with the kind permission of 
//	Dmitrij Frishman, PhD
//	Institute for Bioinformatics
//	GSF - Forschungszentrum fr Umwelt und Gesundheit, GmbH
//	Ingolstdter Landstrae 1,
//	D-85764 Neuherberg, Germany
//	Telephone: +49-89-3187-4201
//	Fax: +49-89-31873585
//	e-mail: d.frishman@gsf.de
//	WWW: http://mips.gsf.de/mips/staff/frishman/

#include "stdafx.h"

#include "QmolView.h"
#include "SelectObject.h"

// Return the selected object
SelectObj CQmolView::pickScene(list<molecule *> &molList, int x, int y)
{
	// Do we have any atoms to pick?
	if(molList.IsEmpty())
		return SelectObj();
	
	// Compute the total number of selectable objects
	ListIterator<molecule *> next(molList);
	molecule **mol;
	int totalNumObjs = 0;

	while(mol = next()){
		totalNumObjs += (*mol)->NumAtoms();
		totalNumObjs += (*mol)->NumBonds();
	}

	// Allocate memory for the selection buffer
	// (buffer size is four times the total number of
	// selectable objects).
	int bufSize = 4*totalNumObjs;
	GLuint *selectionBuffer = new GLuint [bufSize];

	if(selectionBuffer == NULL)
		throw QmolError("Unable to allocate memory for selection!",
			__FILE__, __LINE__);

	GLint viewport[4];
  	float aspect;
  	int hits;

  	glSelectBuffer(bufSize, selectionBuffer);
  	glRenderMode(GL_SELECT);

	glInitNames();
	
	// Stick an element on the stack so we can call glLoadName()
	// later on.
	glPushName(0);

	// Set up the projection matrix for the select render mode
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	
	// The viewport
	viewport[0] = 0;
	viewport[1] = 0;
	viewport[2] = viewWidth;
	viewport[3] = viewHeight;

	// The size of the selection box, currently 10x10, is determined
	// here. This should probably be defined elsewhere.
	if(displayStereo == QM_STEREO_SPLIT_SCREEN){
		if(x < viewWidth/2){
			viewport[2] = viewWidth/2;
		}
		else{
			viewport[0] = viewWidth/2;
			viewport[2] = viewWidth/2;
		}
	}
	
	gluPickMatrix(x, viewport[3] - y, 10.0, 10.0, viewport);
	aspect = ((float) viewport[2])/viewport[3];

	gluPerspective(perspective, aspect, QM_NEAR_CLIP_Z, QM_FAR_CLIP_Z);
  	glMatrixMode(GL_MODELVIEW);
	
	RenderScene(molList, QM_RENDER_SELECT);
	
  	glMatrixMode(GL_PROJECTION);
  	glPopMatrix();
  	glMatrixMode(GL_MODELVIEW);
	
  	hits = glRenderMode(GL_RENDER);
		
  	SelectObj selectedObj = processHits(molList, hits, bufSize,
		selectionBuffer);

	if(selectionBuffer)
		delete [] selectionBuffer;

	return selectedObj;
}

SelectObj CQmolView::processHits(list<molecule *> &molList, GLint hits, 
						   int bufSize, GLuint *buffer)
{
	int i;
	GLint names;
	int match = -1;
	GLuint *ptr;
	GLuint minz, z;

	if(hits < 0 ){
		AfxMessageBox("WARNING: select buffer overflow!");
		return SelectObj();
	}

  	if(hits == 0)
		return SelectObj();
		
	minz = 0xffffffff;
	ptr = (GLuint *) buffer;
	
	for(i = 0; i < hits; i++){  /* for each hit  */
		names = *ptr;
		ptr++;
		z = *ptr;
		ptr++;
		ptr++;
		
		if(z <= minz){
			minz = z;
			match = *ptr;
		}
		ptr += names;
	}
	
	
	if(match == -1)
		throw QMOL_ERROR("Selection buffer underflow -> match not found!");
	
	if(match >= bufSize)
		throw QMOL_ERROR("Selection buffer overflow!");
	
	// Now we need to determine which object was selected and which
	// molecule it belongs to
	SelectObj matchObj;

	ListIterator<molecule *> next(molList);
	molecule **mol;

	while(mol = next()){
		int numAtoms = (*mol)->NumAtoms();
		int numObjs = numAtoms + (*mol)->NumBonds();

		if(match < numObjs){
			// The object is in the current molecule
			matchObj.SetMol(*mol);

			if(match < numAtoms){
				// The object is an atom
				matchObj.SetType(SelectObj::ATOM_OBJ);
				matchObj.SetIndex(match);
			}
			else{
				// The object is a bond
				matchObj.SetType(SelectObj::BOND_OBJ);
				matchObj.SetIndex(match - numAtoms);
			}

			break;
		}

		match -= numObjs;
	}

	return matchObj;
}
