//	(c) 2000-2002 J. Gans (jdg9@cornell.edu), Shalloway Lab, 
//	Cornell University. You can modify and freely distribute 
//	this code, but please keep this header intact.
//
//	Look-and-feel inspired by Xmol 
//	(http://www.networkcs.com/msc/docs/xmol/).
//
//	OpenGL selection code and quaterion rotation matrix
//	formulation from molview demo program by Mark Kilgard
//	(http://reality.sgi.com/opengl/OpenGLforX.html). PostScript
//	rendering also from Mark Kilgard.
//
//	The printing routines are based on code from
//	Craig Fahrnbach <craignan@home.com> via
//	Uwe Kotyczka <kotyczka@bnm-gmbh.de>.
//
//	Covalent bond determination code based on VMD
//	http://www.ks.uiuc.edu/Research/vmd/
//
//	Stride (http://www.embl-heidelberg.de/stride/stride.html)
//	is used for secondary structure determination and has been 
//	included in Qmol with the kind permission of 
//	Dmitrij Frishman, PhD
//	Institute for Bioinformatics
//	GSF - Forschungszentrum fr Umwelt und Gesundheit, GmbH
//	Ingolstdter Landstrae 1,
//	D-85764 Neuherberg, Germany
//	Telephone: +49-89-3187-4201
//	Fax: +49-89-31873585
//	e-mail: d.frishman@gsf.de
//	WWW: http://mips.gsf.de/mips/staff/frishman/

//////////////////////////////////////////////////////////////////
// Classes and functions for computing molecular surfaces
//////////////////////////////////////////////////////////////////
#ifndef __QMOL_SURFACE
#define __QMOL_SURFACE

struct Triangle{
	Triangle()
	{
		vertexPtr[0] = vertexData[0];
		vertexPtr[1] = vertexData[1];
		vertexPtr[2] = vertexData[2];

		vertexIndex[0] = -1;
		vertexIndex[1] = -1;
		vertexIndex[2] = -1;
	};
	
	GLint vertexIndex[3];
	GLfloat* vertexPtr[3];
	GLfloat vertexData[3][3];

	void SetVertex(int i, GLfloat v[3])
	{
		memcpy(vertexData[i], v, 3*sizeof(GLfloat));
	};

	Triangle& operator=(const Triangle &copy)
	{
		vertexIndex[0] = copy.vertexIndex[0];
		vertexIndex[1] = copy.vertexIndex[1];
		vertexIndex[2] = copy.vertexIndex[2];

		memcpy(vertexData[0], copy.vertexData[0], 3*sizeof(GLfloat));
		memcpy(vertexData[1], copy.vertexData[1], 3*sizeof(GLfloat));
		memcpy(vertexData[2], copy.vertexData[2], 3*sizeof(GLfloat));

		// Don't copy the vertexPtr array -- it always points to 
		// vertexData unless specifically changed.

		return (*this);
	};
};

struct VertexInfo{
	Triangle* parent;
	GLint index;
};

// A forward definition of atom for the MolSurface class
class atom;

class MolSurface{
private:

	BOOL validSurface;

	array<atom> *ptrAtoms;
	list <array <float> > *ptrTraj;

	GLfloat fTargetValue;
	GLfloat targetValueCutOff;
	GLfloat probeSphere;
	GLfloat meshSize;

	// Which structure are we generating a surface for?
	// traj_index == -1 if primary structure selected.
	int traj_index;

	int surfaceType; // The type of molecular surface. Currently 
					 // HEURISTIC

	int renderStyle; // The rendering style for the surface. Currently
					 // either SOLID, MESH or POINT
	
	// Should solvent atoms be included when rendering a molecular surface?
	BOOL include_solvent_atoms;

	GLfloat boxLength;
	int maxNumBox;
	GLfloat maxLength;
	array< list<int>* > atomBox;
	array< GLfloat > cubeBox;
	array< list<VertexInfo>* > vertexBox;
	list<Triangle> triangleList;

	array< int > indexArray;
	array< GLfloat > vertexArray;
	array< GLfloat > normalArray;
	array< int > colorArray;

	GLfloat boxSize[3];
	GLfloat invBoxSize[3];
	int xytotb, totb;	// Number of grid cells in atom box

	int cubeSize[3];
	int cubeXY;

	GLfloat molMin[3];
	GLfloat molMax[3];
	GLint numBox[3];
	
	BOOL color_by_atom;
	GLubyte transparency;
	GLubyte color[4];

	GLvoid vMarchCube(GLfloat fX, GLfloat fY, GLfloat fZ, 
		int xIndex, int yIndex, int zIndex, GLfloat fScale);

	// fGetOffset finds the approximate point of intersection of the surface
	// between two points with the values fValue1 and fValue2.
	inline GLfloat fGetOffset(GLfloat fValue1, GLfloat fValue2, GLfloat fValueDesired)
	{
		GLfloat fDelta = fValue2 - fValue1;

        if(fDelta == 0.0f)
        {
                return 0.5f;
        }
        return (fValueDesired - fValue1)/fDelta;
	};

	inline GLfloat ScalarField(GLfloat fX, GLfloat fY, GLfloat fZ)
	{
		return ScalarFieldHeuristic(fX, fY, fZ);
	};

	inline void GradientAndColorIndex(GLfloat r[3], GLfloat grad[3], int &colorIndex)
	{
		GradientAndColorIndexHeuristic(r, grad, colorIndex);
	};

	GLfloat ScalarFieldHeuristic(GLfloat fX, GLfloat fY, GLfloat fZ);
	void GradientAndColorIndexHeuristic(GLfloat r[3], GLfloat grad[3], int &colorIndex);

	void BuildAtomBox();

	void DeleteAtomBox();

	void BuildVertexArray();

	float TriangleArea(int a, int b, int c);
public:

	// The different surface types
	enum {HEURISTIC};
	
	// The different rendering styles
	enum {SOLID, MESH, POINT};

	MolSurface();
	~MolSurface();

	inline void InvalidateSurface()
	{
		validSurface = FALSE;

		indexArray.Alloc(0);
		vertexArray.Alloc(0);
		normalArray.Alloc(0);
		colorArray.Alloc(0);
	};

	inline BOOL ValidSurface()
	{
		return validSurface;
	};

	inline GLfloat GetProbeSphereRadius()
	{
		return probeSphere;
	};

	inline void SetProbeSphereRadius(float m_radius)
	{
		probeSphere = m_radius;

		// Update the box length
		boxLength = 3.0f + probeSphere*probeSphere;
	};

	int GetSurfQuality();
	void SetSurfQuality(int m_quality);

	BOOL BuildSurface(array<atom> *m_ptrAtoms, 
		list <array <float> > *m_ptrTraj, int m_currentStep);

	void RenderSurface(array<atom> &m_atoms);

	inline void ColorByAtom(BOOL m_b)
	{
		color_by_atom = m_b;
	};

	inline int ColorByAtom()
	{
		return color_by_atom;
	};

	inline void SetTransparency(int m_a)
	{
		transparency = m_a;
	};

	inline int GetTransparency()
	{
		return transparency;
	};

	inline GetSurfColor(int &m_r, int &m_g, int &m_b, int &m_a)
	{
		m_r = color[0];
		m_g = color[1];
		m_b = color[2];
		m_a = color[3];
	};

	inline SetSurfColor(int m_r, int m_g, int m_b, int m_a)
	{
		color[0] = m_r;
		color[1] = m_g;
		color[2] = m_b;
		color[3] = m_a;
	};

	inline int GetSurfType()
	{
		return surfaceType;
	};

	inline void SetSurfType(int m_type)
	{
		surfaceType = m_type;
	};

	inline int GetRenderStyle()
	{
		return renderStyle;
	};

	inline void SetRenderStyle(int m_style)
	{
		if((m_style < SOLID) || (m_style > POINT))
			throw QmolError("Invalid Render Type", __FILE__, __LINE__);

		renderStyle = m_style;
	}

	void GetAtomSurfaceArea(array<float> &m_area);

	inline BOOL UseSolventAtoms()
	{
		return include_solvent_atoms;
	};

	inline void UseSolventAtoms(BOOL m_use_solvent)
	{
		include_solvent_atoms = m_use_solvent;
	};
};

#endif // __QMOL_SURFACE